#ifndef DIAG_TYPES_H
#define DIAG_TYPES_H

/*===========================================================================

                   S T A N D A R D    D E C L A R A T I O N S

DESCRIPTION
  This header file contains general types and macros that are of use
  to all modules.  The values or definitions are dependent on the specified
  target.  T_WINNT specifies Windows NT based targets, otherwise the
  default is for ARM targets.

       T_WINNT  Software is hosted on an NT platforn, triggers macro and
                type definitions, unlike definition above which triggers
                actual OS calls

DEFINED TYPES

       Name      Definition
       -------   --------------------------------------------------------
       byte      8  bit unsigned value
       word      16 bit unsigned value
       dword     32 bit unsigned value

       uint1     byte
       uint2     word
       uint4     dword

       uint8     8  bit unsigned value
       uint16    16 bit unsigned value
       uint32    32 bit unsigned value
       uint64    64 bit unsigned value

       boolean   8 bit boolean value

DEFINED CONSTANTS

       Name      Definition
       -------   --------------------------------------------------------
       TRUE      Asserted boolean condition (Logical 1)
       FALSE     Deasserted boolean condition (Logical 0)

       ON        Asserted condition
       OFF       Deasserted condition

       NULL      Pointer to nothing

       PACKED    Used to indicate structures which should use packed
                 alignment
       
  -----------------------------------------------------------------------------
  Copyright (c) 1990, 1992-1994, 1997-2007 QUALCOMM Incorporated.
  All Rights Reserved. QUALCOMM Proprietary and Confidential.
  -----------------------------------------------------------------------------
===========================================================================*/


/*===========================================================================

                      EDIT HISTORY FOR FILE

This section contains comments describing changes made to this file.
Notice that changes are listed in reverse chronological order.

$Header: //linux/pkgs/proprietary/oncrpc/main/source/inc/comdef.h#3 $

when       who     what, where, why
--------   ---     ----------------------------------------------------------
10/01/08   sj	   Created this file out of comdef.h for diag related data types only 
===========================================================================*/


/*===========================================================================

                            Data Declarations

===========================================================================*/

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------
** Constants
** ------------------------------------------------------------------------ */

#ifdef TRUE
#undef TRUE
#endif

#ifdef FALSE
#undef FALSE
#endif

#define TRUE   1   /* Boolean true value. */
#define FALSE  0   /* Boolean false value. */

#ifndef NULL
  #define NULL  0
#endif

/* -----------------------------------------------------------------------
** Standard Types
** ----------------------------------------------------------------------- */

/* The following definitions are the same accross platforms.  This first
** group are the sanctioned types.
*/
#ifndef _ARM_ASM_
typedef  unsigned char      boolean;     /* Boolean value type. */

typedef  unsigned long int  uint32;      /* Unsigned 32 bit value */
typedef  unsigned short     uint16;      /* Unsigned 16 bit value */
typedef  unsigned char      uint8;       /* Unsigned 8  bit value */

typedef  signed long int    int32;       /* Signed 32 bit value */
typedef  signed short       int16;       /* Signed 16 bit value */
typedef  signed char        int8;        /* Signed 8  bit value */

/* This group are the deprecated types.  Their use should be
** discontinued and new code should use the types above
*/
typedef  unsigned char     byte;         /* Unsigned 8  bit value type. */
typedef  unsigned short    word;         /* Unsinged 16 bit value type. */
typedef  unsigned long     dword;        /* Unsigned 32 bit value type. */

/* ---------------------------------------------------------------------
** Compiler Keyword Macros
** --------------------------------------------------------------------- */
/************************** The PACK() macro *****************************

 

  This block sets up the semantics for PACK() macro based upon

  upon the target compiler.

 

  PACK() is necessary to ensure portability of C variable/struct/union

  packing across many platforms.  For example, ARM compilers require the

  following:

    typedef __packed struct { ... } foo_t;

 

  But GCC requires this to achieve the same effect:

    typedef struct __attribute((__packed__)) struct { ... } foo_t;

 

  And, of course, Microsoft VC++ requires an alignment #pragma prologue

  and epilogue.

 

  To satisfy all three, the following form is recommended:

 

    #ifdef _WIN32

    #pragma pack(push,1) // Save previous, and turn on 1 byte alignment

    #endif

 

    typedef PACK(struct)

    {

      ...

    } such_and_such_t;

 

    typedef PACK(struct)

    {

      ...

    } this_and_that_t;

 

    typedef PACK(struct)

    {

      PACK(struct)

      {

        ...

      } hdr;

      PACK(union)

      {

        such_and_such_t sas;

        this_and_that_t tat;

      } payload;

    } cmd_t;

 

    #ifdef _WIN32

    #pragma pack(pop) // Revert alignment to what it was previously

    #endif

 

*********************** BEGIN PACK() Definition ***************************/

#if defined __GNUC__

  #define PACK(x)       x __attribute__((__packed__))

#elif defined __arm

  #define PACK(x)       __packed x

#elif defined _WIN32

  #define PACK(x)       x /* Microsoft uses #pragma pack() prologue/epilogue */

#else

  #error No PACK() macro defined for this compiler

#endif

/********************** END PACK() Definition *****************************/


#if (! defined T_WINNT) && (! defined TARGET_OS_SOLARIS)
	#ifndef SWIG  /* The SWIG preprocessor gets confused by these */
	#if defined(__ARMCC_VERSION) 
	  #define PACKED __packed
	#else  /* __GNUC__ */
	  #define PACKED 
	#endif /* defined (__GNUC__) */
	#endif /* SWIG */

#else /* T_WINNT || TARGET_OS_SOLARIS */

  /* WINNT or SOLARIS based targets
  */
  #define PACKED

#endif /* T_WINNT */

#endif // #ifndef _ARM_ASM_

/* ----------------------------------------------------------------------
**                          STANDARD MACROS
** ---------------------------------------------------------------------- */

#ifndef SWIG /* these confuse the SWIG preprocessor and aren't needed for it */


/*===========================================================================

MACRO FPOS

DESCRIPTION
  This macro computes the offset, in bytes, of a specified field
  of a specified structure or union type.

PARAMETERS
  type          type of the structure or union
  field         field in the structure or union to get the offset of

DEPENDENCIES
  None

RETURN VALUE
  The byte offset of the 'field' in the structure or union of type 'type'.

SIDE EFFECTS
  The lint error "Warning 545: Suspicious use of &" is suppressed within
  this macro.  This is due to the desire to have lint not complain when
  'field' is an array.

===========================================================================*/

#define FPOS( type, field ) \
    /*lint -e545 */ ( (dword) &(( type *) 0)-> field ) /*lint +e545 */



/*===========================================================================

MACRO FSIZ

DESCRIPTION
  This macro computes the size, in bytes, of a specified field
  of a specified structure or union type.

PARAMETERS
  type          type of the structure or union
  field         field in the structure or union to get the size of

DEPENDENCIES
  None

RETURN VALUE
  size in bytes of the 'field' in a structure or union of type 'type'

SIDE EFFECTS
  None

===========================================================================*/

#define FSIZ( type, field ) sizeof( ((type *) 0)->field )

/*===========================================================================

MACRO MAX
MACRO MIN

DESCRIPTION
  Evaluate the maximum/minimum of 2 specified arguments.

PARAMETERS
  x     parameter to compare to 'y'
  y     parameter to compare to 'x'

DEPENDENCIES
  'x' and 'y' are referenced multiple times, and should remain the same
  value each time they are evaluated.

RETURN VALUE
  MAX   greater of 'x' and 'y'
  MIN   lesser of 'x' and 'y'

SIDE EFFECTS
  None

===========================================================================*/
#ifndef MIN
   #define  MIN( x, y ) ( ((x) < (y)) ? (x) : (y) )
#endif

/* -------------------------------------------------------------------------
**                       Debug Declarations
** ------------------------------------------------------------------------- */


/*===========================================================================

                      FUNCTION DECLARATIONS

===========================================================================*/



#endif /* SWIG */

#ifdef __cplusplus
}
#endif
#endif  /* COMDEF_H */
