/* Shared library add-on to iptables for DSCP
 *
 * (C) 2002 by Harald Welte <laforge@gnumonks.org>
 *
 * This program is distributed under the terms of GNU GPL v2, 1991
 *
 * libipt_dscp.c borrowed heavily from libipt_tos.c
 *
 * --class support added by Iain Barnes
 * 
 * For a list of DSCP codepoints see 
 * http://www.iana.org/assignments/dscp-registry
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_dscprange.h>

/* This is evil, but it's my code - HW*/
#include "libipt_dscp_helper.c"

static void init(struct ipt_entry_match *m, unsigned int *nfcache) 
{
	*nfcache |= NFC_IP_TOS;
}

static void help(void) 
{
	printf(
        "DSCP range match v%s options\n"
        "[!] --dscp min-max		Match DSCP range codepoint with numerical value\n"
        "  		                This range can be in decimal (ex: 32)\n"
        "               		or in hex (ex: 0x20)\n"
        "\n"
        "				These two options are mutually exclusive !\n"
				, IPTABLES_VERSION
);
}

static struct option opts[] = {
	{ "dscprange", 1, 0, 'F' },
	//{ "dscp-class", 1, 0, 'G' },
	{ 0 }
};

static void
parse_dscprange(const unsigned char *s, struct ipt_dscp_range_info *dinfo)
{
	unsigned int dscp_min = 0;
    unsigned int dscp_max = 0;
    char *str = NULL;
    char *dash = NULL;

    if ((NULL == s) || (NULL == dinfo))
    {
		exit_error(OTHER_PROBLEM, "error, no enough space to store it.\r\n");
    }

    dash = strchr(s, '-');
    if (dash)
    {
        *dash = '\0';
    }

    str = s;
       
	if (string_to_number(str, 0, 255, &dscp_min) == -1)
	{
		exit_error(PARAMETER_PROBLEM, "Invalid dscp min `%s'\n", str);
	}
    
    if (dash)
    {
        str = dash + 1;
        if (string_to_number(str, 0, 255, &dscp_max) == -1)
        {
            exit_error(PARAMETER_PROBLEM, "Invalid dscp max `%s'\n", str);
        }
    }
    else
    {
        dscp_max = dscp_min;
    }

	if ((dscp_min > IPT_DSCP_MASK) || (dscp_max > IPT_DSCP_MASK) || (dscp_min > dscp_max))
	{
		exit_error(PARAMETER_PROBLEM, "DSCP `%d`-'%d' error: (out of range or min>max)\n", 
            dscp_min, dscp_max);
	}

	dinfo->dscp_min = (u_int8_t )dscp_min;
    dinfo->dscp_max = (u_int8_t)dscp_max;
	return;
}


#if 0
static void
parse_class(const char *s, struct ipt_dscp_info *dinfo)
{
	unsigned int dscp = class_to_dscp(s);

	/* Assign the value */
	dinfo->dscp = (u_int8_t)dscp;
}
#endif


static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_dscp_range_info *dinfo
		= (struct ipt_dscp_range_info *)(*match)->data;

	switch (c) {
	case 'F':
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
			           "DSCP match: Only use --dscp ONCE!");
		check_inverse(optarg, &invert, &optind, 0);
		parse_dscprange(argv[optind-1], dinfo);
		if (invert)
			dinfo->invert = 1;
		*flags = 1;
		break;
        
#if 0
	case 'G':
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
					"DSCP match: Only use --dscp-class ONCE!");
		check_inverse(optarg, &invert, &optind, 0);
		parse_class(argv[optind - 1], dinfo);
		if (invert)
			dinfo->invert = 1;
		*flags = 1;
		break;
#endif

	default:
		return 0;
	}

	return 1;
}

static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
		           "DSCP range match: Parameter --dscprange is required");
}

static void
print_dscprange(const struct ipt_dscp_range_info  *dscp, int numeric)
{

    if (NULL == dscp)
    {
        printf("Null");
    }
    else
    {
     	printf("0x%02x ", dscp->dscp_min);
        printf("- ");
        printf("0x%02x ", dscp->dscp_max);
    }
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	const struct ipt_dscp_range_info *dinfo =
		(const struct ipt_dscp_range_info *)match->data;
	printf("DSCP range match ");
    if (dinfo->invert)
        printf(" ! ");
	print_dscprange(dinfo, numeric);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	const struct ipt_dscp_range_info *dinfo =
		(const struct ipt_dscp_range_info *)match->data;

    if (dinfo->invert)
        printf(" ! ");
	printf("--dscprange ");
	print_dscprange(dinfo, 1);
}

static
struct iptables_match dscp
= { NULL,
    "dscprange",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_dscp_range_info)),
    IPT_ALIGN(sizeof(struct ipt_dscp_range_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&dscp);
}
