/* Shared library add-on to iptables to add MAC address support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_mac_range.h>



/* Function which prints out usage message. */
static void
help(void)
{
	printf(
        "MACRANGE v%s options:\n"
        "[!] --srcmac-range XX:XX:XX:XX:XX:XX-YY:YY:YY:YY:YY:YY\n"
        "				Match source MAC address range\n"
        "[!] --dstmac-range  XX:XX:XX:XX:XX:XX-YY:YY:YY:YY:YY:YY\n"
        "               Match dest MAC address range\n"
        "\n", IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "srcmac-range",  1, 0, '1' },
    { "dstmac-range", 1, 0, '2' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}

static int trans(const char *arg, unsigned char mac[])
{
    unsigned int i = 0;

    if (strlen(arg) != ETH_ALEN*3 - 1)
    {
        printf("error: %s error", arg);
        return -1;
    }

    for (i = 0; i < ETH_ALEN; i++)
    {
        long value;
        char *end;
        value = strtol(arg + i*3, &end, 16);

        if ((end == arg + i*3 + 2) && (value >= 0) && (value <= 255))
        {
            mac[i] = value;
        }
        else
        {
            exit_error(PARAMETER_PROBLEM, "\r\nBad mac address `%s'", arg);
            return -1;
        }
    }

    return 0;
}


static void
parse_mac(char *arg, struct ipt_macrange *info)
{
	unsigned int i = 0;
    char *dash = NULL;
    char *str = NULL;
    unsigned char macaddr[ETH_ALEN] = {0};


    if (NULL == arg)
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad mac address\r\n");
    }


	if (strlen(arg) != ((ETH_ALEN*3-1) * 2 + 1))
    {
		exit_error(PARAMETER_PROBLEM, "\r\nBad mac address `%s' len: %d and need:%d", 
            arg, strlen(arg), ((ETH_ALEN*3 - 1)*2 + 1));
    }

    dash = strchr(arg, '-');
    if (dash)
        *dash = '\0';

    str = arg;
    if (0 != trans(str, macaddr))
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad mac address `%s'", arg);
    }
    
    memcpy(info->min_mac, macaddr, ETH_ALEN);

    if (dash)
    {
        str = dash + 1;
        if (0 != trans(str, macaddr))
        {
            exit_error(PARAMETER_PROBLEM, "\r\nBad mac address `%s'", str);
        }
        memcpy(info->max_mac, macaddr, ETH_ALEN);
    }
    else
    {
        memcpy(info->max_mac, macaddr, ETH_ALEN);
    }

    if (memcmp(info->min_mac, info->max_mac, ETH_ALEN) > 0)
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad mac address, the range is invalid\r\n");
    }

    /*
	for (i = 0; i < ETH_ALEN; i++) {
		long number;
		char *end;

		number = strtol(mac + i*3, &end, 16);

		if (end == mac + i*3 + 2
		    && number >= 0
		    && number <= 255)
			info->srcaddr[i] = number;
		else
			exit_error(PARAMETER_PROBLEM,
				   "Bad mac address `%s'", mac);
	}
    */
}

static void print_macrange(const struct ipt_macrange *range);

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_macrange_info *macinfo = (struct ipt_macrange_info *)(*match)->data;

	switch (c) {
	    case '1':
            if (*flags & MACRANGE_SRC)
            {
                exit_error(PARAMETER_PROBLEM, 
                    "macrange match: Only use --src-range Once!");
            }
            *flags |= MACRANGE_SRC;
            macinfo->flags |= MACRANGE_SRC;
#if 0
            check_inverse(optarg, &invert, &optind, 0);
            if (invert)
            {
                macinfo->flags |= MACRANGE_SRC_INV;
                printf("set invert......\r\n");
            }
#endif
            parse_mac(optarg, &macinfo->src);

            break;

        case '2':
            if (*flags & MACRANGE_DST)
            {
                exit_error(PARAMETER_PROBLEM,
                    "\r\nmacrange match: Only use --dst-range Once!");
            }
            *flags |= MACRANGE_DST;
            macinfo->flags |= MACRANGE_DST;
#if 0
		    check_inverse(optarg, &invert, &optind, 0);
		    if (invert)
            {
			    macinfo->flags |= MACRANGE_DST_INV;
                printf("\r\nmarang dst invert......\r\n");
            }
#endif
            parse_mac(optarg, &macinfo->dst);

            //*flags = 1;
		    break;

	    default:
		    return 0;
	}

	return 1;
}

static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "\r\nmacrange match: You must specify `--src-range' or `--dst-range'\r\n");
}


static void print_macrange(const struct ipt_macrange *range)
{
    if (NULL == range)
    {
        printf("nothing ");
        return ;
    }
    
	printf(" %02x:%02x:%02x:%02x:%02x:%02x-%02x:%02x:%02x:%02x:%02x:%02x ",
        range->min_mac[0], range->min_mac[1], range->min_mac[2],
        range->min_mac[3], range->min_mac[4], range->min_mac[5],
        range->max_mac[0], range->max_mac[1], range->max_mac[2],
        range->max_mac[3], range->max_mac[4], range->max_mac[5]);
}

static void print(const struct ipt_ip *ip, 
    const struct ipt_entry_match *match,
    int numeric)
{
    struct ipt_macrange_info *info = (struct ipt_macrange_info *)match->data;


    if (info->flags & MACRANGE_SRC)
    {
        printf("sMAC range ");
        if (info->flags & MACRANGE_SRC_INV)
        {
            printf("! ");
        }
         print_macrange(&info->src);
    }

    if (info->flags & MACRANGE_DST)
    {
        printf("dMAC range ");
        if (info->flags & MACRANGE_DST_INV)
        {
            printf("! ");
        }
        print_macrange(&info->dst);
    }
}


/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
    struct ipt_macrange_info *info = (struct ipt_macrange_info *)match->data;

    if (info->flags & MACRANGE_SRC)
    {
        if (info->flags & MACRANGE_SRC_INV)
        {
            printf("! ");
        }
        printf("--srcmac-range ");
         print_macrange(&info->src);
         if (info->flags & MACRANGE_DST)
             fputc(' ', stdout);
    }
    if (info->flags & MACRANGE_DST)
    {
        if (info->flags & MACRANGE_DST_INV)
        {
            printf("! ");
        }
        printf("--dstmac-range ");
        print_macrange(&info->dst);
    }
}

static
struct iptables_match macrange
= { NULL,
    "macrange",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_macrange_info)),
    IPT_ALIGN(sizeof(struct ipt_macrange_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&macrange);
}
