/* Shared library add-on to iptables to add TOS matching support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_tosrange.h>

/* TOS names and values. */
static
struct TOS_value
{
	unsigned char TOS;
	const char *name;
} TOS_values[] = {
	{ IPTOS_LOWDELAY,    "Minimize-Delay" },
	{ IPTOS_THROUGHPUT,  "Maximize-Throughput" },
	{ IPTOS_RELIABILITY, "Maximize-Reliability" },
	{ IPTOS_MINCOST,     "Minimize-Cost" },
	{ IPTOS_NORMALSVC,   "Normal-Service" },
};

/* Function which prints out usage message. */
static void
help(void)
{
	unsigned int i;

	printf(
        "TOS range match v%s options:\n"
        "[!] --tosrange min - max                 Match Type of Service field from range of the\n"
        "                                following numeric or descriptive values:\n",
        IPTABLES_VERSION);

	for (i = 0; i < sizeof(TOS_values)/sizeof(struct TOS_value);i++)
		printf("                                     %s %u (0x%02x)\n",
		       TOS_values[i].name,
                       TOS_values[i].TOS,
                       TOS_values[i].TOS);
	fputc('\n', stdout);
}

static struct option opts[] = {
	{ "tosrange", 1, 0, '1' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	*nfcache |= NFC_IP_TOS;
}

static void
parse_tosrange(const unsigned char *s, struct ipt_tos_range_info *info)
{
	unsigned int tos_min = 0;
    unsigned int tos_max = 0;
    char *dash = NULL;
    char *str = s;

    if ((NULL == s) || (NULL == info))
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad tos value\r\n");
    }

    dash = strchr(s, '-');
    if (dash)
    {
        *dash = '\0';
    }

    if (string_to_number(str, 0, 255, &tos_min) == -1)
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad tos value\r\n");
    }


    if (dash)
    {
        
        str = dash + 1;
        if (string_to_number(str, 0, 255, &tos_max) == -1)
        {
            exit_error(PARAMETER_PROBLEM, "\r\nBad tos value\r\n");
        }
    }
    else
    {
        tos_max = tos_min;
    }


    if ((tos_min > IPT_TOS_MAX_VALUE) || (tos_max > IPT_TOS_MAX_VALUE) || (tos_min > tos_max))
    {
        exit_error(PARAMETER_PROBLEM, "\r\nBad tos value, min (0x%02x) > max (0x%02x)\r\n", 
            tos_min, tos_max);
        return ;
    }

    info->tos_min = tos_min;
    info->tos_max = tos_max;
    
    return ;

#if 0
	if (string_to_number(s, 0, 255, &tos) != -1) {
		if (tos == IPTOS_LOWDELAY
		    || tos == IPTOS_THROUGHPUT
		    || tos == IPTOS_RELIABILITY
		    || tos == IPTOS_MINCOST
		    || tos == IPTOS_NORMALSVC) {
		    	info->tos = (u_int8_t )tos;
		    	return;
		}
	} else {
		for (i = 0; i<sizeof(TOS_values)/sizeof(struct TOS_value); i++)
			if (strcasecmp(s,TOS_values[i].name) == 0) {
				info->tos = TOS_values[i].TOS;
				return;
			}
	}
	exit_error(PARAMETER_PROBLEM, "Bad TOS value `%s'", s);
#endif
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_tos_range_info *tosinfo = (struct ipt_tos_range_info *)(*match)->data;

	switch (c) {
	case '1':
		check_inverse(optarg, &invert, &optind, 0);
		parse_tosrange(argv[optind-1], tosinfo);
		if (invert)
			tosinfo->invert = 1;
		*flags = 1;
		break;

	default:
		return 0;
	}
	return 1;
}

static void print_tosrange(const struct ipt_tos_range_info *tosinfo, int numeric)
{
    unsigned int i = 0;
    if (!numeric)
    {
        for (i = 0; i < sizeof(TOS_values)/sizeof(struct TOS_value); i++)
        {
            if (tosinfo->tos_min & TOS_values[i].TOS)
            {
                printf("%s ", TOS_values[i].name);
            }
        }
        printf("(min: 0x%02x) ", tosinfo->tos_min);

        for (i = 0; i < sizeof(TOS_values)/sizeof(struct TOS_value); i++)
        {
            if (tosinfo->tos_max & TOS_values[i].TOS)
            {
                printf("%s ", TOS_values[i].name);
            }
        }
        printf("(max: 0x%02x) ", tosinfo->tos_max);
    }
#if 0
	unsigned int i;

	if (!numeric) {
		for (i = 0; i<sizeof(TOS_values)/sizeof(struct TOS_value); i++)
			if (TOS_values[i].TOS == tos) {
				printf("%s ", TOS_values[i].name);
				return;
			}
	}
	printf("0x%02x ", tos);
#endif
}

/* Final check; must have specified --tos. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "TOS RANGE match: You must specify `--tosrange'");
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	const struct ipt_tos_range_info *info = (const struct ipt_tos_range_info *)match->data;
    
	printf("TOS range match ");
	if (info->invert)
		printf("!");
	print_tosrange(info, numeric);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	const struct ipt_tos_range_info *info = (const struct ipt_tos_range_info *)match->data;
    
	if (info->invert)
		printf("! ");
	printf("--tosrange ");
	print_tosrange(info, 0);
}

static
struct iptables_match tosrange
= { NULL,
    "tosrange",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_tos_range_info)),
    IPT_ALIGN(sizeof(struct ipt_tos_range_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&tosrange);
}
