#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <iptables.h>
#include <linux/if_vlan.h>
#include <linux/netfilter_ipv4/ipt_vlan.h>


#define VLAN_GROUP_ARRAY_MAX 4096
#define VLAN_PRIO_MAX       8
static void
help(void)
{
	printf(
    "  vlan match v%s options:\n"
    "  --vlan [!] value\n"
    "               match vlan \n"
    "  --vlanprio [!] prioi \n"
    "               match 802.1P priority\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "vlan", 1, 0, '1' },
    { "vlanprio", 1, 0, '2'},
	{0}
};

static u_int16_t
parse_vlan_arg(const char *vlanstr, u_int8_t flag)
{
	unsigned int vlanvalue;
	char* ep;

	vlanvalue =  strtoul(vlanstr,&ep,0) ;

	if ( vlanstr == ep ) {
		exit_error(PARAMETER_PROBLEM,
			   "vlan no valid digits in spi `%s'", vlanstr);
	}
	if ( *vlanstr != '\0'  && *ep != '\0' ) {
		exit_error(PARAMETER_PROBLEM,
			   "vlan error parsing value `%s'", vlanstr);
	}

    if (VLAN_MATCH_VLAN_ID_RANGE == flag)
    {
        if (vlanvalue >= VLAN_GROUP_ARRAY_MAX)
        {
            exit_error(PARAMETER_PROBLEM,
                   "vlan `%s' specified too big (0 -- 4095): would overflow", vlanstr);
        }
    }
    else if (VLAN_MATCH_VLAN_PRIO_RANGE == flag)
    {
        if (vlanvalue >= VLAN_PRIO_MAX)
        {
            exit_error(PARAMETER_PROBLEM,
                   "vlan `%s' specified too big (0 -- 7): would overflow", vlanstr);
        }
    }
    else
    {
        exit_error(PARAMETER_PROBLEM,
               "vlan `%s'  unknown", vlanstr);
    }
    
	return (u_int32_t) vlanvalue;
}

static void
parse_vlan_args(char *vlanstring, u_int16_t *vlan, u_int8_t flag)
{
	char *cp;

	if ((cp = strchr(vlanstring, ':')) == NULL)
		vlan[0] = vlan[1] = parse_vlan_arg(vlanstring, flag);
	else {
		*cp = '\0';
		cp++;

		vlan[0] = vlanstring[0] ? parse_vlan_arg(vlanstring, flag) : 0;
		vlan[1] = cp[0] ? parse_vlan_arg(cp, flag) : vlan[0];
	}

    if (vlan[0] > vlan[1])
    {
		exit_error(PARAMETER_PROBLEM,
			   "specified '%s' too big: would overflow", vlanstring);
    }
}

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	struct ipt_vlan_info *vlaninfo = (struct ipt_vlan_info *)m->data;

	vlaninfo->ulVlanId[0] = 0xFFFF;
	vlaninfo->ulVlanId[1] = 0xFFFF;
    vlaninfo->ulVlanPrio[0] = 0xFFFF;
    vlaninfo->ulVlanPrio[1] = 0xFFFF;
}


/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_vlan_info *vlaninfo = (struct ipt_vlan_info *)(*match)->data;

	switch (c) {
	case '1':
		if (*flags & VLAN_MATCH_VLAN_ID_RANGE)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--vlanid' allowed ");
		check_inverse(optarg, &invert, &optind, 0);
		parse_vlan_args(argv[optind-1], vlaninfo->ulVlanId, VLAN_MATCH_VLAN_ID_RANGE);
		if (invert)
			vlaninfo->ulinvflags |= IPT_VLAN_ID_INV;
		*flags |= VLAN_MATCH_VLAN_ID_RANGE;
		break;
        
    case '2':
        if (*flags & VLAN_MATCH_VLAN_PRIO_RANGE)
            exit_error(PARAMETER_PROBLEM,
                   "Only one `--vlanprio' allowed");
        check_inverse(optarg, &invert, &optind, 0);
        parse_vlan_args(argv[optind-1], vlaninfo->ulVlanPrio, VLAN_MATCH_VLAN_PRIO_RANGE);
        if (invert)
            vlaninfo->ulinvflags |= IPT_VLAN_PRIO_INV;
        *flags |= VLAN_MATCH_VLAN_PRIO_RANGE;
        break;
        
	default:
		return 0;
	}

	return 1;
}

/* Final check; we don't care. */
static void final_check(unsigned int flags)
{
}

static void print_vlanargs(struct ipt_vlan_info *vlaninfo)
{

    if (vlaninfo->ulVlanId[0] != 0xFFFF)
    {
        printf("%s ", "vlanid");
        if (vlaninfo->ulinvflags & IPT_VLAN_ID_INV)
            printf(" ! ");
        
        printf("%d:%d",  vlaninfo->ulVlanId[0], vlaninfo->ulVlanId[1]);
    }

    if (vlaninfo->ulVlanPrio[0] != 0xFFFF)
    {
        printf("%s ", "vlanprio");
        if (vlaninfo->ulinvflags & IPT_VLAN_PRIO_INV)
            printf(" ! ");
        
        printf(" %d:%d ",  vlaninfo->ulVlanPrio[0], vlaninfo->ulVlanPrio[1]);
    }
    printf("  ");
    
}

/* Prints out the union ipt_matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match, int numeric)
{
	struct ipt_vlan_info *vlaninfo = (struct ipt_vlan_info *)match->data;

	printf("vlan ");
	print_vlanargs(vlaninfo);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
    const struct ipt_vlan_info *vlaninfo = (struct ipt_vlan_info *)match->data;

    if (vlaninfo->ulVlanId[0] != 0xFFFF) {
    	printf(" --vlanid %s", 
    		(vlaninfo->ulinvflags & IPT_VLAN_ID_INV) ? "! " : " ");
    	if (vlaninfo->ulVlanId[0]
    	    != vlaninfo->ulVlanId[1])
    		printf("%u:%u ",
    		       vlaninfo->ulVlanId[0],
    		       vlaninfo->ulVlanId[1]);
    	else
    		printf("%u ",
    		       vlaninfo->ulVlanId[0]);
    }

    if (vlaninfo->ulVlanPrio[0] != 0xFFFF) {
    printf(" --vlanid %s", 
        (vlaninfo->ulinvflags & IPT_VLAN_PRIO_INV) ? "! " : " ");
    if (vlaninfo->ulVlanPrio[0]
        != vlaninfo->ulVlanPrio[1])
        printf("%u:%u ",
               vlaninfo->ulVlanPrio[0],
               vlaninfo->ulVlanPrio[1]);
    else
        printf("%u ",
               vlaninfo->ulVlanPrio[0]);
    }

}

static
struct iptables_match vlan
= { NULL,
    "vlan",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_vlan_info)),
    IPT_ALIGN(sizeof(struct ipt_vlan_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void
_init(void)
{
	register_match(&vlan);
}

