/* Shared library add-on to iptables for DSCP
 *
 * (C) 2000- 2002 by Matthew G. Marsh <mgm@paktronix.com>,
 * 		     Harald Welte <laforge@gnumonks.org>
 *
 * This program is distributed under the terms of GNU GPL v2, 1991
 *
 * libipt_dscp_j.c borrowed heavily from libipt_tos_j.c
 *
 * --set-class added by Iain Barnes
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_vlan_j.h>


#define VLAN_VID_MASK 0xFFF

static void init(struct ipt_entry_target *t, unsigned int *nfcache) 
{
    struct ipt_vlan_j_info *vlaninfo = (struct ipt_vlan_j_info *)t->data;
    vlaninfo->ulVlanTag = 0;
    vlaninfo->vlan_flags = 0;
}

static void help(void) 
{
	printf(
    "VLAN target options\n"
    "  --set-vlan value		Set vlan Tag\n"
    "  		                This value can be in decimal (ex: 32)\n"
    "               		or in hex (ex: 0x20)\n"
    "                       the vlan tag format is as follow:\n"
    "                       priority: 3 bit (0 -- 7)\n"
    "                       CFI:      1 bit\n"
    "                       VlanId:   12 bit (0 -- 4095)\n"
    "  --set-vlanid value   set vlan id only\n"
    "  --set-vlanprio value set the vlan priority only\n"
    "\n"
    );
}

static struct option opts[] = {
	{ "set-vlan", 1, 0, 'F' },
	{ "set-vlanid", 1, 0, 'I'},
	{ "set-vlanprio", 1, 0, 'P'},
	{ 0 }
};

static void
parse_vlan(const unsigned char *s, struct ipt_vlan_j_info *vlaninfo)
{
	unsigned int vlan;
       
	if (string_to_number(s, 0, 0xFFFF, &vlan) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid vlantag  `%s'\n", s);


    if (vlan > (0xE000 | VLAN_VID_MASK))
    {
		exit_error(PARAMETER_PROBLEM,
			   "vlan tag  `%d` out of range\n", vlan);
    }

    if (vlan & (~(0xE000 | VLAN_VID_MASK)))
    {
		exit_error(PARAMETER_PROBLEM,
			   "vlan tag  `0x%0x` error, no support  set CFI bit now\n", vlan);
    }

    vlaninfo->ulVlanTag = vlan;
    vlaninfo->vlan_flags |= ATP_VLAN_SET_ALL;


#if 0
    if (vlan & VLAN_VID_MASK)
    {
        vlaninfo->ulVlanTag = vlan;
        vlaninfo->vlan_flags |= ATP_VLAN_SET_ID;
    }

    if (vlan & 0xE000)
    {
        vlaninfo->ulVlanTag = vlan;
        vlaninfo->vlan_flags |= ATP_VLAN_SET_PRIO;
    }
#endif

	return;
}


static void parse_vlanprio(const unsigned char *s, struct ipt_vlan_j_info *vlaninfo)
{
    unsigned int vlanprio = 0;

    
	if (string_to_number(s, 0, 7, &vlanprio) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid vlanprio `%s'\n", s);

    vlaninfo->ucVlanprio = vlanprio;
    vlaninfo->vlan_flags |= ATP_VLAN_SET_PRIO;

#if 0
    vlaninfo->ulVlanTag = (vlanprio << 13) & 0xE000;
    vlaninfo->vlan_flags = ATP_VLAN_SET_PRIO;
#endif


    return ;
    
    
}

static void parse_vlanid(const unsigned char *s, struct ipt_vlan_j_info *vlaninfo)
{
    unsigned int vlanid = 0;

    
	if (string_to_number(s, 0, VLAN_VID_MASK, &vlanid) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid vlanprio `%s'\n", s);

#if 1
    vlaninfo->usVlanid = vlanid & VLAN_VID_MASK;
    vlaninfo->vlan_flags |= ATP_VLAN_SET_ID;
#else
    vlaninfo->ulVlanTag = vlanid & VLAN_VID_MASK;
    vlaninfo->vlan_flags = ATP_VLAN_SET_ID;
#endif

    return ;
}


static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_vlan_j_info *vlaninfo
		= (struct ipt_vlan_j_info *)(*target)->data;

	switch (c) {

        /* VLANǰIDȼ */
	case 'F':
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
			           "vlan target: Only use --set-vlan ONCE!");
		parse_vlan(optarg, vlaninfo);
		*flags = 1;
		break;
        
    case 'P':
        if (*flags)
        {
			exit_error(PARAMETER_PROBLEM,
			           "vlan target: Only use --set-vlan ONCE!");
        }
        parse_vlanprio(optarg, vlaninfo);
        *flags = 1;
        break;

    case 'I':
        if (*flags)
        {
            exit_error(PARAMETER_PROBLEM,
                       "vlan target: Only use --set-vlan ONCE!");
        }
        parse_vlanid(optarg, vlaninfo);
        *flags = 1;
        break;
        

	default:
		return 0;
	}

	return 1;
}

static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
		           "VLAN target: Parameter --set-vlan is required");
}

static void print_vlan(struct ipt_vlan_j_info vlantag, int numeric)
{
    if (ATP_VLAN_SET_ID == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf(" vlanid 0x%0x ", vlantag.usVlanid);
    }
    else if (ATP_VLAN_SET_PRIO == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf(" vlanprio 0x%0x ", vlantag.ucVlanprio);
    }
    else if (ATP_VLAN_SET_ALL == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf("  vlantag 0x%0x ", vlantag.ulVlanTag);
    }
    else
    {
        printf("unknown....");
    }

    return ;
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_vlan_j_info *vlaninfo =
		(const struct ipt_vlan_j_info *)target->data;
	print_vlan(*vlaninfo, numeric);

    return ;
}



static void print_vlanset(struct ipt_vlan_j_info vlantag)
{
    if (ATP_VLAN_SET_ID == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf(" --set-vlanid 0x%0x ", vlantag.usVlanid);
    }
    else if (ATP_VLAN_SET_PRIO == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf(" --set-vlanprio 0x%0x ", vlantag.ucVlanprio);
    }
    else if (ATP_VLAN_SET_ALL == (vlantag.vlan_flags & ATP_VLAN_SET_ALL))
    {
        printf("  --set-vlantag 0x%0x ", vlantag.ulVlanTag);
    }
    else
    {
        printf("unknown....");
    }
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_vlan_j_info *vlaninfo =
		(const struct ipt_vlan_j_info *)target->data;

	print_vlanset(*vlaninfo);

    return ;
}

static
struct iptables_target vlan
= { NULL,
    "VLAN",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_vlan_j_info)),
    IPT_ALIGN(sizeof(struct ipt_vlan_j_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&vlan);
}

