/* arch/arm/mach-msm/clock.c
 *
 * MSM architecture clock driver
 *
 * Copyright (C) 2007 Google, Inc.
 * Copyright (c) 2007-2008 QUALCOMM USA, INC.
 * Author: San Mehat <san@android.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/mutex.h>
#include <linux/debugfs.h>
#include <linux/spinlock.h>
#include <linux/cpufreq.h>
#include <linux/sort.h>
#include <asm/io.h>
#include <asm/arch/msm_iomap.h>
#include <asm/mach-types.h>
#include "clock.h"

#include "proc_comm.h"

#define PERF_SWITCH_DEBUG 0
#define PERF_SWITCH_STEP_DEBUG 0

#if PERF_SWITCH_DEBUG
#define dprintk(x...) printk(x)
#else
#define dprintk(x...) do {} while (0)
#endif

#if PERF_SWITCH_STEP_DEBUG
#define sdprintk(x...) printk(x)
#else
#define sdprintk(x...) do {} while (0)
#endif

#define cdprintk(msg...) \
	cpufreq_debug_printk(CPUFREQ_DEBUG_DRIVER, "cpufreq-msm", msg)

#if defined(CONFIG_ARCH_QSD)
#define SHOT_SWITCH  4
#define HOP_SWITCH   5
#define SIMPLE_SLEW  6
#define COMPLEX_SLEW 7

static bool sc_pll;
#endif

struct clock_state
{
	struct clkctl_acpu_speed	*current_speed;
	struct mutex			lock;
	uint32_t			acpu_switch_time_us;
	uint32_t			max_speed_delta_khz;
	uint32_t			vdd_switch_time_us;
	struct clk	 		*active_clocks;
	int 				active_clock_count;
};

static struct clock_state drv_state = { 0 };
static DEFINE_MUTEX(clocks_mutex);
static LIST_HEAD(clocks);

/*
 * Bitmap of enabled clocks, excluding ACPU which is always
 * enabled
 */
static DECLARE_BITMAP(clock_map_enabled, NR_CLOCKS);
static DEFINE_SPINLOCK(clock_map_lock);

static void acpuclk_init(struct clk *clk);
static int acpuclk_enable(struct clk *clk);
static void acpuclk_disable(struct clk *clk);
static unsigned long acpuclk_get_rate(struct clk *clk);
int acpuclk_set_rate(struct clk *clk, unsigned long rate, int for_power_collapse);

/*
 * ACPU speed table
 */
#if !defined(CONFIG_ARCH_QSD)
/* The table is initialized with the most common clock plan. Table fix up is
 * done during boot up to match the actual clock plan supported by the
 * hardware. */
struct clkctl_acpu_speed  acpu_freq_tbl[] = {
	{ 0, 19200, ACPU_PLL_TCXO, 0, 0, 19200, 0, 30720, 0},
	{ 0, 61440, ACPU_PLL_0, 4, 3, 61440, 0, 30720, 0}, /* VDD assumed */
	{ 0, 81920, ACPU_PLL_0, 4, 2, 40960, 1, 61440, 0}, /* VDD assumed */
	{ 0, 96000, ACPU_PLL_1, 1, 7, 48000, 1, 61440, 0}, /* VDD assumed */
	{ 1, 122880, ACPU_PLL_0, 4, 1, 61440, 1, 61440, 3},
	{ 0, 128000, ACPU_PLL_1, 1, 5, 64000, 1, 61440, 3}, /* VDD assumed */
	{ 0, 176000, ACPU_PLL_2, 2, 5, 88000, 1, 61440, 3}, /* VDD assumed */
	{ 0, 192000, ACPU_PLL_1, 1, 3, 64000, 2, 61440, 3}, /* VDD assumed */
	{ 1, 245760, ACPU_PLL_0, 4, 0, 81920, 2, 61440, 4},
	{ 1, 256000, ACPU_PLL_1, 1, 2, 128000, 1, 128000, 5},
	{ 0, 264000, ACPU_PLL_2, 2, 3, 88000, 2, 128000, 5}, /* VDD assumed */
	{ 0, 352000, ACPU_PLL_2, 2, 2, 88000, 3, 128000, 5}, /* VDD assumed */
	{ 1, 384000, ACPU_PLL_1, 1, 1, 128000, 2, 128000, 6},
	{ 1, 528000, ACPU_PLL_2, 2, 1, 132000, 3, 128000, 7},
	{ 0, 0, 0, 0, 0, 0, 0, 0},
};
#else
struct clkctl_acpu_speed acpu_freq_tbl[] = {
	{ 0, 19200, ACPU_PLL_TCXO, 0, 0, 0, 0, 0, 0, 0},
	{ 0, 48000, ACPU_PLL_1, 1, 0xF, 0, 0, 0, 0, 0},
	{ 0, 64000, ACPU_PLL_1, 1, 0xB, 0, 0, 0, 0, 0},
	{ 0, 96000, ACPU_PLL_1, 1, 7, 0, 0, 0, 0, 0},
	{ 0, 128000, ACPU_PLL_1, 1, 5, 0, 0, 2, 0, 0},
	{ 0, 192000, ACPU_PLL_1, 1, 3, 0, 0, 0, 0, 0},
	{ 1, 245000, ACPU_PLL_0, 4, 0, 0, 0, 0, 0, 0}, /* 235.93 on CDMA */
	{ 0, 256000, ACPU_PLL_1, 1, 2, 0, 0, 0, 0, 0},
	{ 1, 384000, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xA, 0},
	{ 0, 422400, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xB, 0},
	{ 0, 460800, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xC, 0},
	{ 0, 499200, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xD, 0},
	{ 0, 537600, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xE, 0},
	{ 1, 576000, ACPU_PLL_3, 3, 0, 0, 0, 1, 0xF, 0},
	{ 0, 614400, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x10, 0},
	{ 0, 652800, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x11, 0},
	{ 0, 691200, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x12, 0},
	{ 0, 729600, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x13, 0},
	{ 1, 768000, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x14, 0},
	{ 0, 806400, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x15, 0},
	{ 0, 844800, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x16, 0},
	{ 0, 883200, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x17, 0},
	{ 0, 921600, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x18, 0},
	{ 0, 960000, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x19, 0},
	{ 1, 998400, ACPU_PLL_3, 3, 0, 0, 0, 1, 0x1A, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
};
#endif

static int pc_clk_enable(unsigned id)
{
	int rc = msm_proc_comm(PCOM_CLKCTL_RPC_ENABLE, &id, 0);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

static void pc_clk_disable(unsigned id)
{
	msm_proc_comm(PCOM_CLKCTL_RPC_DISABLE, &id, 0);
}

static int pc_clk_set_rate(unsigned id, unsigned rate)
{
	/* The rate _might_ be rounded off to the nearest KHz value by the
	 * remote function. So a return value of 0 doesn't necessarily mean
	 * that the exact rate was set successfully.
	 */
	int rc = msm_proc_comm(PCOM_CLKCTL_RPC_SET_RATE, &id, &rate);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

static int pc_clk_set_min_rate(unsigned id, unsigned rate)
{
	int rc = msm_proc_comm(PCOM_CLKCTL_RPC_MIN_RATE, &id, &rate);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

static int pc_clk_set_max_rate(unsigned id, unsigned rate)
{
	int rc = msm_proc_comm(PCOM_CLKCTL_RPC_MAX_RATE, &id, &rate);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

static int pc_clk_set_flags(unsigned id, unsigned flags)
{
	int rc = msm_proc_comm(PCOM_CLKCTL_RPC_SET_FLAGS, &id, &flags);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

static unsigned pc_clk_get_rate(unsigned id)
{
	if (msm_proc_comm(PCOM_CLKCTL_RPC_RATE, &id, 0))
		return 0;
	else
		return id;
}

static unsigned pc_clk_is_enabled(unsigned id)
{
	if (msm_proc_comm(PCOM_CLKCTL_RPC_ENABLED, &id, 0))
		return 0;
	else
		return id;
}

static int pc_pll_request(unsigned id, unsigned on)
{
	int rc;
	on = !!on;
	rc = msm_proc_comm(PCOM_CLKCTL_RPC_PLL_REQUEST, &id, &on);
	if (rc < 0)
		return rc;
	else
		return ((int)id < 0 ? -EINVAL : 0);
}

/*----------------------------------------------------------------------------
 * Architecture level clock registration
 *---------------------------------------------------------------------------*/

int clk_register(struct clk *clk)
{
	if (!clk || IS_ERR(clk))
		return -EINVAL;
	mutex_lock(&clocks_mutex);
	spin_lock_init(&clk->lock);
	list_add_tail(&clk->list, &clocks);
	if (clk->id == ACPU_CLK)
		acpuclk_init(clk);
	mutex_unlock(&clocks_mutex);
	return 0;
}

/*----------------------------------------------------------------------------
 * Standard clock functions defined in include/linux/clk.h
 *---------------------------------------------------------------------------*/
struct clk *clk_get(struct device *dev, const char *id)
{
	struct clk *c_clk;

	mutex_lock(&clocks_mutex);
	list_for_each_entry(c_clk, &clocks, list) {
		if (!strcmp(id, c_clk->name) && try_module_get(c_clk->owner)) {
			mutex_unlock(&clocks_mutex);
			return c_clk;
		}
	}
	mutex_unlock(&clocks_mutex);
	return ERR_PTR(-ENOENT);
}
EXPORT_SYMBOL(clk_get);

int clk_enable(struct clk *clk)
{
	unsigned long flags;
	if (clk->id == ACPU_CLK)
		return acpuclk_enable(clk);
	spin_lock_irqsave(&clk->lock, flags);
	clk->count++;
	if (clk->count == 1) {
		pc_clk_enable(clk->id);
		spin_lock(&clock_map_lock);
		clock_map_enabled[BIT_WORD(clk->id)] |= BIT_MASK(clk->id);
		spin_unlock(&clock_map_lock);
	}
	spin_unlock_irqrestore(&clk->lock, flags);
	return 0;
}
EXPORT_SYMBOL(clk_enable);

void clk_disable(struct clk *clk)
{
	unsigned long flags;
	if (clk->id == ACPU_CLK)
		return acpuclk_disable(clk);
	spin_lock_irqsave(&clk->lock, flags);
	BUG_ON(clk->count == 0);
	clk->count--;
	if (clk->count == 0) {
		pc_clk_disable(clk->id);
		spin_lock(&clock_map_lock);
		clock_map_enabled[BIT_WORD(clk->id)] &= ~BIT_MASK(clk->id);
		spin_unlock(&clock_map_lock);
	}
	spin_unlock_irqrestore(&clk->lock, flags);
}
EXPORT_SYMBOL(clk_disable);

unsigned long clk_get_rate(struct clk *clk)
{
	if (clk->id == ACPU_CLK)
		return acpuclk_get_rate(clk);
	return pc_clk_get_rate(clk->id);
}
EXPORT_SYMBOL(clk_get_rate);

void clk_put(struct clk *clk)
{
	module_put(clk->owner);
}
EXPORT_SYMBOL(clk_put);

int clk_set_rate(struct clk *clk, unsigned long rate)
{
	int ret = -EPERM;
	if (clk->id == ACPU_CLK)
		return acpuclk_set_rate(clk, rate, 0);
	if (clk->flags & CLKFLAG_USE_MAX_TO_SET) {
		ret = pc_clk_set_max_rate(clk->id, rate);
		if (ret) {
			printk(KERN_ERR "pc_clk_set_max_rate failed (%s)\n",
							clk->name);
			return ret;
		}
	}
	if (clk->flags & CLKFLAG_USE_MIN_TO_SET) {
		ret = pc_clk_set_min_rate(clk->id, rate);
		if (ret) {
			printk(KERN_ERR "pc_clk_set_min_rate failed (%s)\n",
							clk->name);
			return ret;
		}
	}
	if (!(clk->flags & CLKFLAG_USE_NO_SET)) {
		ret = pc_clk_set_rate(clk->id, rate);
		if (ret)
			printk(KERN_ERR "pc_clk_set_rate failed (%s)\n",
							clk->name);
	}
	return ret;
}
EXPORT_SYMBOL(clk_set_rate);

int clk_set_parent(struct clk *clk, struct clk *parent)
{
	/*
	 * XXX: Waiting for proper clock tree info from QC
	 */
	return -ENOSYS;
}

struct clk *clk_get_parent(struct clk *clk)
{
	/*
	 * XXX: Waiting for peroper clock tree info from QC
	 */
	return ERR_PTR(-ENOSYS);
}

int clk_set_flags(struct clk *clk, unsigned long flags)
{
	if (clk == NULL || IS_ERR(clk))
		return -EINVAL;
	if (clk->id == ACPU_CLK)
		return -EINVAL;
	return pc_clk_set_flags(clk->id, flags);
}
EXPORT_SYMBOL(clk_set_flags);

/*----------------------------------------------------------------------------
 * ARM11 'owned' clock control
 *---------------------------------------------------------------------------*/

#if defined(CONFIG_ARCH_QSD)
static void pll_set_freq(uint32_t lval, unsigned freq_switch)
{
	uint32_t regval;

	if (lval > 33)
		lval = 33;
	if (lval < 10)
		lval = 10;

	/* wait for any calibrations or frequency switches to finish */
	while (readl(PLL_STATUS_ADDR) & 0x3);

	/* write the new L val and switch mode */
	regval = readl(PLL_FSM_CTL_EXT_ADDR);
	regval &= ~(0x3f << 3);
	regval |= (lval << 3);
	if (freq_switch == SIMPLE_SLEW)
		regval |= (0x1 << 9);

	regval &= ~(0x3 << 0);
	regval |= (freq_switch << 0);
	writel(regval, PLL_FSM_CTL_EXT_ADDR);

	/* set bypassnl, reset_n, scpll to normal mode */
	writel(0x00600007, PLL_CTL_ADDR);
}

static void pll_apps_enable(bool state)
{
	uint32_t regval;

	/* Wait for any frequency switches to finish. */
	while (readl(PLL_STATUS_ADDR) & 0x1);

	/* put the pll in standby mode */
	regval = readl(PLL_CTL_ADDR);
	regval &= ~(0x7);
	regval |= (0x2);
	writel(regval, PLL_CTL_ADDR);

	udelay(drv_state.vdd_switch_time_us);
	if (state) {
		/* put the pll in normal mode */
		regval = readl(PLL_CTL_ADDR);
		regval |= (0x7);
		writel(regval, PLL_CTL_ADDR);
	} else {
		/* put the pll in power down mode */
		regval = readl(PLL_CTL_ADDR);
		regval &= ~(0x7);
		writel(regval, PLL_CTL_ADDR);
	}
	udelay(drv_state.vdd_switch_time_us);
	sc_pll = state;
}

static void pll_init(uint32_t lval)
{
	/* power down scpll */
	writel(0x0, PLL_CTL_ADDR);

	/* set bypassnl, put into standby */
	writel(0x00400002, PLL_CTL_ADDR);

	/* set bypassnl, reset_n, full calibration */
	writel(0x00600004, PLL_CTL_ADDR);

	/* Ensure register write to initiate calibration has taken
	effect before reading status flag */
	dmb();

	/* wait for cal_all_done */
	while (readl(PLL_STATUS_ADDR) & 0x2);

	/* power down scpll */
	writel(0x0, PLL_CTL_ADDR);

	/* switch scpll to desired freq */
	pll_set_freq(lval, SHOT_SWITCH);
}
#endif

static int acpuclk_enable(struct clk *clk)
{
	return 0; /* acpuclk is always enabled */
}

static void acpuclk_disable(struct clk *clk)
{
	/* acpuclk is always enabled */
}

static int acpuclk_set_vdd_level(int vdd)
{
	uint32_t current_vdd;

	current_vdd = readl(A11S_VDD_SVS_PLEVEL_ADDR) & 0x07;

	dprintk(KERN_INFO "clock: Switching VDD from %u -> %d\n",
		current_vdd, vdd);
	writel((1 << 7) | (vdd << 3), A11S_VDD_SVS_PLEVEL_ADDR);
	udelay(drv_state.vdd_switch_time_us);
	if ((readl(A11S_VDD_SVS_PLEVEL_ADDR) & 0x7) != vdd) {
		dprintk(KERN_ERR "clock: VDD set failed\n");
		return -EIO;
	}

	dprintk(KERN_INFO "clock: VDD switched\n");
	return 0;
}

int acpuclk_set_rate(struct clk *clk, unsigned long rate, int for_power_collapse)
{
	uint32_t reg_clkctl, reg_clksel;
	struct clkctl_acpu_speed *cur_s, *tgt_s, *strt_s;
	int ramp_up = 0; /* 1 == up, 0 == down */
#if !defined(CONFIG_ARCH_QSD)
	uint32_t clk_div;
#endif
	int rc = 0;

	strt_s = cur_s = drv_state.current_speed;

	if (rate == (cur_s->a11clk_khz * 1000))
		return 0;

	for (tgt_s = acpu_freq_tbl; tgt_s->a11clk_khz != 0; tgt_s++) {
		if (tgt_s->a11clk_khz == (rate / 1000))
			break;
	}

	if (tgt_s->a11clk_khz == 0)
		return -EINVAL;

	if (tgt_s->a11clk_khz > strt_s->a11clk_khz)
		ramp_up = 1;
	else
		ramp_up = 0;

	if (!for_power_collapse)
		mutex_lock(&drv_state.lock);

	dprintk(KERN_INFO "clock: Switching from ACPU rate %u -> %u\n",
		strt_s->a11clk_khz * 1000, tgt_s->a11clk_khz * 1000);

#if defined(CONFIG_ARCH_QSD)
	/* Loop from the first step (up or down), until the target new clock
	   freq is achieved */
	while (cur_s != tgt_s) {
		if (ramp_up)
			cur_s++;
		else
			cur_s--;

		/* If the current and the target PLLs are both Scorpion PLL,
		   then slew directly to the target frequency */
		if (drv_state.current_speed->pll == ACPU_PLL_3 &&
			 tgt_s->pll == ACPU_PLL_3)
			cur_s = tgt_s;

		/* If the current PLL is the Scorpion PLL but the target PLL is
		   NOT the Scorpion PLL, then skip all intermediate frequencies
		   that use the Scorpion PLL and slew directly to the last
		   frequency associated with the Scorpion PLL */
		if (drv_state.current_speed->pll == ACPU_PLL_3 &&
		    tgt_s->pll != ACPU_PLL_3) {
			if (ramp_up)
				while ((cur_s + 1)->pll == ACPU_PLL_3)
					cur_s++;
			else
				while ((cur_s - 1)->pll == ACPU_PLL_3)
					cur_s--;
		}

		/* Skip over speeds on the wrong PLL */
		while (cur_s->pll != drv_state.current_speed->pll &&
		       cur_s->pll != tgt_s->pll) {
			if (ramp_up)
				cur_s++;
			else
				cur_s--;
		}

		/* If switching from non-Scorpion PLL to the Scorpion
		   PLL, then use the SHOT_SWITCH mode */
		if (drv_state.current_speed->pll != ACPU_PLL_3 &&
		    cur_s->pll == ACPU_PLL_3) {
			/* Make sure the Scorpion PLL is turned */
			pll_apps_enable(1);
			pll_set_freq(cur_s->sc_l_value, SHOT_SWITCH);

			/* Program the new Scorpion core source selection */
			reg_clksel = readl(SPSS_CLK_SEL_ADDR);
			reg_clksel &= ~(0x3 << 1);
			reg_clksel |= (cur_s->sc_core_src_sel_mask << 1);
			writel(reg_clksel, SPSS_CLK_SEL_ADDR);
		}

		/* If switching speeds derived from the Scorpion PLL,
		   then use the automatic slewing method */
		else if (drv_state.current_speed->pll == ACPU_PLL_3 &&
			 cur_s->pll == ACPU_PLL_3) {
			pll_set_freq(cur_s->sc_l_value, COMPLEX_SLEW);

			/* Program the new Scorpion core source selection */
			reg_clksel = readl(SPSS_CLK_SEL_ADDR);
			reg_clksel &= ~(0x3 << 1);
			reg_clksel |= (cur_s->sc_core_src_sel_mask << 1);
			writel(reg_clksel, SPSS_CLK_SEL_ADDR);
		}

		/* Otherwise, step through all intermediate frequencies */
		else {
			/* Get the current clock source selection */
			reg_clksel = readl(SPSS_CLK_SEL_ADDR) & 0x1;

			/* Use the alternative clock source selection */
			switch (reg_clksel) {
			case 0x0:
				/* Program the clock source according to the new
				   frequency */
				reg_clkctl = readl(SPSS_CLK_CNTL_ADDR);
				reg_clkctl &= ~(0x7 << 4);
				reg_clkctl |= (cur_s->a11clk_src_sel << 4);
				writel(reg_clkctl, SPSS_CLK_CNTL_ADDR);

				/* Program the clock divider according to the
				   new frequency */
				reg_clkctl = readl(SPSS_CLK_CNTL_ADDR);
				reg_clkctl &= ~(0xf << 0);
				reg_clkctl |= (cur_s->a11clk_src_div << 0);
				writel(reg_clkctl, SPSS_CLK_CNTL_ADDR);

				/* Program the clock source selection */
				reg_clksel = readl(SPSS_CLK_SEL_ADDR);
				reg_clksel |= 0x1;
				writel(reg_clksel, SPSS_CLK_SEL_ADDR);

				break;

			case 0x1:
				/* Program the clock source according to the new
				   frequency */
				reg_clkctl = readl(SPSS_CLK_CNTL_ADDR);
				reg_clkctl &= ~(0x7 << 12);
				reg_clkctl |= (cur_s->a11clk_src_sel << 12);
				writel(reg_clkctl, SPSS_CLK_CNTL_ADDR);

				/* Program the clock divider according to the
				   new frequency */
				reg_clkctl = readl(SPSS_CLK_CNTL_ADDR);
				reg_clkctl &= ~(0xf << 8);
				reg_clkctl |= (cur_s->a11clk_src_div << 8);
				writel(reg_clkctl, SPSS_CLK_CNTL_ADDR);

				/* Program the clock source selection */
				reg_clksel = readl(SPSS_CLK_SEL_ADDR);
				reg_clksel &= ~(0x1);
				writel(reg_clksel, SPSS_CLK_SEL_ADDR);

				break;

			default:
				break;
			}
			/* Program the new Scorpion core source selection */
			reg_clksel = readl(SPSS_CLK_SEL_ADDR);
			reg_clksel &= ~(0x3 << 1);
			reg_clksel |= (cur_s->sc_core_src_sel_mask << 1);
			writel(reg_clksel, SPSS_CLK_SEL_ADDR);
		}

		/* Update the driver state with the new clock freq */
		drv_state.current_speed = cur_s;

		/* Wait a few usec to stabilize the clock */
		udelay(drv_state.acpu_switch_time_us);
	}

	/* Make sure the Scorpion PLL is shut off when not being used */
	if (tgt_s->pll != ACPU_PLL_3 && sc_pll != 0)
		pll_apps_enable(0);

#else

	if (tgt_s->vdd > strt_s->vdd) {
		rc = acpuclk_set_vdd_level(tgt_s->vdd);
		if (rc < 0) {
			printk(KERN_ERR "clock: Unable to switch ACPU vdd\n");
			goto failed_unlock;
		}
	}

	/* Set wait states for CPU inbetween frequency changes */
	reg_clkctl = readl(A11S_CLK_CNTL_ADDR);
	reg_clkctl |= (100 << 16); /* set WT_ST_CNT */
	writel(reg_clkctl, A11S_CLK_CNTL_ADDR);

	if (!for_power_collapse &&
	    (strt_s->pll != tgt_s->pll) && (tgt_s->pll != ACPU_PLL_TCXO)) {
		dprintk(KERN_DEBUG "%s: Enabling PLL %d\n",
			__func__, tgt_s->pll);
		rc = pc_pll_request(tgt_s->pll, 1);
		if (rc < 0) {
			printk(KERN_ERR "PLL enable failed (%d)\n", rc);
			goto failed_unlock;
		}
		dprintk(KERN_DEBUG "%s: PLL %d enabled\n",
			__func__, tgt_s->pll);
		cdprintk("PLL %d enabled\n", tgt_s->pll);
	}

	while (cur_s != tgt_s) {
		struct clkctl_acpu_speed *hunt_s = cur_s;
		while (1) {
			int d;

			if (!ramp_up && hunt_s->a11clk_khz == 19200) {
				hunt_s = NULL;
				break;
			}
			if (ramp_up && hunt_s->a11clk_khz == 528000) {
				hunt_s = NULL;
				break;
			}

			if (ramp_up)
				hunt_s++; /* next clk in table */
			else
				hunt_s--; /* prev clk in table */

			/* Only allow clocks which match the target PLL */
			if ((hunt_s->pll != cur_s->pll)
			 && (hunt_s->pll != tgt_s->pll)) {
				continue; /* Skip speed; not at tgt PLL */
			}

			d = abs((int)(cur_s->a11clk_khz - hunt_s->a11clk_khz));
			if (d > drv_state.max_speed_delta_khz) {
				hunt_s = NULL;
				break; /* Unable to find a step */
			}

			/* Valid speed found */
			break;
		}

		if (!hunt_s) {
			printk(KERN_ERR "clock: Invalid ACPU step\n");
			rc = -EINVAL;
			goto failed_unlock;
		}
		sdprintk(KERN_DEBUG "%s: STEP khz = %u, pll = %d\n", __func__,
			 hunt_s->a11clk_khz, hunt_s->pll);

		/* AHB_CLK_DIV */
		clk_div = (readl(A11S_CLK_SEL_ADDR) >> 1) & 0x03;

		/*
		 * If the new clock divider is higher than the previous, then
		 * program the divider before switching the clock
		 */
		if (hunt_s->ahbclk_div > clk_div) {
			reg_clksel = readl(A11S_CLK_SEL_ADDR);
			reg_clksel &= ~(0x3 << 1);
			reg_clksel |= (hunt_s->ahbclk_div << 1);
			writel(reg_clksel, A11S_CLK_SEL_ADDR);
		}
		if ((readl(A11S_CLK_SEL_ADDR) & 0x01) == 0) {
			/* SRC0 */

			/* Program clock source */
			reg_clkctl = readl(A11S_CLK_CNTL_ADDR);
			reg_clkctl &= ~(0x07 << 4);
			reg_clkctl |= (hunt_s->a11clk_src_sel << 4);
			writel(reg_clkctl, A11S_CLK_CNTL_ADDR);

			/* Program clock divider */
			reg_clkctl = readl(A11S_CLK_CNTL_ADDR);
			reg_clkctl &= ~0xf;
			reg_clkctl |= hunt_s->a11clk_src_div;
			writel(reg_clkctl, A11S_CLK_CNTL_ADDR);

			/* Program clock source selection */
			reg_clksel = readl(A11S_CLK_SEL_ADDR);
			reg_clksel |= 1; /* CLK_SEL_SRC1NO  == SRC1 */
			writel(reg_clksel, A11S_CLK_SEL_ADDR);
		} else {
			/* SRC1 */

			/* Program clock source */
			reg_clkctl = readl(A11S_CLK_CNTL_ADDR);
			reg_clkctl &= ~(0x07 << 12);
			reg_clkctl |= (hunt_s->a11clk_src_sel << 12);
			writel(reg_clkctl, A11S_CLK_CNTL_ADDR);

			/* Program clock divider */
			reg_clkctl = readl(A11S_CLK_CNTL_ADDR);
			reg_clkctl &= ~(0xf << 8);
			reg_clkctl |= (hunt_s->a11clk_src_div << 8);
			writel(reg_clkctl, A11S_CLK_CNTL_ADDR);

			/* Program clock source selection */
			reg_clksel = readl(A11S_CLK_SEL_ADDR);
			reg_clksel &= ~1; /* CLK_SEL_SRC1NO  == SRC0 */
			writel(reg_clksel, A11S_CLK_SEL_ADDR);
		}

		/*
		 * If the new clock divider is lower than the previous, then
		 * program the divider after switching the clock
		 */
		if (hunt_s->ahbclk_div < clk_div) {
			reg_clksel = readl(A11S_CLK_SEL_ADDR);
			reg_clksel &= ~(0x3 << 1);
			reg_clksel |= (hunt_s->ahbclk_div << 1);
			writel(reg_clksel, A11S_CLK_SEL_ADDR);
		}
		cur_s = hunt_s;
		drv_state.current_speed = cur_s;
		udelay(drv_state.acpu_switch_time_us);
	}

	if (!for_power_collapse &&
	    (strt_s->pll != tgt_s->pll) && (strt_s->pll != ACPU_PLL_TCXO)) {
		dprintk(KERN_DEBUG "%s: Disabling PLL %d\n",
			__func__, strt_s->pll);
		rc = pc_pll_request(strt_s->pll, 0);
		if (rc < 0) {
			printk(KERN_ERR "PLL disable failed (%d)\n", rc);
			goto failed_unlock;
		}
		dprintk(KERN_DEBUG "%s: PLL %d disabled\n",
			__func__, strt_s->pll);
		cdprintk("PLL %d disabled\n", strt_s->pll);
	}

	/* Drop VDD level if we can. */
	if (tgt_s->vdd < strt_s->vdd) {
		if (acpuclk_set_vdd_level(tgt_s->vdd) < 0)
			printk(KERN_ERR "clock: Unable to drop ACPU vdd\n");
	}

	/* Set the bus speed. */
	if (strt_s->ebi1clk_khz != tgt_s->ebi1clk_khz) {
		rc = pc_clk_set_min_rate(EBI1_CLK, tgt_s->ebi1clk_khz * 1000);
		/* Failing to set the bus speed is not a failure in scaling
		 * the CPU frequency.  It is just a suboptimal power
		 * consumption scenario.
		 */
		if (rc < 0)
			printk(KERN_ERR "EBI1 min rate failed (%d)\n", rc);
		else
			cdprintk("AXI bus request: %d KHz\n",
				 tgt_s->ebi1clk_khz);
	}

#endif /* defined(CONFIG_ARCH_QSD) */

	dprintk(KERN_DEBUG "%s: ACPU speed change complete\n", __func__);

failed_unlock:
	if (!for_power_collapse)
		mutex_unlock(&drv_state.lock);
	return rc;
}

#if !defined(CONFIG_ARCH_QSD)
static void acpuclk_init(struct clk *clk)
{
	struct clkctl_acpu_speed *speed;
	uint32_t div, sel, current_vdd;

	/*
	 * Determine the rate of ACPU clock
	 */

	if (!(readl(A11S_CLK_SEL_ADDR) & 0x01)) { /* CLK_SEL_SRC1N0 */
		/* CLK_SRC0_SEL */
		sel = (readl(A11S_CLK_CNTL_ADDR) >> 12) & 0x7;
		/* CLK_SRC0_DIV */
		div = (readl(A11S_CLK_CNTL_ADDR) >> 8) & 0x0f;
	} else {
		/* CLK_SRC1_SEL */
		sel = (readl(A11S_CLK_CNTL_ADDR) >> 4) & 0x07;
		/* CLK_SRC1_DIV */
		div = readl(A11S_CLK_CNTL_ADDR) & 0x0f;
	}

	for (speed = acpu_freq_tbl; speed->a11clk_khz != 0; speed++) {
		if (speed->a11clk_src_sel == sel
		 && (speed->a11clk_src_div == div))
			break;
	}
	if (speed->a11clk_khz == 0) {
		printk(KERN_WARNING "Warning - ACPU clock reports invalid speed\n");
		return;
	}

	drv_state.current_speed = speed;


	printk(KERN_INFO "ACPU running at %d KHz\n", speed->a11clk_khz);
}
#else
static void acpuclk_init(struct clk *clk)
{
	struct clkctl_acpu_speed *speed;
	uint32_t div, sel, regval;

	/* initialize scpll at 384Mhz */
	pll_init(0xA);

	/* Determine the source of the Scorpion clock. */
	regval = readl(SPSS_CLK_SEL_ADDR);
	switch ((regval & 0x6) >> 1) {
	case 0: /* raw source clock */
	case 3: /* low jitter PLL1 (768Mhz) */
		if (regval & 0x1) {
			sel = ((readl(SPSS_CLK_CNTL_ADDR) >> 4) & 0x7);
			div = ((readl(SPSS_CLK_CNTL_ADDR) >> 0) & 0xf);
		} else {
			sel = ((readl(SPSS_CLK_CNTL_ADDR) >> 12) & 0x7);
			div = ((readl(SPSS_CLK_CNTL_ADDR) >> 8) & 0xf);
		}

		/* Find the matching clock rate. */
		for (speed = acpu_freq_tbl; speed->a11clk_khz != 0; speed++) {
			if (speed->a11clk_src_sel == sel &&
			    speed->a11clk_src_div == div)
				break;
		}
		sc_pll = 0;
		break;

	case 1: /* unbuffered scorpion pll (384Mhz to 998.4Mhz) */
		sel = ((readl(PLL_FSM_CTL_EXT_ADDR) >> 3) & 0x3f);

		/* Find the matching clock rate. */
		for (speed = acpu_freq_tbl; speed->a11clk_khz != 0; speed++) {
			if (speed->sc_l_value == sel &&
			    speed->sc_core_src_sel_mask == 1)
				break;
		}
		sc_pll = 1;
		break;

	case 2: /* AXI bus clock (128Mhz) */
	default:
		sc_pll = 0;
		speed = &acpu_freq_tbl[4];
	}

	if (speed->a11clk_khz == 0) {
		printk(KERN_WARNING "Warning - ACPU clock reports invalid "
			"speed\n");
		return;
	}

	drv_state.current_speed = speed;


	printk(KERN_INFO "ACPU running at %d KHz\n", speed->a11clk_khz);

	/* switch scpll to 998Mhz */
	acpuclk_set_rate(NULL, 998400000, 0);
}
#endif

static unsigned long acpuclk_get_rate(struct clk *clk)
{
	return drv_state.current_speed->a11clk_khz * 1000;
}

/*
 * Find out whether any clock is enabled that needs the TCXO clock.
 *
 * On exit, the buffer 'reason' holds a bitmap of ids of all enabled
 * clocks found that require TCXO.
 *
 * reason: buffer to hold the bitmap; must be compatible with
 *         linux/bitmap.h
 * nbits: number of bits that the buffer can hold; 0 is ok
 *
 * Return value:
 *      0: does not require the TCXO clock
 *      1: requires the TCXO clock
 */
int msm_clock_require_tcxo(unsigned long *reason, int nbits)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&clock_map_lock, flags);
	ret = !bitmap_empty(clock_map_enabled, NR_CLOCKS);
	if (nbits > 0)
		bitmap_copy(reason, clock_map_enabled, min(nbits, NR_CLOCKS));
	spin_unlock_irqrestore(&clock_map_lock, flags);

	return ret;
}

/*
 * Find the clock matching the given id and copy its name to the
 * provided buffer.
 *
 * Return value:
 * -ENODEV: there is no clock matching the given id
 * -EBUSY : unable to acquire the mutex
 *       0: success
 */
int msm_clock_get_name(uint32_t id, char *name, uint32_t size)
{
	struct clk *c_clk;
	int ret = -ENODEV;

	if (!mutex_trylock(&clocks_mutex))
		return -EBUSY;

	list_for_each_entry(c_clk, &clocks, list) {
		if (id == c_clk->id) {
			strlcpy(name, c_clk->name, size);
			ret = 0;
			break;
		}
	}
	mutex_unlock(&clocks_mutex);

	return ret;
}

#ifdef CONFIG_ARCH_MSM
static int cmp_acpu_speed(const void *a, const void *b)
{
	return ((struct clkctl_acpu_speed *)a)->a11clk_khz
		- ((struct clkctl_acpu_speed *)b)->a11clk_khz;
}

#define DIV2REG(n)		((n)-1)
#define REG2DIV(n)		((n)+1)
#define SLOWER_BY(div,factor) 	div = DIV2REG(REG2DIV(div) * factor)

static void pll0_a11_ahb_fixup(unsigned long pll0_l,
					struct clkctl_acpu_speed *t)
{
#define PLL0_196608_KHZ		10
#define PLL0_245760_KHZ		12
#define PLL0_491520_KHZ		25
#define PLL0_983000_KHZ		51

	switch (pll0_l) {
	case PLL0_983000_KHZ:
		SLOWER_BY(t->a11clk_src_div, 4);
		break;
	case PLL0_491520_KHZ:
		SLOWER_BY(t->a11clk_src_div, 2);
		break;
	case PLL0_245760_KHZ:
		/* Initialization correct. */
		break;
	case PLL0_196608_KHZ:
		t->a11clk_khz = 196608 / REG2DIV(t->a11clk_src_div);
		switch (REG2DIV(t->a11clk_src_div)) {
		case 4: /* 49.152 MHz */
			t->ahbclk_khz = t->a11clk_khz;
			break;
		case 3: /* 65.536 MHz */
			t->ahbclk_khz = t->a11clk_khz;
			t->ahbclk_div = DIV2REG(1);
			break;
		case 2: /* 98.304 MHz */
			/* TODO: Increase AHB freq for sRoc
			 * and Roc 2.0 */
			t->ahbclk_khz = t->a11clk_khz >> 1;
			break;
		case 1: /* 196.608 MHz */
			/* TODO: Increase AHB freq for sRoc
			 * and Roc 2.0 */
			t->ahbclk_khz = t->a11clk_khz / 3;
			break;
		}
		break;
	default:
		printk(KERN_CRIT "Unknown PLL0 speed.\n");
		BUG();
		break;
	}
}

static void pll1_a11_ahb_fixup(unsigned long pll1_l,
					struct clkctl_acpu_speed *t)
{
#define PLL1_768000_KHZ		40
#define PLL1_960000_KHZ		50

	switch (pll1_l) {
	case PLL1_960000_KHZ:
		t->a11clk_khz = 960000 / REG2DIV(t->a11clk_src_div);
		/* XXX: Necessary for AHB freq of 128 MHz, but not
		 * necessary for lower AHB freqs. All ACPU freqs used
		 * for scaling have AHB of 128 MHz, so unconditional
		 * increment might be okay.
		 */
		t->ahbclk_div++;
		t->ahbclk_khz = t->a11clk_khz / REG2DIV(t->ahbclk_div);
		break;
	case PLL1_768000_KHZ:
		/* Initialization correct. */
		break;
	default:
		printk(KERN_CRIT "Unknown PLL1 speed.\n");
		BUG();
		break;
	}
}

static void pll2_a11_ahb_fixup(unsigned long pll2_l,
					struct clkctl_acpu_speed *t)
{
#define PLL2_960000_KHZ		50
#define PLL2_1056000_KHZ	55

	switch (pll2_l) {
	case PLL2_1056000_KHZ:
		/* Initialization correct. */
		break;
	case PLL2_960000_KHZ:
		t->a11clk_khz = 960000 / REG2DIV(t->a11clk_src_div);
		if (t->ahbclk_div > 1)
			t->ahbclk_div--;
		t->ahbclk_khz = t->a11clk_khz / REG2DIV(t->ahbclk_div);
		break;
	default:
		printk(KERN_CRIT "Unknown PLL2 speed.\n");
		BUG();
		break;
	}
}

static void msm_clock_acpu_tbl_fix_up(void)
{
	unsigned long pll0_l, pll1_l, pll2_l;
	int turbo_mode = 0;
	struct clkctl_acpu_speed *t;
	unsigned int n = 0;

	/* This call is used just to make sure the modem side clock code is
	 * up and running and has set up the PLL0 freq.
	 */
	while (!pc_clk_get_rate(EBI1_CLK));

	pll0_l = readl(PLLn_L_VAL(0)) & 0x3f;
	pll1_l = readl(PLLn_L_VAL(1)) & 0x3f;
	pll2_l = readl(PLLn_L_VAL(2)) & 0x3f;

	printk(KERN_INFO "L val: PLL0: %d, PLL1: %d, PLL2: %d\n",
				(int)pll0_l, (int)pll1_l, (int)pll2_l);

	/* No fix up needed. */
	if (pll0_l == PLL0_245760_KHZ && pll1_l == PLL1_768000_KHZ
	    && pll2_l == PLL2_1056000_KHZ)
		goto print_info;

	/* Fix the dividers and the allowed clock rates based on the PLL
	 * frequencies.
	 *
	 * The CPU freq is based off the PLL frequencies. So if the actual
	 * PLL freq is different from the PLL freq assumed in the default
	 * table, then the dividers are adjusted to run the CPU at
	 * frequencies closer to the ones in the default table.
	 *
	 * The AHB freq is based off the CPU freq. Attempt is made to
	 * generally keep the AHB freq as close as possible to the AXI bus
	 * freq. But for each VDD level there is an upper limit for the AHB
	 * bus freq. Apart from the per VDD upper limit the AHB also has an
	 * absolute upper limit of 160 MHz. In most cases, these limitations
	 * would explain why the AHB divider is changed when a PLL freq
	 * changes.
	 */
	for (t = &acpu_freq_tbl[0]; t->a11clk_khz != 0; t++) {
		n++;

		switch (t->pll) {
		case ACPU_PLL_0:
			pll0_a11_ahb_fixup(pll0_l, t);
			break;
		case ACPU_PLL_1:
			pll1_a11_ahb_fixup(pll1_l, t);
			break;
		case ACPU_PLL_2:
			pll2_a11_ahb_fixup(pll2_l, t);
			break;
		}

		if (pll0_l == PLL0_196608_KHZ && t->a11clk_khz <= 196608)
			t->ebi1clk_khz = 24576;
	}

	/* The ACPU table is expected to be in ascending order by the rest
	 * of the code. This is the case after most fix ups. One example
	 * exception is when PLL1 runs at 960 MHz. In this case the entries
	 * corresponding to (256 and 264) need to be swapped.
	 */
	sort(acpu_freq_tbl, n, sizeof(struct clkctl_acpu_speed),
					cmp_acpu_speed, NULL);

	/* The AXI bus freq is also based off the PLL frequencies. For the
	 * AXI bus to be able to run at 160 MHz it needs a PLL that is
	 * running at an integer multiple of 160 MHz. Hence the turbo mode
	 * (AXI running at 160 MHz) is only enabled if one of the PLLs run at
	 * 960 MHz.
	 */
	turbo_mode = (pll1_l == PLL1_960000_KHZ || pll2_l == PLL2_960000_KHZ);

	/* Most SURFs do not have RAM that supports a memory bus freq of
	 * 160MHz.
	 */
	if (turbo_mode && !machine_is_msm7201a_surf()
	    && !machine_is_msm7x25_surf() && !machine_is_msm7601a_surf()
	    && !machine_is_qst1105_surf() && !machine_is_msm7501a_surf()
	    && !machine_is_qst1500_surf() && !machine_is_qst1600_surf()) {
		t = acpu_freq_tbl + n - 1;
		t->ebi1clk_khz = 160000;
	}

print_info:
	printk(KERN_INFO "U  CPU-Freq  PLL  DIV  AHB-Freq  ADIV  AXI-Freq\n");
	for (t = &acpu_freq_tbl[0]; t->a11clk_khz != 0; t++)
		printk(KERN_INFO "%d  %8d  %3d  %3d  %8d  %4d  %8d\n",
			t->use_for_scaling, t->a11clk_khz, t->pll,
			t->a11clk_src_div + 1, t->ahbclk_khz,
			t->ahbclk_div + 1, t->ebi1clk_khz);
	if (turbo_mode)
		printk(KERN_INFO
			"Turbo mode supported (not necessarily enabled).\n");
	else
		printk(KERN_INFO "Turbo mode not supported.\n");
}
#endif

void __init msm_clock_init(struct msm_clock_platform_data *clkdata)
{
	unsigned n;

#ifdef CONFIG_ARCH_MSM
	msm_clock_acpu_tbl_fix_up();
#endif

	/* Start MSM clock driver */
	clock_init(clkdata->acpu_switch_time_us,
				clkdata->max_speed_delta_khz,
				clkdata->vdd_switch_time_us,
				clkdata->active_clocks,
				clkdata->active_clock_count);

	/* Register all the clocks for the MSM7k */
	for (n = 0; n < drv_state.active_clock_count; n++)
		clk_register(drv_state.active_clocks + n);
}

#if defined(CONFIG_DEBUG_FS)
static struct clk *msm_clock_get_nth(unsigned index)
{
	if (index < drv_state.active_clock_count) {
		return drv_state.active_clocks + index;
	} else {
		return 0;
	}
}

static int clock_debug_set(void *data, u64 val)
{
	struct clk *clock = data;
	int ret;

	ret = clk_set_rate(clock, val);
	if (ret != 0)
		printk(KERN_ERR "clk_set_rate failed (%d)\n", ret);
	return ret;
}

static int clock_debug_get(void *data, u64 *val)
{
	struct clk *clock = data;
	*val = clk_get_rate(clock);
	return 0;
}

static int clock_enable_set(void *data, u64 val)
{
	struct clk *clock = data;
	int rc = 0;

	if (val) {
		rc = pc_clk_enable(clock->id);
	} else {
		pc_clk_disable(clock->id);
	}
	return rc;
}

static int clock_enable_get(void *data, u64 *val)
{
	struct clk *clock = data;
	if (clock->id == ACPU_CLK)
		*val = 1;
	else
		*val = pc_clk_is_enabled(clock->id);
	return 0;
}

DEFINE_SIMPLE_ATTRIBUTE(clock_rate_fops, clock_debug_get, clock_debug_set, "%llu\n");
DEFINE_SIMPLE_ATTRIBUTE(clock_enable_fops, clock_enable_get, clock_enable_set, "%llu\n");

static int __init clock_debug_init(void)
{
	struct dentry *dent_rate;
	struct dentry *dent_enable;
	struct clk *clock;
	unsigned n = 0;

	dent_rate = debugfs_create_dir("clk_rate", 0);
	if (IS_ERR(dent_rate))
		return PTR_ERR(dent_rate);

	dent_enable = debugfs_create_dir("clk_enable", 0);
	if (IS_ERR(dent_enable))
		return PTR_ERR(dent_enable);

	while ((clock = msm_clock_get_nth(n++)) != 0) {
		debugfs_create_file(clock->name, 0644, dent_rate,
				    clock, &clock_rate_fops);
		debugfs_create_file(clock->name, 0644, dent_enable,
				    clock, &clock_enable_fops);
	}
	return 0;
}

device_initcall(clock_debug_init);
#endif

/*----------------------------------------------------------------------------
 * Clock driver initialization
 *---------------------------------------------------------------------------*/
void __init clock_init(uint32_t acpu_switch_time_us,
						uint32_t max_speed_delta_khz,
						uint32_t vdd_switch_time_us,
						struct clk *active_clocks,
						int active_clock_count)
{
	pr_info("clock_init()\n");

	mutex_init(&drv_state.lock);
	drv_state.acpu_switch_time_us = acpu_switch_time_us;
	drv_state.max_speed_delta_khz = max_speed_delta_khz;
	drv_state.vdd_switch_time_us = vdd_switch_time_us;
	drv_state.active_clocks = active_clocks;
	drv_state.active_clock_count = active_clock_count;
}

/* The bootloader and/or AMSS may have left various clocks enabled.
 * Disable any clocks that belong to us (CLKFLAG_AUTO_OFF) but have
 * not been explicitly enabled by a clk_enable() call.
 */
static int __init clock_late_init(void)
{
	unsigned long flags;
	struct clk *clk;
	unsigned count = 0;

	mutex_lock(&clocks_mutex);
	list_for_each_entry(clk, &clocks, list) {
		if (clk->flags & CLKFLAG_AUTO_OFF) {
			spin_lock_irqsave(&clk->lock, flags);
			if (!clk->count) {
				count++;
				pc_clk_disable(clk->id);
			}
			spin_unlock_irqrestore(&clk->lock, flags);
		}
	}
	mutex_unlock(&clocks_mutex);
	pr_info("clock_late_init() disabled %d unused clocks\n", count);
	return 0;
}

late_initcall(clock_late_init);

