/*
 *  @file linux/arch/arm/mach-msm/cpu.c
 *  @brief cpufreq_driver callbacks for msm cpu
 *
 *  Copyright (C) 2001-2002 Deep Blue Solutions Ltd.
 *  Copyright (C) 2007-2008 QUALCOMM USA, INC.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * CPU support functions
 */

/*
 *  TODO
 *  -ensure idle load switches to lowest freq
 *  -verify correct hardware support in module_init
 *  -ditto for msm_init
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/cpufreq.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/init.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/mach-types.h>

#include "clock.h"

#define dprintk(msg...) \
	cpufreq_debug_printk(CPUFREQ_DEBUG_DRIVER, "cpufreq-msm", msg)

static struct cpufreq_driver msm_cpufreq_driver;
static struct clk *acpu_clk;

static struct cpufreq_frequency_table msm_freq[20];

#define hz_to_khz(hz) (hz/1000)
#define khz_to_hz(khz) (khz*1000)

/*
 * Validate the speed policy.
 */
static int msm_verify_policy(struct cpufreq_policy *policy)
{
	return cpufreq_frequency_table_verify(policy, msm_freq);
}

static int msm_set_target(struct cpufreq_policy *policy,
			  unsigned int target_freq, unsigned int relation)
{
	cpumask_t cpus_allowed;
	int cpu = policy->cpu;
	struct cpufreq_freqs freqs;
	unsigned int index;
	int rc;

	/*
	 * Save this threads cpus_allowed mask.
	 */
	cpus_allowed = current->cpus_allowed;

	/*
	 * Bind to the specified CPU.  When this call returns,
	 * we should be running on the right CPU.
	 */
	set_cpus_allowed(current, cpumask_of_cpu(cpu));
	BUG_ON(cpu != smp_processor_id());

	cpufreq_frequency_table_target(policy, msm_freq,
				       target_freq, relation, &index);
	/*
	 * - keep close to "target_freq"
	 * - policy->min <= new_freq <= policy->max (THIS MUST BE VALID!!!)
	 * - if relation==CPUFREQ_REL_L, try to select a new_freq higher than
	 *   or equal target_freq. ("L for lowest, but no lower than")
	 * - if relation==CPUFREQ_REL_H, try to select a new_freq lower than
	 *   or equal target_freq. ("H for highest, but no higher than")
	 */

	dprintk("target_freq: %u target_perf: %u\n", msm_freq[index].frequency,
		msm_freq[index].index);

	freqs.old = hz_to_khz(clk_get_rate(acpu_clk));
	freqs.new = msm_freq[index].frequency;
	freqs.cpu = policy->cpu;

	if (freqs.old == freqs.new) {
		set_cpus_allowed(current, cpus_allowed);
		return 0;
	}

	cpufreq_notify_transition(&freqs, CPUFREQ_PRECHANGE);

	rc = clk_set_rate(acpu_clk, khz_to_hz(freqs.new));
	if (rc) {
		printk(KERN_ERR
		       "cpu_msm: clk_set_rate(acpu_clk, %d)=%d\n",
		       khz_to_hz(freqs.new), rc);
	}

	/*
	 * Restore the CPUs allowed mask.
	 */
	set_cpus_allowed(current, cpus_allowed);

	cpufreq_notify_transition(&freqs, CPUFREQ_POSTCHANGE);

	return rc;
}

static unsigned int msm_get(unsigned int cpu)
{
	cpumask_t cpus_allowed;
	unsigned int current_freq;

	cpus_allowed = current->cpus_allowed;

	set_cpus_allowed(current, cpumask_of_cpu(cpu));
	BUG_ON(cpu != smp_processor_id());

	current_freq = hz_to_khz(clk_get_rate(acpu_clk));

	set_cpus_allowed(current, cpus_allowed);

	return current_freq;
}

static int msm_cpufreq_init(struct cpufreq_policy *policy)
{
	int ret, i;

	/* set default policy and cpuinfo */
	policy->governor = CPUFREQ_DEFAULT_GOVERNOR;

	/* latency in ns, 5 ms, RPCs are long and winding */
	policy->cpuinfo.transition_latency = 5000000;

	/* initialize policy limits and hardware(cpuinfo) limits */
	ret = cpufreq_frequency_table_cpuinfo(policy, msm_freq);
	if (ret)
		return (ret);

	cpufreq_frequency_table_get_attr(msm_freq, policy->cpu);

	/* set policy limits for CPU */
	policy->min = ~0;
	policy->max = 0;
	for (i = 0; i < (ARRAY_SIZE(msm_freq) - 1)
		    && msm_freq[i].frequency != CPUFREQ_TABLE_END; i++) {
		if (msm_freq[i].frequency < policy->min)
			policy->min = msm_freq[i].frequency;
		if (msm_freq[i].frequency > policy->max)
			policy->max = msm_freq[i].frequency;
	}
	policy->cur = hz_to_khz(clk_get_rate(acpu_clk));

	return 0;
}

static int msm_cpufreq_exit(struct cpufreq_policy *policy)
{
	cpufreq_frequency_table_put_attr(policy->cpu);
	return 0;
}

static struct cpufreq_driver msm_cpufreq_driver = {
	.name    = "cpu_msm",
	.owner   = THIS_MODULE,
	.flags   = CPUFREQ_CONST_LOOPS,
	.init    = msm_cpufreq_init,
	.verify  = msm_verify_policy,
	.target  = msm_set_target,
	.get     = msm_get,
	.getavg  = NULL,
	.exit    = msm_cpufreq_exit,
	.suspend = NULL,
	.resume  = NULL,
	.attr    = NULL,
};

static int __init msm_cpu_init(void)
{
	unsigned int i;
	unsigned int freq_cnt = 0;

	/* Construct the msm_freq table by looking at the supported
	 * frequencies in the acpu_freq_tbl instead of having a static table
	 * which might not always have the right values for a machine.
	 */
	for (i = 0; acpu_freq_tbl[i].a11clk_khz != 0
			&& freq_cnt < ARRAY_SIZE(msm_freq)-1; i++) {
		if (acpu_freq_tbl[i].use_for_scaling) {
			msm_freq[freq_cnt].index = freq_cnt;
			msm_freq[freq_cnt].frequency
				= acpu_freq_tbl[i].a11clk_khz;
			freq_cnt++;
		}
	}

	if (acpu_freq_tbl[i].a11clk_khz != 0) {
		printk(KERN_ERR "%s: Too many scaling frequencies.\n",
			__FUNCTION__);
		return -ENOMEM;
	}

	msm_freq[freq_cnt].index = freq_cnt;
	msm_freq[freq_cnt].frequency = CPUFREQ_TABLE_END;

	dprintk("%d scaling frequencies supported.\n", freq_cnt);

	acpu_clk = clk_get(NULL, "acpu_clk");
	if (IS_ERR(acpu_clk)) {
		printk(KERN_ERR "clk_get(acpu_clk)= %d\n", (int)acpu_clk);
		return (int)acpu_clk;
	}

	return cpufreq_register_driver(&msm_cpufreq_driver);
}

static void __exit msm_cpu_exit(void)
{
	cpufreq_unregister_driver(&msm_cpufreq_driver);
}

MODULE_AUTHOR("QUALCOMM Inc");
MODULE_DESCRIPTION("cpufreq driver for Qualcomm MSM ARM CPUs");
MODULE_LICENSE("GPL v2");

module_init(msm_cpu_init);
module_exit(msm_cpu_exit);
