/* arch/arm/mach-msm/pm.c
 *
 * MSM Power Management Routines
 *
 * Copyright (C) 2007 Google, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/clk.h>
#include <linux/init.h>
#include <linux/pm.h>
#include <linux/pm_qos_params.h>
#include <linux/proc_fs.h>
#include <linux/suspend.h>
#include <linux/reboot.h>
#include <linux/uaccess.h>
#include <asm/arch/msm_iomap.h>
#include <asm/arch/system.h>
#include <asm/io.h>

#include "smd_private.h"
#include "clock.h"
#include "proc_comm.h"
#ifdef FEATURE_E5832
#include <linux/debugfs.h>

static int  sigmap = ( 1 << USER_MIC_SIG_BIT_18);
static int holdbit_mask = 0xFFFFFFFF;
#endif
#ifdef FEATURE_E5832
#define sleep_wb_timer    20*HZ
#define sleep_sdio_timer    10*HZ
#define sleep_neigh_timer  20*HZ
static int  app_deep_sleep_flag = 0;
#endif

enum {
	MSM_PM_DEBUG_SUSPEND = 1U << 0,
	MSM_PM_DEBUG_POWER_COLLAPSE = 1U << 1,
	MSM_PM_DEBUG_STATE = 1U << 2,
	MSM_PM_DEBUG_CLOCK = 1U << 3,
	MSM_PM_DEBUG_RESET_VECTOR = 1U << 4,
	MSM_PM_DEBUG_SMSM_STATE = 1U << 5,
	MSM_PM_DEBUG_IDLE = 1U << 6,
};
static int msm_pm_debug_mask;
module_param_named(debug_mask, msm_pm_debug_mask, int, S_IRUGO | S_IWUSR | S_IWGRP);

enum {
	MSM_PM_SLEEP_MODE_POWER_COLLAPSE_SUSPEND,
	MSM_PM_SLEEP_MODE_POWER_COLLAPSE,
	MSM_PM_SLEEP_MODE_APPS_SLEEP,
	MSM_PM_SLEEP_MODE_RAMP_DOWN_AND_WAIT_FOR_INTERRUPT,
	MSM_PM_SLEEP_MODE_WAIT_FOR_INTERRUPT,
};
static int msm_pm_sleep_mode = CONFIG_MSM_SLEEP_MODE;
module_param_named(sleep_mode, msm_pm_sleep_mode, int, S_IRUGO | S_IWUSR | S_IWGRP);
static int msm_pm_idle_sleep_mode = CONFIG_MSM_IDLE_SLEEP_MODE;
module_param_named(idle_sleep_mode, msm_pm_idle_sleep_mode, int, S_IRUGO | S_IWUSR | S_IWGRP);
static int msm_pm_idle_sleep_min_time = CONFIG_MSM_IDLE_SLEEP_MIN_TIME;
module_param_named(idle_sleep_min_time, msm_pm_idle_sleep_min_time, int, S_IRUGO | S_IWUSR | S_IWGRP);

#define A11S_CLK_SLEEP_EN (MSM_CSR_BASE + 0x11c)
#define A11S_PWRDOWN (MSM_CSR_BASE + 0x440)
#define A11S_STANDBY_CTL (MSM_CSR_BASE + 0x108)
#define A11RAMBACKBIAS (MSM_CSR_BASE + 0x508)

static struct clk *acpu_clk;
unsigned long pm_saved_acpu_clk_rate;

int acpuclk_set_rate(struct clk *clk, unsigned long rate, int for_power_collapse);

int msm_pm_collapse(void);
int msm_arch_idle(void);
void msm_pm_collapse_exit(void);

int64_t msm_timer_enter_idle(void);
void msm_timer_exit_idle(int low_power);
int msm_irq_idle_sleep_allowed(void);

static uint32_t *msm_pm_reset_vector;

static uint32_t msm_pm_max_sleep_time;

#ifdef CONFIG_MSM_IDLE_STATS
enum msm_pm_time_stats_id {
	MSM_PM_STAT_REQUESTED_IDLE,
	MSM_PM_STAT_IDLE_WFI,
	MSM_PM_STAT_IDLE_SLEEP,
	MSM_PM_STAT_IDLE_FAILED_SLEEP,
	MSM_PM_STAT_NOT_IDLE,
	MSM_PM_STAT_COUNT
};

enum {
	SUSPEND_REQUESTED,
	SUSPEND_FINISHED,
	SUSPEND_COUNT_MAX,
};

static unsigned int msm_pm_suspend_count[SUSPEND_COUNT_MAX];
static uint32_t msm_pm_sleep_limit;
static DECLARE_BITMAP(msm_pm_clocks_no_tcxo_shutdown, NR_CLOCKS);
static DEFINE_MUTEX(pm_stats_mutex);

static struct msm_pm_time_stats {
	const char *name;
	int bucket[CONFIG_MSM_IDLE_STATS_BUCKET_COUNT];
	int64_t min_time[CONFIG_MSM_IDLE_STATS_BUCKET_COUNT];
	int64_t max_time[CONFIG_MSM_IDLE_STATS_BUCKET_COUNT];
	int count;
	int64_t total_time;
} msm_pm_stats[MSM_PM_STAT_COUNT] = {
	[MSM_PM_STAT_REQUESTED_IDLE].name = "idle-request",
	[MSM_PM_STAT_IDLE_WFI].name = "idle-wfi",
	[MSM_PM_STAT_IDLE_SLEEP].name = "idle-sleep",
	[MSM_PM_STAT_IDLE_FAILED_SLEEP].name = "idle-failed-sleep",
	[MSM_PM_STAT_NOT_IDLE].name = "not-idle",
};

static void msm_pm_add_stat(enum msm_pm_time_stats_id id, int64_t t)
{
	int i;
	int64_t bt;

	msm_pm_stats[id].total_time += t;
	msm_pm_stats[id].count++;
	bt = t;
	do_div(bt, CONFIG_MSM_IDLE_STATS_FIRST_BUCKET);
	if (bt < 1ULL << (CONFIG_MSM_IDLE_STATS_BUCKET_SHIFT *
				(CONFIG_MSM_IDLE_STATS_BUCKET_COUNT - 1)))
		i = DIV_ROUND_UP(fls((uint32_t)bt),
					CONFIG_MSM_IDLE_STATS_BUCKET_SHIFT);
	else
		i = CONFIG_MSM_IDLE_STATS_BUCKET_COUNT - 1;
	msm_pm_stats[id].bucket[i]++;
	if (t < msm_pm_stats[id].min_time[i] || !msm_pm_stats[id].max_time[i])
		msm_pm_stats[id].min_time[i] = t;
	if (t > msm_pm_stats[id].max_time[i])
		msm_pm_stats[id].max_time[i] = t;
}
#endif

#define TARGET_CLOCK_RATE 19200000

static int
msm_pm_wait_state(uint32_t wait_state_all_set, uint32_t wait_state_all_clear,
                  uint32_t wait_state_any_set, uint32_t wait_state_any_clear)
{
	int i;
	uint32_t state;

	for (i = 0; i < 100000; i++) {
		state = smsm_get_state(SMSM_MODEM_STATE);
		if (((state & wait_state_all_set) == wait_state_all_set) &&
		    ((~state & wait_state_all_clear) == wait_state_all_clear) &&
		    (wait_state_any_set == 0 || (state & wait_state_any_set) ||
		     wait_state_any_clear == 0 || (state & wait_state_any_clear)))
			return 0;
	}
	printk(KERN_ERR "msm_pm_wait_state(%x, %x, %x, %x) failed %x\n",
	       wait_state_all_set, wait_state_all_clear,
	       wait_state_any_set, wait_state_any_clear, state);
	return -ETIMEDOUT;
}

static int msm_sleep(int sleep_mode, uint32_t sleep_delay,
	uint32_t sleep_limit, int from_idle)
{
	uint32_t saved_vector[2];
	int collapsed;
	void msm_irq_enter_sleep1(bool arm9_wake, int from_idle);
	int msm_irq_enter_sleep2(bool arm9_wake, int from_idle);
	void msm_irq_exit_sleep1(void);
	void msm_irq_exit_sleep2(void);
	void msm_irq_exit_sleep3(void);
	void msm_gpio_enter_sleep(int from_idle);
	void msm_gpio_exit_sleep(void);
	void smd_sleep_exit(void);
	uint32_t enter_state;
	uint32_t enter_wait_set = 0;
	uint32_t enter_wait_clear = 0;
	uint32_t exit_state;
	uint32_t exit_wait_clear = 0;
	uint32_t exit_wait_set = 0;
	int ret;
	int rv = -EINTR;

	if (msm_pm_debug_mask & MSM_PM_DEBUG_SUSPEND)
		printk(KERN_INFO "msm_sleep(): "
			"mode %d delay %u limit %u idle %d\n",
			sleep_mode, sleep_delay, sleep_limit, from_idle);

#ifdef CONFIG_MSM_IDLE_STATS
	msm_pm_sleep_limit = sleep_limit;
#endif

	switch (sleep_mode) {
	case MSM_PM_SLEEP_MODE_POWER_COLLAPSE:
		enter_state = SMSM_PWRC;
		enter_wait_set = SMSM_RSA;
		exit_state = SMSM_WFPI;
		exit_wait_clear = SMSM_RSA;
		break;
	case MSM_PM_SLEEP_MODE_POWER_COLLAPSE_SUSPEND:
		enter_state = SMSM_PWRC_SUSPEND;
		enter_wait_set = SMSM_RSA;
		exit_state = SMSM_WFPI;
		exit_wait_clear = SMSM_RSA;
		break;
	case MSM_PM_SLEEP_MODE_APPS_SLEEP:
		enter_state = SMSM_SLEEP;
		exit_state = SMSM_SLEEPEXIT;
		exit_wait_set = SMSM_SLEEPEXIT;
		break;
	default:
		enter_state = 0;
		exit_state = 0;
	}

	msm_irq_enter_sleep1(!!enter_state, from_idle);
	msm_gpio_enter_sleep(from_idle);

	if (enter_state) {
		if (sleep_delay == 0 && sleep_mode >= MSM_PM_SLEEP_MODE_APPS_SLEEP)
			sleep_delay = 192000*5;
		smsm_set_sleep_duration(sleep_delay);
		smsm_set_sleep_limitation(sleep_limit);
		ret = smsm_change_state(SMSM_APPS_STATE, SMSM_RUN, enter_state);
		if (ret) {
			printk(KERN_ERR "msm_sleep(): smsm_change_state %x failed\n", enter_state);
			enter_state = 0;
			exit_state = 0;
		}
		ret = msm_pm_wait_state(enter_wait_set, enter_wait_clear, 0, 0);
		if (ret) {
			printk(KERN_INFO "msm_sleep(): "
			       "msm_pm_wait_state failed, %x\n",
			       smsm_get_state(SMSM_MODEM_STATE));
			goto enter_failed;
		}
	}
	if (msm_irq_enter_sleep2(!!enter_state, from_idle))
		goto enter_failed;

	if (enter_state) {
		writel(0x1f, A11S_CLK_SLEEP_EN);
		writel(1, A11S_PWRDOWN);

		writel(0, A11S_STANDBY_CTL);
		writel(0, A11RAMBACKBIAS);

		if (msm_pm_debug_mask & MSM_PM_DEBUG_STATE)
			printk(KERN_INFO "msm_sleep(): enter "
			       "A11S_CLK_SLEEP_EN %x, A11S_PWRDOWN %x, "
			       "smsm_get_state %x\n", readl(A11S_CLK_SLEEP_EN),
			       readl(A11S_PWRDOWN),
			       smsm_get_state(SMSM_MODEM_STATE));
	}

	if (sleep_mode <= MSM_PM_SLEEP_MODE_RAMP_DOWN_AND_WAIT_FOR_INTERRUPT) {
		pm_saved_acpu_clk_rate = clk_get_rate(acpu_clk);
		if (msm_pm_debug_mask & MSM_PM_DEBUG_CLOCK)
			printk(KERN_INFO "msm_sleep(): change clk %ld -> %d"
			       "\n", pm_saved_acpu_clk_rate, TARGET_CLOCK_RATE);
		acpuclk_set_rate(acpu_clk, TARGET_CLOCK_RATE, 1);
	}
	if (sleep_mode < MSM_PM_SLEEP_MODE_APPS_SLEEP) {
		if (msm_pm_debug_mask & MSM_PM_DEBUG_SMSM_STATE)
			smsm_print_sleep_info();
		saved_vector[0] = msm_pm_reset_vector[0];
		saved_vector[1] = msm_pm_reset_vector[1];
		msm_pm_reset_vector[0] = 0xE51FF004; /* ldr pc, 4 */
		msm_pm_reset_vector[1] = virt_to_phys(msm_pm_collapse_exit);
		if (msm_pm_debug_mask & MSM_PM_DEBUG_RESET_VECTOR)
			printk(KERN_INFO "msm_sleep(): vector %x %x -> "
			       "%x %x\n", saved_vector[0], saved_vector[1],
			       msm_pm_reset_vector[0], msm_pm_reset_vector[1]);
		collapsed = msm_pm_collapse();
		msm_pm_reset_vector[0] = saved_vector[0];
		msm_pm_reset_vector[1] = saved_vector[1];
		if (collapsed) {
			cpu_init();
			local_fiq_enable();
			rv = 0;
		}
		if (msm_pm_debug_mask & MSM_PM_DEBUG_POWER_COLLAPSE)
			printk(KERN_INFO "msm_pm_collapse(): returned %d\n",
			       collapsed);
		if (msm_pm_debug_mask & MSM_PM_DEBUG_SMSM_STATE)
			smsm_print_sleep_info();
	} else {
		msm_arch_idle();
		rv = 0;
	}

	if (sleep_mode <= MSM_PM_SLEEP_MODE_RAMP_DOWN_AND_WAIT_FOR_INTERRUPT) {
		if (msm_pm_debug_mask & MSM_PM_DEBUG_CLOCK)
			printk(KERN_INFO "msm_sleep(): change clk %d -> %ld"
			       "\n", TARGET_CLOCK_RATE, pm_saved_acpu_clk_rate);
		if (acpuclk_set_rate(acpu_clk, pm_saved_acpu_clk_rate, 1) < 0)
			printk(KERN_ERR "msm_sleep(): clk_set_rate %ld "
			       "failed\n", pm_saved_acpu_clk_rate);
	}
	if (msm_pm_debug_mask & MSM_PM_DEBUG_STATE)
		printk(KERN_INFO "msm_sleep(): exit A11S_CLK_SLEEP_EN %x, "
		       "A11S_PWRDOWN %x, smsm_get_state %x\n",
		       readl(A11S_CLK_SLEEP_EN), readl(A11S_PWRDOWN),
		       smsm_get_state(SMSM_MODEM_STATE));
	msm_irq_exit_sleep1();
enter_failed:
	if (enter_state) {
		writel(0x00, A11S_CLK_SLEEP_EN);
		writel(0, A11S_PWRDOWN);
		smsm_change_state(SMSM_APPS_STATE, enter_state, exit_state);
		msm_pm_wait_state(exit_wait_set, exit_wait_clear, 0, 0);
		if (msm_pm_debug_mask & MSM_PM_DEBUG_STATE)
			printk(KERN_INFO "msm_sleep(): sleep exit "
			       "A11S_CLK_SLEEP_EN %x, A11S_PWRDOWN %x, "
			       "smsm_get_state %x\n", readl(A11S_CLK_SLEEP_EN),
			       readl(A11S_PWRDOWN),
			       smsm_get_state(SMSM_MODEM_STATE));
		if (msm_pm_debug_mask & MSM_PM_DEBUG_SMSM_STATE)
			smsm_print_sleep_info();
	}
	msm_irq_exit_sleep2();
	if (enter_state) {
		smsm_change_state(SMSM_APPS_STATE, exit_state, SMSM_RUN);
		msm_pm_wait_state(SMSM_RUN, 0, 0, 0);
		if (msm_pm_debug_mask & MSM_PM_DEBUG_STATE)
			printk(KERN_INFO "msm_sleep(): sleep exit "
			       "A11S_CLK_SLEEP_EN %x, A11S_PWRDOWN %x, "
			       "smsm_get_state %x\n", readl(A11S_CLK_SLEEP_EN),
			       readl(A11S_PWRDOWN),
			       smsm_get_state(SMSM_MODEM_STATE));
	}
	msm_irq_exit_sleep3();
	msm_gpio_exit_sleep();
	smd_sleep_exit();
	return rv;
}

#ifdef FEATURE_E5832
extern void adjust_wb_timer(unsigned int sleep_wb_time);
typedef void (*adjust_sdio_timer)(unsigned int sleep_sdio_time);
extern void adjust_br_timer(unsigned int hello_time, unsigned int forword_time);
typedef void (*adjust_neigh_timer)(unsigned int sleep_neigh_time);
adjust_neigh_timer pneighfunc = NULL;
adjust_sdio_timer psdiofunc = NULL;
#ifdef FEATURE_WIFI_INTERRUPT
typedef int (*wlan_callback)(void);
wlan_callback wlan_suspend_func = NULL;
wlan_callback wlan_resume_func = NULL;
void unregister_wlan_suspend_func(void)
{
    wlan_suspend_func = NULL;
}
void unregister_wlan_resume_func(void)
{
    wlan_resume_func = NULL;
}
void register_wlan_resume_func(wlan_callback pfunc)
{
  wlan_resume_func = pfunc;
}
void register_wlan_suspend_func(wlan_callback pfunc)
{
  printk("Function:register_wlan_suspend_func()\n");
  wlan_suspend_func = pfunc;
}
EXPORT_SYMBOL(register_wlan_resume_func);
EXPORT_SYMBOL(wlan_resume_func);
EXPORT_SYMBOL(register_wlan_suspend_func);
EXPORT_SYMBOL(unregister_wlan_suspend_func);
EXPORT_SYMBOL(unregister_wlan_resume_func);
#endif
void register_adjust_sdio_func( adjust_sdio_timer pfunc )
{
    psdiofunc = pfunc;
}
void register_adjust_neigh_func( adjust_neigh_timer pfunc )
{
    pneighfunc = pfunc;
}
EXPORT_SYMBOL(register_adjust_sdio_func);
void unregister_adjust_sdio_func(void)
{
    psdiofunc = NULL;
}
EXPORT_SYMBOL(unregister_adjust_sdio_func);
static void enter_app_deep_sleep(void)
{    
    if( app_deep_sleep_flag == 0)
    {
        app_deep_sleep_flag = 1;
        
        if(NULL != adjust_wb_timer)
        {
            printk(KERN_EMERG "adjust wb timer time to 10HZ!\n");
            adjust_wb_timer(sleep_wb_timer);
        }
        if(NULL != psdiofunc)
        {
            printk(KERN_EMERG "adjust to add sdio polling time 5HZ!\n");
            psdiofunc(sleep_sdio_timer);
        }

        if (NULL != pneighfunc)
        {
            printk(KERN_EMERG "adjust to add sdio polling time 5HZ!\n");
            pneighfunc(sleep_neigh_timer);
        }

    }
}

void exit_app_deep_sleep(void)
{
    if (1 == app_deep_sleep_flag)
    {
        app_deep_sleep_flag = 0;
        if(NULL != adjust_wb_timer)
        {
            printk(KERN_EMERG "restore wb timer time!\n");
            adjust_wb_timer(0);
        }
        if(NULL != psdiofunc)
        {
            printk(KERN_EMERG "restore sdio polling time!\n");
            psdiofunc(0);
        }

        if (NULL != pneighfunc)
    	 {
           printk(KERN_EMERG "restore neigh polling time!\n");
           pneighfunc(0);
        }

#ifdef FEATURE_WIFI_INTERRUPT
        if(NULL != wlan_resume_func)
        {
          wlan_resume_func();
        }
#endif
    }
}
EXPORT_SYMBOL(exit_app_deep_sleep);
/*------------------------------------------------------------
  ԭ : int driver_vote_sleep(int map, int flag)
       : driver vote to enable sleep
       : map, identify which driver vote for sleep, flag, 1: disable or 0 : enable sleep
       : null
  ֵ   : null
-------------------------------------------------------------*/
int vote_for_sleep(int map, int flag)
{
     printk(KERN_EMERG "sigmap = 0x%x, vote = 0x%x, value = %d\n", sigmap, (1 << map) , flag);
     if(0 == flag)
     {
          sigmap = sigmap & ~(1 << map);
     }
     else if(1 == flag)
     {
          exit_app_deep_sleep();
          sigmap = sigmap | (1 << map);
     }
     else
    {
        printk(KERN_EMERG "map = %x, value = %x is incorrect \n", map, flag);
     }
     printk(KERN_EMERG "sigmap = 0x%x, vote = 0x%x, value = %d\n", sigmap, (1 << map) , flag);
     return 0;
}

/*------------------------------------------------------------
  ԭ : int driver_idle_sleep_allowed()
       : if drivers can be sleep
       : null
       : null
  ֵ   : 1: can be sleep, 0: can not sleep
-------------------------------------------------------------*/
int krl_idle_sleep_allowed(void)
{
    return  !(sigmap & (( 1 << KRL_USB_SIG_BIT_1) | ( 1 << KRL_GPIO_SIG_BIT_2) | ( 1 << KRL_SERIAL_SIG_BIT_3))); /// BB5D01710.DIAG: yangjianping 2009-6-30 ASD SERIAL Vote 
}

/*------------------------------------------------------------
  ԭ : int userspace_idle_sleep_allowed()
       : if user space process can be sleep
       : null
       : null
  ֵ   : 1: can be sleep, 0: can not sleep
-------------------------------------------------------------*/
int app_idle_sleep_allowed(void)
{
    return  !(sigmap & (( 1 << USER_MIC_SIG_BIT_18) |( 1 << USER_WLAN_SIG_BIT_17)|(1 << USER_CALL_SIG_BIT_19))); // <BB5D01125.SLEEP: yangjianping 2009-6-23 ADD Call vote
}

EXPORT_SYMBOL(vote_for_sleep);

int get_sigmap(void)
{
    return  sigmap;
}
EXPORT_SYMBOL(get_sigmap);
#ifdef FEATURE_WIFI_INTERRUPT
EXPORT_SYMBOL(wlan_suspend_func);
#endif

#endif
void arch_idle(void)
{
	int ret;
	int64_t sleep_time;
	int low_power = 0;
#ifdef CONFIG_MSM_IDLE_STATS
	int64_t t1;
	static int64_t t2;
	int exit_stat;
#endif
	int latency_qos = pm_qos_requirement(PM_QOS_INTERRUPT_LATENCY);
	uint32_t sleep_limit = 0;
	int allow_sleep =
		msm_pm_idle_sleep_mode < MSM_PM_SLEEP_MODE_WAIT_FOR_INTERRUPT &&
		msm_irq_idle_sleep_allowed();
#ifdef FEATURE_E5832
       allow_sleep &= krl_idle_sleep_allowed() && app_idle_sleep_allowed();
#endif

	sleep_time = msm_timer_enter_idle();
#ifdef CONFIG_MSM_IDLE_STATS
	t1 = ktime_to_ns(ktime_get());
	msm_pm_add_stat(MSM_PM_STAT_NOT_IDLE, t1 - t2);
	msm_pm_add_stat(MSM_PM_STAT_REQUESTED_IDLE, sleep_time);
#endif

	if (CONFIG_MSM_TCXO_SHUTDOWN_LATENCY < latency_qos) {
		sleep_limit = 0;
	} else if (CONFIG_MSM_POWER_COLLAPSE_LATENCY < latency_qos) {
		sleep_limit = 1;
	} else if (CONFIG_MSM_SWFI_LATENCY < latency_qos) {
		allow_sleep = false;
	} else {
		/* no time even for SWFI */
		return;
	}

	if (msm_pm_debug_mask & MSM_PM_DEBUG_IDLE)
		printk(KERN_INFO "arch_idle: sleep time %llu, allow_sleep %d\n",
		       sleep_time, allow_sleep);
	if (sleep_time < msm_pm_idle_sleep_min_time || !allow_sleep) {
		msm_arch_idle();
#ifdef CONFIG_MSM_IDLE_STATS
		exit_stat = MSM_PM_STAT_IDLE_WFI;
#endif
  	} else {
#ifdef CONFIG_MSM_IDLE_STATS
		DECLARE_BITMAP(clk_ids, NR_CLOCKS);

		ret = msm_clock_require_tcxo(clk_ids, NR_CLOCKS);
		if (ret > 0)
			bitmap_copy(msm_pm_clocks_no_tcxo_shutdown, clk_ids,
				NR_CLOCKS);
#else
		ret = msm_clock_require_tcxo(NULL, 0);
#endif
		if (ret)
			sleep_limit = 1;

		low_power = 1;
		do_div(sleep_time, NSEC_PER_SEC / 32768);
		if (sleep_time > 0x6DDD000) {
			printk("sleep_time too big %lld\n", sleep_time);
			sleep_time = 0x6DDD000;
		}
#ifdef FEATURE_WIFI_INTERRUPT
      if(NULL != wlan_suspend_func)
      {
  			if(0 != wlan_suspend_func())
  			{
  			  printk(KERN_INFO"Go to sleeping\n");
  				return;
  			}			
      }    
#endif
      enter_app_deep_sleep();
      ret = msm_sleep(msm_pm_idle_sleep_mode, sleep_time,sleep_limit, 1);
#ifdef CONFIG_MSM_IDLE_STATS
		if (ret)
			exit_stat = MSM_PM_STAT_IDLE_FAILED_SLEEP;
		else
             exit_stat = MSM_PM_STAT_IDLE_SLEEP;
#endif
	}
	msm_timer_exit_idle(low_power);
#ifdef CONFIG_MSM_IDLE_STATS
	t2 = ktime_to_ns(ktime_get());
	msm_pm_add_stat(exit_stat, t2 - t1);
#endif
}

static int msm_pm_enter(suspend_state_t state)
{
	int ret;
#ifdef CONFIG_MSM_IDLE_STATS
	DECLARE_BITMAP(clk_ids, NR_CLOCKS);

	ret = msm_clock_require_tcxo(clk_ids, NR_CLOCKS);
	if (ret > 0)
		bitmap_copy(msm_pm_clocks_no_tcxo_shutdown, clk_ids,
			NR_CLOCKS);
#else
	ret = msm_clock_require_tcxo(NULL, 0);
#endif

	ret = msm_sleep(msm_pm_sleep_mode,
		msm_pm_max_sleep_time, (uint32_t)ret, 0);

#ifdef CONFIG_MSM_IDLE_STATS
	if (msm_pm_sleep_mode == MSM_PM_SLEEP_MODE_POWER_COLLAPSE_SUSPEND) {
		msm_pm_suspend_count[SUSPEND_REQUESTED]++;
		if (!ret)
			msm_pm_suspend_count[SUSPEND_FINISHED]++;
	}
#endif

	return 0;
}

static struct platform_suspend_ops msm_pm_ops = {
	.enter		= msm_pm_enter,
	.valid		= suspend_valid_only_mem,
};

static uint32_t restart_reason = 0x776655AA;

static void msm_pm_power_off(void)
{
	msm_proc_comm(PCOM_POWER_DOWN, 0, 0);
	for (;;) ;
}

static void msm_pm_restart(char str)
{
	/* If there's a hard reset hook and the restart_reason
	 * is the default, prefer that to the (slower) proc_comm
	 * reset command.
	 */
	if ((restart_reason == 0x776655AA) && msm_reset_hook) {
		msm_reset_hook(str);
	} else {
		msm_proc_comm(PCOM_RESET_CHIP, &restart_reason, 0);
	}
	for (;;) ;
}

static int msm_reboot_call(struct notifier_block *this, unsigned long code, void *_cmd)
{
	if((code == SYS_RESTART) && _cmd) {
		char *cmd = _cmd;
		if (!strcmp(cmd, "bootloader")) {
			restart_reason = 0x77665500;
		} else if (!strcmp(cmd, "recovery")) {
			restart_reason = 0x77665502;
		} else if (!strcmp(cmd, "eraseflash")) {
			restart_reason = 0x776655EF;
		} else if (!strncmp(cmd, "oem-", 4)) {
			unsigned code = simple_strtoul(cmd + 4, 0, 16) & 0xff;
			restart_reason = 0x6f656d00 | code;
		} else {
			restart_reason = 0x77665501;
		}
	}
	return NOTIFY_DONE;
}

static struct notifier_block msm_reboot_notifier =
{
	.notifier_call = msm_reboot_call,
};

#ifdef CONFIG_MSM_IDLE_STATS
static int msm_pm_read_proc(char *page, char **start, off_t off,
                               int count, int *eof, void *data)
{
	int len = 0;
	int i, j;
	char *p = page;
	char clk_name[16];

	p += sprintf(p, "Last clocks against TCXO shutdown:\n");
	for_each_bit(i, msm_pm_clocks_no_tcxo_shutdown, NR_CLOCKS) {
		clk_name[0] = '\0';
		msm_clock_get_name(i, clk_name, sizeof(clk_name));
		p += sprintf(p, "  %s (id=%d)\n", clk_name, i);
	}

	p += sprintf(p, "Last power collapse voted ");
	if (msm_pm_sleep_limit)
		p += sprintf(p, "against TCXO shutdown\n\n");
	else
		p += sprintf(p, "for TCXO shutdown\n\n");

	mutex_lock(&pm_stats_mutex);
	p += sprintf(p, "suspend count:\n\trequested=%u, finished=%u\n",
			msm_pm_suspend_count[SUSPEND_REQUESTED],
			msm_pm_suspend_count[SUSPEND_FINISHED]);

	for (i = 0; i < ARRAY_SIZE(msm_pm_stats); i++) {
		int64_t bucket_time;
		int64_t s;
		uint32_t ns;
		s = msm_pm_stats[i].total_time;
		ns = do_div(s, NSEC_PER_SEC);
		p += sprintf(p,
			"%s:\n"
			"  count: %7d\n"
			"  total_time: %lld.%09u\n",
			msm_pm_stats[i].name,
			msm_pm_stats[i].count,
			s, ns);
		bucket_time = CONFIG_MSM_IDLE_STATS_FIRST_BUCKET;
		for (j = 0; j < CONFIG_MSM_IDLE_STATS_BUCKET_COUNT - 1; j++) {
			s = bucket_time;
			ns = do_div(s, NSEC_PER_SEC);
			p += sprintf(p, "   <%2lld.%09u: %7d (%lld-%lld)\n",
				s, ns, msm_pm_stats[i].bucket[j],
				msm_pm_stats[i].min_time[j],
				msm_pm_stats[i].max_time[j]);
			bucket_time <<= CONFIG_MSM_IDLE_STATS_BUCKET_SHIFT;
		}
		p += sprintf(p, "  >=%2lld.%09u: %7d (%lld-%lld)\n",
			s, ns, msm_pm_stats[i].bucket[j],
			msm_pm_stats[i].min_time[j],
			msm_pm_stats[i].max_time[j]);
	}
	mutex_unlock(&pm_stats_mutex);
	*start = page + off;

	len = p - page;
	if (len > off)
		len -= off;
	else
		len = 0;

	return len < count ? len  : count;
}

static int msm_pm_write_proc(struct file *file, const char __user *buf,
		unsigned long count, void *data)
{
	char input[2] = {0};
	int i, j;

	if (count != 2)
		return -EINVAL;

	if (copy_from_user(input, buf, count))
		return -EFAULT;

	if (input[0] == '0') {
		mutex_lock(&pm_stats_mutex);
		for (i = 0; i < MSM_PM_STAT_COUNT; i++) {
			for (j = 0;
				j < CONFIG_MSM_IDLE_STATS_BUCKET_COUNT;
				j++) {
				msm_pm_stats[i].bucket[j] = 0;
				msm_pm_stats[i].min_time[j] = 0;
				msm_pm_stats[i].max_time[j] = 0;
			}
			msm_pm_stats[i].count = 0;
			msm_pm_stats[i].total_time = 0;
		}
		memset(msm_pm_suspend_count, 0, sizeof(msm_pm_suspend_count));
		mutex_unlock(&pm_stats_mutex);
	}

	return count;
}
#endif

void msm_pm_set_max_sleep_time(int64_t max_sleep_time_ns)
{
	int64_t max_sleep_time_bs = max_sleep_time_ns;

	/* Convert from ns -> BS units */
	do_div(max_sleep_time_bs, NSEC_PER_SEC / 32768);

	if (max_sleep_time_bs > 0x6DDD000)
		msm_pm_max_sleep_time = (uint32_t) 0x6DDD000;
	else
		msm_pm_max_sleep_time = (uint32_t) max_sleep_time_bs;

	if (msm_pm_debug_mask & MSM_PM_DEBUG_SUSPEND)
		printk("%s: Requested %lldns (%lldbs), Giving %ubs\n",
		       __func__, max_sleep_time_ns, 
		       max_sleep_time_bs, 
		       msm_pm_max_sleep_time);
}
EXPORT_SYMBOL(msm_pm_set_max_sleep_time);

#ifdef FEATURE_E5832
int sleep_app_knl_set(void *data, u64 val)
{
     sigmap = val;
     return 0;
}
EXPORT_SYMBOL(sleep_app_knl_set);

static int sleep_app_knl_get(void *data, u64 *val)
{
	*val = sigmap;
	return 0;
}
static int sleep_vote_bit_clr(void *data, u64 val)
{
     holdbit_mask = (1 << val);
     vote_for_sleep(val, 0);
     return 0;
}

static int sleep_vote_bit_get(void *data, u64 *val)
{
	*val = sigmap & holdbit_mask;
	return 0;
}

static int sleep_vote_bit_set(void *data, u64 val)
{
     holdbit_mask = (1 << val);
     vote_for_sleep(val, 1);
     return 0;
}

DEFINE_SIMPLE_ATTRIBUTE(sleep_app_knl_fops, sleep_app_knl_get, sleep_app_knl_set, "%llu\n");
DEFINE_SIMPLE_ATTRIBUTE(sleep_clr_fops, sleep_vote_bit_get, sleep_vote_bit_clr, "%llu\n");
DEFINE_SIMPLE_ATTRIBUTE(sleep_set_fops, sleep_vote_bit_get, sleep_vote_bit_set, "%llu\n");
#endif

static int __init msm_pm_init(void)
{
#ifdef CONFIG_MSM_IDLE_STATS
	struct proc_dir_entry *pe;
#endif

	struct dentry *dent;

	pm_power_off = msm_pm_power_off;
	arm_pm_restart = msm_pm_restart;
	msm_pm_max_sleep_time = 0;

	register_reboot_notifier(&msm_reboot_notifier);

	acpu_clk = clk_get(NULL, "acpu_clk");
	if (acpu_clk == NULL) {
		printk(KERN_ERR "msm_pm_init: failed get acpu_clk\n");
		return -ENODEV;
	}
#if defined(CONFIG_ARCH_QSD)
	/* The bootloader is responsible for initializing many of Scorpion's
	 * coprocessor registers for things like cache timing. The state of
	 * these coprocessor registers is lost on reset, so part of the
	 * bootloader must be re-executed. Do not overwrite the reset vector
	 * or bootloader area.
	 */
	msm_pm_reset_vector = PAGE_OFFSET;
#else
	msm_pm_reset_vector = ioremap(0, PAGE_SIZE);
	if (msm_pm_reset_vector == NULL) {
		printk(KERN_ERR "msm_pm_init: failed to map reset vector\n");
		return -ENODEV;
	}
#endif /* CONFIG_ARCH_QSD */

	suspend_set_ops(&msm_pm_ops);

#ifdef CONFIG_MSM_IDLE_STATS
	pe = create_proc_entry("msm_pm_stats", 0644, NULL);
	if (!pe)
		return -ENOMEM;
	pe->read_proc = msm_pm_read_proc;
	pe->write_proc = msm_pm_write_proc;
#endif
	dent = debugfs_create_dir("sleep", 0);
	if (IS_ERR(dent))
	{
	    return 0;
	}

	debugfs_create_file("summary", 0644, dent, NULL,  &sleep_app_knl_fops);
	debugfs_create_file("sleep_enable_bit", 0644, dent, NULL,  &sleep_clr_fops);
	debugfs_create_file("sleep_disable_bit", 0644, dent, NULL,  &sleep_set_fops);

	return 0;
}

__initcall(msm_pm_init);
