/*
 * SOC Info Routines
 *
 * Copyright (C) 2008 QUALCOMM USA, INC.
 * 
 * All source code in this file is licensed under the following license
 * except where indicated.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 *
 */

#include <linux/types.h>
#include <linux/sysdev.h>
#include "socinfo.h"
#include "smd_private.h"

/* Used to parse shared memory.  Must match the modem. */
struct socinfo {
	u32 format;
	u32 id;
	u32 version;
	char build_id[32];
};

static struct socinfo *socinfo;

int socinfo_get_id(void)
{
	return (socinfo) ? socinfo->id : 0;
}

int socinfo_get_version(void)
{
	return (socinfo) ? socinfo->version : 0;
}

char *socinfo_get_build_id(void)
{
	return (socinfo) ? socinfo->build_id : NULL;
}

static ssize_t
socinfo_show_id(struct sys_device *dev, char *buf)
{
	if (!socinfo) {
		printk(KERN_ERR "%s: No socinfo found!", __func__);
		return 0;
	}

	return snprintf(buf, PAGE_SIZE, "%d\n",
			socinfo_get_id());
}

static ssize_t
socinfo_show_version(struct sys_device *dev, char *buf)
{
	int version;

	if (!socinfo) {
		printk(KERN_ERR "%s: No socinfo found!", __func__);
		return 0;
	}

	version = socinfo_get_version();
	return snprintf(buf, PAGE_SIZE, "%d.%d\n",
			SOCINFO_VERSION_MAJOR(version),
			SOCINFO_VERSION_MINOR(version));
}

static ssize_t
socinfo_show_build_version(struct sys_device *dev, char *buf)
{
	if (!socinfo) {
		printk(KERN_ERR "%s: No socinfo found!", __func__);
		return 0;
	}

	return snprintf(buf, PAGE_SIZE, "%-.32s\n",
			socinfo_get_build_id());
}

static SYSDEV_ATTR(id, 0444, socinfo_show_id, NULL);
static SYSDEV_ATTR(version, 0444, socinfo_show_version, NULL);
static SYSDEV_ATTR(build_version, 0444, socinfo_show_build_version, NULL);

struct sysdev_class soc_sysdev_class = {
	.name = "soc",
};

static struct sys_device soc_sys_device = {
	.id = 0,
	.cls = &soc_sysdev_class,
};

static void __init socinfo_init_sysdev(void)
{
	int err;

	err = sysdev_class_register(&soc_sysdev_class);
	if (err) {
		printk(KERN_ERR "%s: sysdev_class_register fail (%d)\n",
		       __func__, err);
		return;
	}
	err = sysdev_register(&soc_sys_device);
	if (err) {
		printk(KERN_ERR "%s: sysdev_register fail (%d)\n",
		       __func__, err);
		return;
	}
	err = sysdev_create_file(&soc_sys_device, &attr_id);
	if (err) {
		printk(KERN_ERR "%s: sysdev_create_file"
		       " (id) fail (%d)\n",
		       __func__, err);
		return;
	}
	err = sysdev_create_file(&soc_sys_device, &attr_version);
	if (err) {
		printk(KERN_ERR "%s: sysdev_create_file"
		       " (version) fail (%d)\n",
		       __func__, err);
		return;
	}
	err = sysdev_create_file(&soc_sys_device, &attr_build_version);
	if (err) {
		printk(KERN_ERR "%s: sysdev_create_file"
		       " (build_version) fail (%d)\n",
		       __func__, err);
		return;
	}
}

int __init socinfo_init(void)
{
	socinfo = smem_alloc(SMEM_HW_SW_BUILD_ID, sizeof(struct socinfo));
	if (!socinfo) {
		printk(KERN_ERR "%s: Can't find SMEM_HW_SW_BUILD_ID\n",
		       __func__);
		return -EIO;
	}
	socinfo_init_sysdev();
	return 0;
}
