/*
 * drivers/ide/arm/s1r72v05.c
 *
 * Copyright (c) 2008 QUALCOMM USA, INC.
 * 
 * All source code in this file is licensed under the following license
 * except where indicated.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 *
 * Platform driver for Epson S1R72V05 IDE interface.
 */

#include <linux/module.h>
#include <linux/ide.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <asm/gpio.h>

#define DRIVER_NAME "s1r72v05"
#define DRIVER_VERSION "1.0"

#define S1R72V05_IO_EXTENT (s1r72v05_data.io_res->end - \
			    s1r72v05_data.io_res->start + 1)
#define S1R72V05_HW_REVISION 0x50

/* parameters for changing power management states */
#define S1R72V05_PM_WAIT_USEC          50
#define S1R72V05_MAX_PM_WAIT_USEC      1000

/* chip config:
   bit 7: XINT active high
   bit 6: XINT 1/0 mode
   bit 5: XDREQ active low
   bit 4: XDACK active low
   bit 3: DMA0/1 is enabled by assertion of chip select and XDACK
   bit 2: endianness (even address is lower byte)
   bit 1-0: 16-bit BE mode
*/
#define S1R72V05_CHIPCONFIG_VALUE      0x0e

/* S1R72V05 IDE taskfile register identifiers */
#define S1R72V05_DATA_TF_REG           0x0
#define S1R72V05_ERROR_TF_REG          0x1
#define S1R72V05_FEATURES_TF_REG       0x1
#define S1R72V05_SECTOR_CNT_TF_REG     0x2
#define S1R72V05_LBA_LOW_TF_REG        0x3
#define S1R72V05_LBA_MID_TF_REG        0x4
#define S1R72V05_LBA_HIGH_TF_REG       0x5
#define S1R72V05_DEVICE_TF_REG         0x6
#define S1R72V05_STATUS_TF_REG         0x7
#define S1R72V05_COMMAND_TF_REG        0x7
#define S1R72V05_DEVICE_CONTROL_TF_REG 0xE
#define S1R72V05_ALT_STATUS_TF_REG     0xE
#define S1R72V05_UNSUPPORTED           -1

#define S1R72V05_TF_WRITE              0x80
#define S1R72V05_TF_READ               0x40

/* S1R72V05 register map (byte offsets from start of chip select) */
#define S1R72V05_MAININTSTAT_OFFSET             0x00
#define   S1R72V05_MAININTSTAT_FINISHEDPM       0x01
#define   S1R72V05_MAININTSTAT_IDE_INTSTAT      0x10
#define   S1R72V05_MAININTSTAT_ALLINTS          0xF9
#define S1R72V05_DEVICEINTSTAT_OFFSET           0x01
#define S1R72V05_HOSTINTSTAT_OFFSET             0x02
#define S1R72V05_CPUINTSTAT_OFFSET              0x03
#define S1R72V05_IDE_INTSTAT_OFFSET             0x04
#define   S1R72V05_IDE_INTSTAT_DETECTINTRQ      0x02
#define   S1R72V05_IDE_INTSTAT_ALLINTS          0xF7
#define S1R72V05_MAININTENB_OFFSET              0x10
#define   S1R72V05_MAININTENB_ENIDE_INTSTAT     0x10
#define S1R72V05_IDE_INTENB_OFFSET              0x14
#define   S1R72V05_IDE_INTENB_ENDETECTINTRQ     0x02
#define S1R72V05_REVISIONNUM_OFFSET             0x20
#define S1R72V05_CHIPRESET_SWAPPED_OFFSET       0x20
#define S1R72V05_CHIPRESET_OFFSET               0x21
#define S1R72V05_REVISIONNUM_SWAPPED_OFFSET     0x21
#define S1R72V05_PM_CONTROL_0_OFFSET            0x22
#define   S1R72V05_PM_CONTROL_0_GOACTIVE60      0x20
#define S1R72V05_PM_CONTROL_1_OFFSET            0x23
#define S1R72V05_IDE_STATUS_OFFSET              0x90
#define S1R72V05_IDE_CONTROL_OFFSET             0x91
#define   S1R72V05_IDE_CONTROL_IDE_CLR          0x40
#define S1R72V05_IDE_CONFIG_0_OFFSET            0x92
#define   S1R72V05_IDE_CONFIG_0_IDE_BUSRESET    0x80
#define S1R72V05_IDE_CONFIG_1_OFFSET            0x93
#define   S1R72V05_IDE_CONFIG_1_SWAP            0x04
#define   S1R72V05_IDE_CONFIG_1_ACTIVEIDE       0x80
#define S1R72V05_IDE_RMOD_OFFSET                0x94
#define   S1R72V05_IDE_RMOD_MAX_TIMING          0xFF
#define S1R72V05_IDE_REGADRS_OFFSET             0xA0
#define S1R72V05_IDE_RDREGVALUE_0_OFFSET        0xA2
#define S1R72V05_IDE_RDREGVALUE_1_OFFSET        0xA3
#define S1R72V05_IDE_WRREGVALUE_0_OFFSET        0xA4
#define S1R72V05_IDE_WRREGVALUE_1_OFFSET        0xA5
#define S1R72V05_CHIPCONFIG_SWAPPED_OFFSET      0xB6
#define S1R72V05_CHIPCONFIG_OFFSET              0xB7
#define S1R72V05_CPU_CHGENDIAN_SWAPPED_OFFSET   0xB8
#define S1R72V05_CPU_CHGENDIAN_OFFSET           0xB9

#define S1R72V05_REG_ADDR(reg) \
  (s1r72v05_data.virt_io_addr + S1R72V05_##reg##_OFFSET)

/* Uncomment to turn on output of taskfile register accesses. */
/* #define S1R72V05_TF_DEBUG */

#ifdef S1R72V05_TF_DEBUG
#define S1R72V05_TF_DBG dev_dbg
#else
#define S1R72V05_TF_DBG(...)
#endif

static struct {
	unsigned char __iomem *virt_io_addr;
	ide_hwif_t *hwif;
	unsigned int hwif_idx;
	struct resource *io_res;
	struct resource *irq_res;
	struct platform_device *plat_dev;
} s1r72v05_data;

static inline void s1r72v05_read_tfreg(unsigned long port)
{
	unsigned long timeout;

	iowrite8(S1R72V05_TF_READ | port,
		 S1R72V05_REG_ADDR(IDE_REGADRS));
	timeout = jiffies + 2;
	while ((ioread8(S1R72V05_REG_ADDR(IDE_REGADRS)) &
		S1R72V05_TF_READ) && time_before(jiffies, timeout))
		cpu_relax();
}

static inline void s1r72v05_write_tfreg(unsigned long port)
{
	unsigned long timeout;

	iowrite8(S1R72V05_TF_WRITE | port, S1R72V05_REG_ADDR(IDE_REGADRS));
	timeout = jiffies + 2;
	while ((ioread8(S1R72V05_REG_ADDR(IDE_REGADRS)) &
		S1R72V05_TF_WRITE) && time_before(jiffies, timeout))
		cpu_relax();
}

static int s1r72v05_locate_hwif(void)
{
	int index;

	/* Use the first hwif that is available. */
	for (index = 0; index < MAX_HWIFS; index++) {
		if (!ide_hwifs[index].proc) {
			s1r72v05_data.hwif = &ide_hwifs[index];
			s1r72v05_data.hwif_idx = index;
			dev_dbg(&s1r72v05_data.plat_dev->dev,
				"s1r72v05 using ide hwif %d\n", index);
			return 0;
		}
	}
	return -1;
}

static u8 s1r72v05_inb(unsigned long port)
{
	u8 retval;

	s1r72v05_read_tfreg(port);
	retval = ioread8(S1R72V05_REG_ADDR(IDE_RDREGVALUE_0));

	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_inb(%#lx) = %#x \n", port, retval);

	return retval;
}

static u16 s1r72v05_inw(unsigned long port)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_inw(%#lx) was called\n", port);

	s1r72v05_read_tfreg(port);

	return ioread16(S1R72V05_REG_ADDR(IDE_RDREGVALUE_0));
}

static void s1r72v05_insw(unsigned long port, void *addr, u32 count)
{
	u16 *to = (u16 *) addr;

	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_insw(%#lx, %p, %#x) was called\n",
			port, addr, count);

	for (; count; count--) {
		s1r72v05_read_tfreg(port);
		*(to++) = ioread16(S1R72V05_REG_ADDR(IDE_RDREGVALUE_0));
	}
}

static void s1r72v05_insl(unsigned long port, void *addr, u32 count)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_insl(%#lx, %p, %#x) was called\n",
			port, addr, count);

	printk(KERN_ALERT "%s: s1r72v05_insl UNIMPLEMENTED!\n", DRIVER_NAME);
}

static void s1r72v05_outb(u8 val, unsigned long port)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_outb(%#x, %#lx) was called\n", val, port);

	iowrite16(val, S1R72V05_REG_ADDR(IDE_WRREGVALUE_0));
	s1r72v05_write_tfreg(port);
}

static void s1r72v05_outbsync(ide_drive_t *drive, u8 addr, unsigned long port)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_outbsync(%p, %#x, %#lx) was called\n",
			drive, addr, port);

	iowrite16(addr, S1R72V05_REG_ADDR(IDE_WRREGVALUE_0));
	s1r72v05_write_tfreg(port);
}

static void s1r72v05_outw(u16 val, unsigned long port)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_outw(%#x, %#lx) was called\n", val, port);

	iowrite16(val, S1R72V05_REG_ADDR(IDE_WRREGVALUE_0));
	s1r72v05_write_tfreg(port);
}

static void s1r72v05_outsw(unsigned long port, void *addr, u32 count)
{
	u16 *from = (u16 *) addr;

	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_outsw(%#lx, %p, %#x) was called\n",
			port, addr, count);

	for (; count; count--) {
		iowrite16(*from++, S1R72V05_REG_ADDR(IDE_WRREGVALUE_0));
		s1r72v05_write_tfreg(port);
	}
}

static void s1r72v05_outsl(unsigned long port, void *addr, u32 count)
{
	S1R72V05_TF_DBG(&s1r72v05_data.plat_dev->dev,
			"s1r72v05_outsl(%#lx, %p, %#x) was called\n",
			port, addr, count);

	printk(KERN_ALERT "%s: s1r72v05_outsl UNIMPLEMENTED!\n", DRIVER_NAME);
}

#ifdef DEBUG
static void s1r72v05_register_dump(void)
{
	int i;

	dev_dbg(&s1r72v05_data.plat_dev->dev, "begin register dump\n");
	dev_dbg(&s1r72v05_data.plat_dev->dev, "offset value|offset value\n");
	for (i = 0; i < S1R72V05_IO_EXTENT; i += 2) {
		dev_dbg(&s1r72v05_data.plat_dev->dev,
			"%#.2x: %#.2x | %#.2x: %#.2x\n",
			i, ioread8(s1r72v05_data.virt_io_addr + i),
			i + 1, ioread8(s1r72v05_data.virt_io_addr + i + 1));
	}
	dev_dbg(&s1r72v05_data.plat_dev->dev, "end register dump\n");
}
#endif

static int s1r72v05_ack_intr(struct hwif_s *hwif)
{
	unsigned char ide_status;

	if (hwif != s1r72v05_data.hwif)
		return 0;

	/* Is this not an IDE related interrupt? */
	if ((ioread8(S1R72V05_REG_ADDR(MAININTSTAT)) &
	     S1R72V05_MAININTSTAT_IDE_INTSTAT) == 0) {
		printk(KERN_ALERT "%s: non-IDE interrupt!\n", DRIVER_NAME);
		printk(KERN_ALERT "%s: 0x0:%#.2x 0x1:%#.2x 0x2:%#.2x\n",
		       DRIVER_NAME, ioread8(S1R72V05_REG_ADDR(MAININTSTAT)),
		       ioread8(S1R72V05_REG_ADDR(DEVICEINTSTAT)),
		       ioread8(S1R72V05_REG_ADDR(HOSTINTSTAT)));
		return 0;
	}

	/* Is this the right kind of IDE interrupt? */
	ide_status = ioread8(S1R72V05_REG_ADDR(IDE_INTSTAT));
	if (!(ide_status & S1R72V05_IDE_INTSTAT_DETECTINTRQ)) {
		printk(KERN_ALERT
		       "%s: IDE interrupt for wrong reason: %#x!\n",
		       DRIVER_NAME, ide_status);
		return 0;
	}

	/* Clear IDE interrupt status. */
	iowrite8(S1R72V05_IDE_INTSTAT_DETECTINTRQ,
		 S1R72V05_REG_ADDR(IDE_INTSTAT));

	/* Make sure interrupt status is clear. */
	if (ioread8(S1R72V05_REG_ADDR(MAININTSTAT)) &
	    S1R72V05_MAININTSTAT_IDE_INTSTAT)
		printk(KERN_ALERT
		       "%s: interrupt status did not clear!\n", DRIVER_NAME);

	return 1;
}

static int s1r72v05_probe(struct platform_device *plat_dev)
{
	unsigned int num_sleeps;
	u8 ide_devices[4] = { 0xff, 0xff, 0xff, 0xff };

	s1r72v05_data.plat_dev = plat_dev;
	dev_dbg(&plat_dev->dev, "s1r72v05_probe() was called\n");

	s1r72v05_data.io_res = platform_get_resource(plat_dev, IORESOURCE_MEM,
						     0);
	if (!s1r72v05_data.io_res) {
		printk(KERN_ALERT
		       "%s: could not get IORESOURCE_MEM\n", DRIVER_NAME);
		return -ENODEV;
	}

	s1r72v05_data.irq_res = platform_get_resource(plat_dev,
						      IORESOURCE_IRQ, 0);
	if (!s1r72v05_data.irq_res) {
		printk(KERN_ALERT "%s: could not get IORESOURCE_IRQ\n",
		       DRIVER_NAME);
		return -ENODEV;
	}

	if (gpio_tlmm_config(GPIO_CFG(s1r72v05_data.irq_res->end,
				      0, GPIO_INPUT, GPIO_NO_PULL,
				      GPIO_2MA),
			     GPIO_ENABLE)) {
		printk(KERN_ALERT "%s: could not configure gpio\n",
		       DRIVER_NAME);
		return -EIO;
	}
	gpio_configure(s1r72v05_data.irq_res->end, IRQF_TRIGGER_FALLING);


	if (!request_mem_region(s1r72v05_data.io_res->start,
				S1R72V05_IO_EXTENT, DRIVER_NAME)) {
		printk(KERN_ALERT
		       "%s: could not allocate io region\n", DRIVER_NAME);
		return -EBUSY;
	}

	s1r72v05_data.virt_io_addr =
	    (unsigned char *)ioremap(s1r72v05_data.io_res->start,
				    S1R72V05_IO_EXTENT);
	if (!s1r72v05_data.virt_io_addr) {
		printk(KERN_ALERT
		       "%s: could not ioremap io region\n", DRIVER_NAME);
		goto release_mem_region;
	}

	/* Reset chip. We don't know if the endianness is wrong, so
	   write to both bytes. The other one is the revision register
	   which is read-only (so this won't break anything).
	 */
	iowrite8(0x1, S1R72V05_REG_ADDR(CHIPRESET));
	mdelay(1);
	iowrite8(0x1, S1R72V05_REG_ADDR(CHIPRESET_SWAPPED));
	mdelay(1);

	/* Before we fix the endianness, there's a strange issue where reading
	   the revision register can return 0 unless you read the reset
	   register first.
	 */
	if (ioread8(S1R72V05_REG_ADDR(REVISIONNUM)) != S1R72V05_HW_REVISION) {
		dev_dbg(&plat_dev->dev,
			"revision register not detected, assuming "
			"endianness is reversed\n");
		/* Make sure revision is showing up in its byte-swapped
		   position.
		 */
		if (ioread8(S1R72V05_REG_ADDR(REVISIONNUM_SWAPPED)) !=
		    S1R72V05_HW_REVISION) {
			printk(KERN_ALERT
			       "%s: could not read revision register before "
			       "setting endianness\n", DRIVER_NAME);
			goto release_iomap;
		}

		/* Initialize chip config register and fix endianness. */
		iowrite8(S1R72V05_CHIPCONFIG_VALUE,
			 S1R72V05_REG_ADDR(CHIPCONFIG_SWAPPED));
		ioread8(S1R72V05_REG_ADDR(CPU_CHGENDIAN_SWAPPED));
	} else {
		dev_dbg(&plat_dev->dev,
			"revision register detected, assuming "
			"endianness is correct\n");
		/* Initialize chip config register. */
		iowrite8(S1R72V05_CHIPCONFIG_VALUE,
			 S1R72V05_REG_ADDR(CHIPCONFIG));
	}

	/* Check revision, it should show up correctly now. */
	if (ioread8(S1R72V05_REG_ADDR(REVISIONNUM)) != S1R72V05_HW_REVISION) {
		printk(KERN_ALERT "%s: could not read revision register "
		       "after setting endianness\n", DRIVER_NAME);
		goto release_iomap;
	}

	/* The chip turns on in sleep mode. Go into Active60. */
	iowrite8(S1R72V05_MAININTSTAT_FINISHEDPM,
		 S1R72V05_REG_ADDR(MAININTSTAT));
	iowrite8(S1R72V05_PM_CONTROL_0_GOACTIVE60,
		 S1R72V05_REG_ADDR(PM_CONTROL_0));

	/* Wait for transition to Active60 state. */
	num_sleeps = 0;
	do {
		udelay(S1R72V05_PM_WAIT_USEC);
		num_sleeps += S1R72V05_PM_WAIT_USEC;
	} while (!(ioread8(S1R72V05_REG_ADDR(MAININTSTAT)) &
		   S1R72V05_MAININTSTAT_FINISHEDPM) &&
		 num_sleeps < S1R72V05_MAX_PM_WAIT_USEC);
	if (!(ioread8(S1R72V05_REG_ADDR(MAININTSTAT)) &
	      S1R72V05_MAININTSTAT_FINISHEDPM)) {
		printk(KERN_ALERT "%s: could not switch to Active60 state\n",
		       DRIVER_NAME);
		goto release_iomap;
	}

	/* Turn on IDE interface in the chip. */
	iowrite8(S1R72V05_IDE_CONFIG_1_SWAP | S1R72V05_IDE_CONFIG_1_ACTIVEIDE,
		 S1R72V05_REG_ADDR(IDE_CONFIG_1));

	/* "Initialize the IDE circuit." */
	iowrite8(S1R72V05_IDE_CONTROL_IDE_CLR, S1R72V05_REG_ADDR(IDE_CONTROL));

	/* Set the assert pulse and negate pulse width to the max for now. */
	iowrite8(S1R72V05_IDE_RMOD_MAX_TIMING, S1R72V05_REG_ADDR(IDE_RMOD));

	/* Reset the IDE bus and set to PIO mode. */
	iowrite8(S1R72V05_IDE_CONFIG_0_IDE_BUSRESET,
		 S1R72V05_REG_ADDR(IDE_CONFIG_0));

	/* Clear any pending interrupts. */
	iowrite8(S1R72V05_MAININTSTAT_ALLINTS, S1R72V05_REG_ADDR(MAININTSTAT));
	iowrite8(S1R72V05_IDE_INTSTAT_ALLINTS, S1R72V05_REG_ADDR(IDE_INTSTAT));

	/* Enable IDE interrupts. */
	iowrite8(S1R72V05_MAININTENB_ENIDE_INTSTAT,
		 S1R72V05_REG_ADDR(MAININTENB));
	iowrite8(S1R72V05_IDE_INTENB_ENDETECTINTRQ,
		 S1R72V05_REG_ADDR(IDE_INTENB));

	dev_dbg(&plat_dev->dev, "chip initialization complete\n");

	/*
	 * Initialize linux IDE data structures and register interface.
	 */

	if (s1r72v05_locate_hwif() < 0) {
		printk(KERN_ALERT "%s: could not find an ide hwif to use\n",
		       DRIVER_NAME);
		goto release_iomap;
	}

	/* Set up hwif_s structure. */
	s1r72v05_data.hwif->io_ports[IDE_DATA_OFFSET] = S1R72V05_DATA_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_ERROR_OFFSET] = S1R72V05_ERROR_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_NSECTOR_OFFSET] =
	    S1R72V05_SECTOR_CNT_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_SECTOR_OFFSET] =
	    S1R72V05_LBA_LOW_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_LCYL_OFFSET] =
		S1R72V05_LBA_MID_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_HCYL_OFFSET] =
	    S1R72V05_LBA_HIGH_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_SELECT_OFFSET] =
	    S1R72V05_DEVICE_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_STATUS_OFFSET] =
	    S1R72V05_STATUS_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_CONTROL_OFFSET] =
	    S1R72V05_DEVICE_CONTROL_TF_REG;
	s1r72v05_data.hwif->io_ports[IDE_IRQ_OFFSET] = S1R72V05_UNSUPPORTED;
	s1r72v05_data.hwif->irq = s1r72v05_data.irq_res->start;
	s1r72v05_data.hwif->noprobe = 0;
	s1r72v05_data.hwif->OUTB = s1r72v05_outb;
	s1r72v05_data.hwif->OUTBSYNC = s1r72v05_outbsync;
	s1r72v05_data.hwif->OUTW = s1r72v05_outw;
	s1r72v05_data.hwif->OUTSW = s1r72v05_outsw;
	s1r72v05_data.hwif->OUTSL = s1r72v05_outsl;
	s1r72v05_data.hwif->INB = s1r72v05_inb;
	s1r72v05_data.hwif->INW = s1r72v05_inw;
	s1r72v05_data.hwif->INSW = s1r72v05_insw;
	s1r72v05_data.hwif->INSL = s1r72v05_insl;

	/* Set up hw_regs_t structure. */
	memcpy(s1r72v05_data.hwif->io_ports, s1r72v05_data.hwif->io_ports,
	       sizeof(s1r72v05_data.hwif->io_ports));
	s1r72v05_data.hwif->ack_intr = s1r72v05_ack_intr;

	ide_devices[0] = s1r72v05_data.hwif_idx;
	if (ide_device_add(ide_devices, NULL)) {
		printk(KERN_ALERT "%s: ide_device_add() failed\n",
		       DRIVER_NAME);
		/* Disable interrupts in the interface chip. */
		iowrite8(0, S1R72V05_REG_ADDR(MAININTENB));
		goto release_iomap;
	}

	dev_dbg(&plat_dev->dev,
		"probe complete, interface registered\n");

	return 0;

release_iomap:
	iounmap((void *)s1r72v05_data.virt_io_addr);

release_mem_region:
	release_mem_region(s1r72v05_data.io_res->start, S1R72V05_IO_EXTENT);
	return -ENOMEM;
}

static int s1r72v05_remove(struct platform_device *plat_dev)
{
	ide_unregister(s1r72v05_data.hwif_idx, 0, 0);
	iounmap((void *)s1r72v05_data.virt_io_addr);
	release_mem_region(s1r72v05_data.io_res->start, S1R72V05_IO_EXTENT);

	return 0;
}

static struct platform_driver s1r72v05_driver = {
	.probe = s1r72v05_probe,
	.remove = s1r72v05_remove,
	.shutdown = NULL,
	.suspend = NULL,
	.suspend_late = NULL,
	.resume_early = NULL,
	.resume = NULL,
	.driver = {
		   .name = DRIVER_NAME,
		   .owner = THIS_MODULE,
	},
};

static int __init s1r72v05_init(void)
{
	return platform_driver_register(&s1r72v05_driver);
}

static void __exit s1r72v05_exit(void)
{
	platform_driver_unregister(&s1r72v05_driver);
}

module_init(s1r72v05_init);
module_exit(s1r72v05_exit);

MODULE_DESCRIPTION("Epson S1R72V05 PATA interface driver");
MODULE_VERSION("0.1");
MODULE_LICENSE("GPL v2");
