/* drivers/input/misc/minitrackball.c
 *
 * Input driver for Panasonic GPIO-based "Jog Ball" (aka mini-trackball)
 * device (part # EVQWJN)
 *
 * Copyright (c) 2008 QUALCOMM USA, INC.
 * 
 * All source code in this file is licensed under the following license
 * except where indicated.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/input.h>

#include <asm/gpio.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>

#define DEFAULT_DELTA_X 5
#define DEFAULT_DELTA_Y 5
#define SELECT_INPUT_BTN BTN_LEFT	/* This is chosen to match with
					 * the mousedev driver */

#define DRIVE_GPIO   8
#define PLUS_X_GPIO  5
#define PLUS_X_IRQ   MSM_GPIO_TO_INT(PLUS_X_GPIO)
#define NEG_X_GPIO   6
#define NEG_X_IRQ    MSM_GPIO_TO_INT(NEG_X_GPIO)
#define PLUS_Y_GPIO  2
#define PLUS_Y_IRQ   MSM_GPIO_TO_INT(PLUS_Y_GPIO)
#define NEG_Y_GPIO   3
#define NEG_Y_IRQ    MSM_GPIO_TO_INT(NEG_Y_GPIO)
#define SELECT_GPIO  4
#define SELECT_IRQ   MSM_GPIO_TO_INT(SELECT_GPIO)

static int delta_x = DEFAULT_DELTA_X;
static int delta_y = DEFAULT_DELTA_Y;
module_param(delta_x, int, S_IRUGO);
module_param(delta_y, int, S_IRUGO);

static struct input_dev *mtb_dev;

#define SELECT_LOW  0
#define SELECT_HIGH (!SELECT_LOW)
static u8 select_state = SELECT_HIGH;

static irqreturn_t plus_x_isr(int irq, void *dummy)
{
	input_report_rel(mtb_dev, REL_X, delta_x);
	input_sync(mtb_dev);

	return IRQ_HANDLED;
}

static irqreturn_t neg_x_isr(int irq, void *dummy)
{
	input_report_rel(mtb_dev, REL_X, -delta_x);
	input_sync(mtb_dev);

	return IRQ_HANDLED;
}

static irqreturn_t plus_y_isr(int irq, void *dummy)
{
	input_report_rel(mtb_dev, REL_Y, delta_y);
	input_sync(mtb_dev);

	return IRQ_HANDLED;
}

static irqreturn_t neg_y_isr(int irq, void *dummy)
{
	input_report_rel(mtb_dev, REL_Y, -delta_y);
	input_sync(mtb_dev);

	return IRQ_HANDLED;
}

static irqreturn_t select_isr(int irq, void *dummy)
{
	select_state = !select_state;
	/* the GPIO is low when the button is pushed */
	input_report_key(mtb_dev,
			 SELECT_INPUT_BTN,
			 (select_state == SELECT_LOW));
	input_sync(mtb_dev);

	return IRQ_HANDLED;
}

static int mtb_open(struct input_dev *dev)
{
	int err;

	gpio_tlmm_config(GPIO_CFG(DRIVE_GPIO,
				  0, /* function */
				  GPIO_OUTPUT,
				  GPIO_NO_PULL,
				  GPIO_16MA),
			 GPIO_ENABLE);
	gpio_tlmm_config(GPIO_CFG(PLUS_X_GPIO,
				  0, /* function */
				  GPIO_INPUT,
				  GPIO_NO_PULL,
				  GPIO_2MA),
			 GPIO_ENABLE);
	gpio_tlmm_config(GPIO_CFG(NEG_X_GPIO,
				  0, /* function */
				  GPIO_INPUT,
				  GPIO_NO_PULL,
				  GPIO_2MA),
			 GPIO_ENABLE);
	gpio_tlmm_config(GPIO_CFG(PLUS_Y_GPIO,
				  0, /* function */
				  GPIO_INPUT,
				  GPIO_NO_PULL,
				  GPIO_2MA),
			 GPIO_ENABLE);
	gpio_tlmm_config(GPIO_CFG(NEG_Y_GPIO,
				  0, /* function */
				  GPIO_INPUT,
				  GPIO_NO_PULL,
				  GPIO_2MA),
			 GPIO_ENABLE);

	err = request_irq(PLUS_X_IRQ,
			  &plus_x_isr,
			  IRQF_TRIGGER_FALLING, "MTB Plus X", NULL);
	if (err)
		goto cleanup5;

	err = request_irq(NEG_X_IRQ,
			  &neg_x_isr, IRQF_TRIGGER_FALLING, "MTB Neg X", NULL);
	if (err)
		goto cleanup4;

	err = request_irq(PLUS_Y_IRQ,
			  &plus_y_isr,
			  IRQF_TRIGGER_FALLING, "MTB Plus Y", NULL);
	if (err)
		goto cleanup3;

	err = request_irq(NEG_Y_IRQ,
			  &neg_y_isr, IRQF_TRIGGER_FALLING, "MTB Neg Y", NULL);
	if (err)
		goto cleanup2;

	err = request_irq(SELECT_IRQ,
			  &select_isr,
			  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			  "MTB Select", NULL);
	if (err)
		goto cleanup1;

	return 0;

cleanup1:
	free_irq(NEG_Y_IRQ, NULL);
cleanup2:
	free_irq(PLUS_Y_IRQ, NULL);
cleanup3:
	free_irq(NEG_X_IRQ, NULL);
cleanup4:
	free_irq(PLUS_X_IRQ, NULL);
cleanup5:
	gpio_direction_input(DRIVE_GPIO);
	return err;
}

static void mtb_close(struct input_dev *dev)
{
	gpio_direction_input(DRIVE_GPIO);

	free_irq(PLUS_X_IRQ, NULL);
	free_irq(NEG_X_IRQ, NULL);
	free_irq(PLUS_Y_IRQ, NULL);
	free_irq(NEG_Y_IRQ, NULL);
	free_irq(SELECT_IRQ, NULL);
}

static __init int mtb_init(void)
{
	int err;

	mtb_dev = input_allocate_device();
	if (!mtb_dev)
		return -ENOMEM;

	mtb_dev->name = "Mini Trackball Device";
	mtb_dev->phys = "/dev/mtb";
	mtb_dev->open = mtb_open;
	mtb_dev->close = mtb_close;
	mtb_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_REL);
	set_bit(SELECT_INPUT_BTN, mtb_dev->keybit);
	mtb_dev->relbit[0] = BIT(REL_X) | BIT(REL_Y);

	err = input_register_device(mtb_dev);
	if (err) {
		input_free_device(mtb_dev);
		return err;
	}

	return 0;
}

/**
*  @brief Exits mtb driver.
*
*  Deregisters input device.
*/
static void __exit mtb_exit(void)
{
	input_unregister_device(mtb_dev);
	input_free_device(mtb_dev);
}

module_init(mtb_init);
module_exit(mtb_exit);

MODULE_DESCRIPTION("Mini-trackball driver");
MODULE_VERSION("0.2");
MODULE_LICENSE("GPL v2");
