/* drivers/titan/audioFIFO.c
 *
 * Copyright (C) 2004,2008 QUALCOMM USA, INC.
 * 
 * All source code in this file is licensed under the following license
 * except where indicated.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 */

#include <linux/completion.h>	/* struct completion, complete()*/
#include <linux/io.h>		/* readl(), writel()*/
#include "raptor2.h"

#include "audioFIFO.h"
#include "platform.h"

#define INIT_CHECK	0

#ifndef MIN
#define MIN(a, b)	((a > b) ? b : a)
#endif

static struct tAudioFIFO gAudioFIFO = {
	/* Control */
	.bInit = 0,
	.pBuf = NULL,
	.nFirst = 0,
	.nLast = 0,
	.len = 0,
	.inCount = 0,
	.outCount = 0,
	.outChunk = 0,
	/* Status */
	.underruns = 0,
	/* Intr Handler */
	.pBaseAddr = NULL,
	.irqOffset = 0,
	.callBack = 0,
	.pCBData = NULL
};

/*===========================================================================
FUNCTION  GetAudioFIFO(void)

DESCRIPTION Return pointer to the audio FIFO.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
tAudioFIFO *GetAudioFIFO(void)
{
	return &gAudioFIFO;
}

/*===========================================================================
FUNCTION int InitAudioFIFO(tAudioFIFO* pCtrl, int* pBuf, int length)

DESCRIPTION Initialize the audion fifo, called before the audion fifo
  can be used.

DEPENDENCIES
	 should be called before AudioFIFO in use!!!

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
int InitAudioFIFO(tAudioFIFO *pCtrl, int *pBuf, int length)
{
	if (length > AUDIO_FIFO_MAX_LEN)
		return -1;

	pCtrl->bInit = 1;
	pCtrl->pBuf = pBuf;
	pCtrl->nFirst = 0;
	pCtrl->nLast = 0;
	pCtrl->len = length;
	pCtrl->inCount = 0;
	pCtrl->outCount = 0;
	/* pCtrl->outChunk = ???; */

	/* reset status */
	pCtrl->underruns = 0;

	/* Init Handler separately
	*pCtrl->pBaseAddr = ???;
	*pCtrl->irqOffset = 0;
	*pCtrl->callBack = ???;
	*pCtrl->pCBack = ???;
	*/
	return 0;
}

/*===========================================================================
FUNCTION int InitAudioFIFOHandler(tAudioFIFO* pCtrl, int *pBaseAddr, int irqNr,
				tAudioFIFOCallBack callBack, void *pCBData )

DESCRIPTION Initialize the audio fifo interrupt handler.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
int InitAudioFIFOHandler(tAudioFIFO *pCtrl, int *pBaseAddr, int irqNr,
			 tAudioFIFOCallBack callBack, void *pCBData)
{
	if ((IRQ_AUDIO_OUT0 != irqNr) && (IRQ_AUDIO_OUT1 != irqNr))
		return -1;

	pCtrl->pBaseAddr = pBaseAddr;
	pCtrl->irqOffset = (irqNr == IRQ_AUDIO_OUT1) ? AUDIO_IRQ_SIZE : 0;
	pCtrl->callBack = callBack;
	pCtrl->pCBData = pCBData;

	return 0;
}

/*===========================================================================
FUNCTION void RestartAudioFIFO(tAudioFIFO* pCtrl)

DESCRIPTION Restart the audio fifo.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
void RestartAudioFIFO(tAudioFIFO *pCtrl)
{
	pCtrl->nFirst = 0;
	pCtrl->nLast = 0;
}

/*===========================================================================
FUNCTION void ClearAudioBuf(tAudioFIFO* pCtrl)

DESCRIPTION Clear out the audio buffer.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
void ClearAudioBuf(tAudioFIFO *pCtrl)
{
#if (INIT_CHECK == 1)
	if (!pCtrl->bInit)
		return;
#endif

	if (pCtrl->pBuf)
		memset(pCtrl->pBuf, 0, (pCtrl->len * sizeof(int)));
}

/*===========================================================================
FUNCTION int RequestAudioFIFOWriteBuffer( tAudioFIFO *pCtrl, int **ppBuf,
						int reqLen )

DESCRIPTION Get audion fifo space and intialize it.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
int RequestAudioFIFOWriteBuffer(tAudioFIFO *pCtrl, int **ppBuf, int reqLen)
{
	int nTst;
	int space;

#if (INIT_CHECK == 1)
	if (!pCtrl->bInit) {
		*ppBuf = NULL;
		return -1;
	}
#endif

	if (pCtrl->nLast == pCtrl->len)
		pCtrl->nLast = 0;

	*ppBuf = &pCtrl->pBuf[pCtrl->nLast];

	space = GetAudioFIFOSpace(pCtrl);
	reqLen = MIN(reqLen, space);
	nTst = pCtrl->nLast + reqLen;
	if (nTst >= pCtrl->len)
		reqLen = pCtrl->len - pCtrl->nLast;

	return reqLen;
}

/*===========================================================================
FUNCTION int DeliverAudioFIFOWriteBuffer( tAudioFIFO *pCtrl, int *pBuf,
						 int len )

DESCRIPTION Return an audio FIFO write buffer to the caller.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
int DeliverAudioFIFOWriteBuffer(tAudioFIFO *pCtrl, int *pBuf, int len)
{
	int nTst;

#if (INIT_CHECK == 1)
	if (!pCtrl->bInit)
		return -1;

#endif

	if ((pBuf != &pCtrl->pBuf[pCtrl->nLast]) ||
	    (len > GetAudioFIFOSpace(pCtrl)))
		return -2;

	nTst = pCtrl->nLast + len;
	if (nTst > pCtrl->len)
		return -3;
	else if (nTst == pCtrl->len)
		pCtrl->nLast = 0;
	else
		pCtrl->nLast = nTst;

	pCtrl->inCount += len;

	return len;
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*Audio Interrupt handler
*/
/*===========================================================================
FUNCTION irqreturn_t HandleAudioFIFOInterrupt ( int irqnr, void *devid,
						struct pt_regs *regs )

DESCRIPTION  Handle the interrupt for the audio.  Tells when the fifo will
  underrun and needs more data, errors, etc.

DEPENDENCIES

RETURN VALUE

SIDE EFFECTS

===========================================================================*/
irqreturn_t HandleAudioFIFOInterrupt(int irqnr, void *devid,
				     struct pt_regs *regs)
{
	tAudioFIFO *pCtrl = (tAudioFIFO *) devid;
	unsigned int portAddr;
	int status;
	int outCnt, inCnt;
	int nTst;

	portAddr = (unsigned int)pCtrl->pBaseAddr + pCtrl->irqOffset;
	status = READL(portAddr + AUDIO_IRQ_STAT);
	WRITEL(status, portAddr + AUDIO_IRQ_CLEAR);

	/* update FIFO outCount */
	outCnt = pCtrl->outCount + pCtrl->outChunk;
	inCnt = pCtrl->inCount;
	/* ...if we underrun, outCount cannot incr past inCount, but keep nFirst
	 indexing the right location */
	if (inCnt < outCnt) {
		pCtrl->outCount = inCnt;
		pCtrl->underruns++;
	} else {
		pCtrl->outCount = outCnt;
	}

	/* update FIFO nFirst index */
	nTst = pCtrl->nFirst + pCtrl->outChunk;
	if (nTst >= pCtrl->len)
		nTst -= pCtrl->len;

	pCtrl->nFirst = nTst;

	if (pCtrl->callBack)
		(pCtrl->callBack) (pCtrl->pCBData);


	return IRQ_HANDLED;
}

