/*===========================================================================

		D S P  ( P C M   A U D I O)   D R I V E R

DESCRIPTION
   This file implements the dsp driver, which interfaces LINUX user
   tasks to PCM audio

EXTERNALIZED FUNCTIONS
   dsp_driver_init    - should be called to initialize this module
   dsp_driver_cleanup - should be called to terminate  this module

   LINUX driver supported functions:
      open(), close(), write(), ioctl()
   All other LINUX driver functions are not supported and will
      return an error code if attempted

INITIALIZATION AND SEQUENCING REQUIREMENTS
   It is assumed that LINUX kernel is up and running and its
   services are available.

   This driver assumes only one user (i.e., multiple opens will fail).
   Otherwise, you can get the case where output samples are mixed
   indeterminately.

Copyright (c) 2005,2007-2008 QUALCOMM USA, INC.

All source code in this file is licensed under the following license
except where indicated.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
version 2 as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, you can find it at http://www.fsf.org
===========================================================================*/

/*===========================================================================

			INCLUDE FILES FOR MODULE

===========================================================================*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/soundcard.h>
#include <linux/cdev.h>
#include <linux/rwsem.h>
#include <linux/uaccess.h>
#include <asm/semaphore.h>
#include "audiohw.h"

#include "audioSS.h"

/*===========================================================================

			DEFINITIONS AND DECLARATIONS FOR MODULE

===========================================================================*/

/*===========================================================================
  Constant / Define Declarations
===========================================================================*/

/*---------------------------------------------------------------------------
*---------------------------------------------------------------------------*/
#define LINUX_SUCCESS 0

/*---------------------------------------------------------------------------
* Default audio settings
*---------------------------------------------------------------------------*/
/* KAS redflag suggested change */
/*#define AUDIO_FRAMES_PER_KSEC 5000         */   /* 31000 */
#define AUDIO_FRAMES_PER_KSEC (31000*2)		/* 31000 */
#define AUDIO_SAMPLES_PER_SEC 44100
#define AUDIO_IS_STEREO       TRUE
#define AUDIO_BUFFER_LEN      64 * 1024 * sizeof(int)

/*---------------------------------------------------------------------------
* Declare the standard set of module parameters along with the
* command line insmod parameters
*---------------------------------------------------------------------------*/
#define DSP_DRIVER_VERSION "01:00.00"

MODULE_LICENSE("Proprietary");
MODULE_VERSION(DSP_DRIVER_VERSION);
MODULE_AUTHOR("QUALCOMM-QCT-Enablement");
MODULE_DESCRIPTION("Qualcomm DSP device driver, version=" DSP_DRIVER_VERSION);

/*---------------------------------------------------------------------------
* Major and minor device numbers for this device.
* - must match the corresponding entries in
*   platform/linux_apps/rootfs/dev.txt
*---------------------------------------------------------------------------*/
#define DSP_MAJOR               14
#define DSP_MINOR               3

/*---------------------------------------------------------------------------
* Constants for use in the file_operations structure
*---------------------------------------------------------------------------*/
#define DSP_NOTIMPLEMENTED      NULL
#define DSP_OWNER               THIS_MODULE

/* The name applications should open to make use of the driver               */
/* Note that this must match the entry in platform/linux_apps/rootfs/dev.txt */
#define DSP_DEVICENAME          "/dev/dsp"

/*---------------------------------------------------------------------------
* This is the name assigned to this driver that will appear in the
* /proc/qcom dir.
*---------------------------------------------------------------------------*/
#define DSP_DRIVER_DEVICE_NAME "dsp"

/*---------------------------------------------------------------------------
* Used to stringify an identifier (ie, it takes an identifier and creates
* a string out of it).
*---------------------------------------------------------------------------*/
#define DSP_DRIVER_STRINGIFY(identifier_to_stringify) #identifier_to_stringify

/*===========================================================================
  Type Declarations
===========================================================================*/

typedef enum {
	AUDIOSS_NOT_CONFIGURED = 1,
	AUDIOSS_CHANGES_ALLOWED = 2,
	AUDIOSS_INIT = AUDIOSS_NOT_CONFIGURED | AUDIOSS_CHANGES_ALLOWED
} audio_state_t;

#define RIFF_CHUNK_ID 0x46464952	/* RIFF (Big Endian) */
#define WAVE_FORMAT   0x45564157	/* WAVE (Big Endian) */
#pragma pack(2)
typedef struct WavHeaderStruct {
	unsigned long ChunkID;
	unsigned long ChunkSize;
	unsigned long Format;
	unsigned long SubChunk1ID;
	unsigned long SubChunk1Size;
	unsigned short AudioFormat;
	unsigned short NumChannels;
	unsigned long SampleRate;
	unsigned long ByteRate;
	unsigned short BlockAlign;
	unsigned short BitsPerSample;
	unsigned long SubChunk2ID;
	unsigned long SubChunk2Size;
} WaveHeaderType;
#pragma pack()

/*---------------------------------------------------------------------------
* These are the data members that are needed by each minor number
*
* allocated          - TRUE if minor is open (allocated for use)
* audio_state        - state flags reflect capabilities (AUDIOSS_...)
* need_module_put    - count of module_put's that have been postponed until
			FIFO empty
* rw_alloc           - rw semaphore protecting allocated value
* usage_mutex        - held while the minor number is doing an open,
			 ioctl or close
* write_mutex        - held while the minor number is doing a write
* pAudioSS           - pointer to Audio SubSystem object
*---------------------------------------------------------------------------*/
typedef struct {
	boolean allocated;
	int audio_state;
	int need_module_put;
	struct rw_semaphore rw_alloc;
	struct mutex usage_mutex;
	struct mutex write_mutex;
	tAudioSS *pAudioSS;
} dsp_driver_info_type;

/*===========================================================================
  Local Object Definitions
===========================================================================*/

/*---------------------------------------------------------------------------
* Set to TRUE after the driver has been successfully initialized and
* before its been terminated.
*---------------------------------------------------------------------------*/
static boolean dsp_driver_initialized = FALSE;

/*---------------------------------------------------------------------------
* These semaphores are used to protect critical code sections from
* multiple threads accessing them.
*---------------------------------------------------------------------------*/
struct mutex dsp_driver_open_mutex;

/*---------------------------------------------------------------------------
* This will hold a pointer to the /proc/qcom dir entry
*---------------------------------------------------------------------------*/
static struct proc_dir_entry *dsp_driver_procfs_qcom_dir;

/*---------------------------------------------------------------------------
* This data structure contains the information associated with the driver
*---------------------------------------------------------------------------*/
static dsp_driver_info_type *dsp_driver_info;

/*---------------------------------------------------------------------------
* The AudioFIFO Buffer length
*---------------------------------------------------------------------------*/
int audioFIFOBufferLen ;
module_param(audioFIFOBufferLen, int, S_IRUSR | S_IRGRP | S_IROTH);
MODULE_PARM_DESC(audioFIFOBufferLen, "Audio FIFO buffer length (bytes)");

/*---------------------------------------------------------------------------
* Default Frames per Kilosecond setting (1000 * frames/second)
*---------------------------------------------------------------------------*/
int audioFramesPerKSec;
module_param(audioFramesPerKSec, int,
	     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
MODULE_PARM_DESC(audioFramesPerKSec,
		 "Audio Frames Per Kilosecond (1000 * frames/sec)");

/*===========================================================================
  Forward Declarations
===========================================================================*/
static struct file_operations dsp_fops;

static struct cdev dsp_cdev;

/*===========================================================================
			Function Definitions
===========================================================================*/
/*===========================================================================
FUNCTION dsp_driver_read_proc_dsp

DESCRIPTION
   This function will be called when the virtual file /proc/qcom/dsp is read.
   It will dump debug information on this driver.

   This is an unsupported interface, which means that it is not documented,
   and bugs or issues with its operation will not be considered a priority.

DEPENDENCIES
   This callback function must have been registered with the kernel
   and associated with the "dsp" file, so that attempts to dump
   the dsp file will call this function to retrieve the text from it.

RETURN VALUE
   The number of bytes of debug text data inserted into the buffer

SIDE EFFECTS
   None
===========================================================================*/
static int dsp_driver_read_proc_dsp
    (char *buffer,
     char **start, off_t offset, int count, int *eof, void *data) {
	return 0;
}				/* dsp_driver_read_proc_dsp() */

/*===========================================================================
FUNCTION dsp_driver_open

DESCRIPTION
   This is the LINUX open device driver call, conforming to the standard
   LINUX device driver interface.

   This open call implements both static and dynamic channel allocation.
   If the minor number is zero, then dynamic allocation is assumed
   and one of the remaining channels for this major number will be allocated.
   If the minor number is non zero, then static allocation is assumed
   and the minor number passed into this open call will attempt to be
   opened.

DEPENDENCIES
   The driver should have been previously initialized

RETURN VALUE
    LINUX_SUCCESS           if open succeeded
   -EINVAL      illegal parameter or state detected
   -ENODEV      channel not available
   -ERESTARTSYS wait interrupted, retry

SIDE EFFECTS
   None
===========================================================================*/
static int dsp_driver_open
		(struct inode *inode, struct file *filp) {
	int open_status = LINUX_SUCCESS;
	int minor_number = MINOR(inode->i_rdev);

	printk(KERN_INFO "dsp-open flags:%x\n", filp->f_flags & O_NONBLOCK);
	/*---------------------------------------------------------------------
	Do these basic sanity checks to make sure we don't crash the driver
	and are in the correct state to provide this service
	---------------------------------------------------------------------*/
	if (!dsp_driver_initialized) {
		open_status = -EINVAL;
		goto bail_open;
	}

	dsp_driver_info->pAudioSS->bNonBlock = FALSE;

	/*----------------------------------------------------------------------
	Increment the usage count so that driver isn't unloaded until open's
	all have been closed.
	----------------------------------------------------------------------*/
	if (try_module_get(THIS_MODULE)) {
		/* Make sure safe to open/allocate minor */
		/* open_status =
		down_write_interruptible( &dsp_driver_info->rw_alloc ); */
		down_write(&dsp_driver_info->rw_alloc);
		if (LINUX_SUCCESS == open_status) {
			/* only one open at a time please */
			open_status =
			    mutex_lock_interruptible(&dsp_driver_open_mutex);
			if (LINUX_SUCCESS == open_status) {
				if (dsp_driver_info->allocated) {
					open_status = -EBUSY;
					/* release locks */
					mutex_unlock(&dsp_driver_open_mutex);
					up_write(&dsp_driver_info->rw_alloc);
				}
			} else {
				up_write(&dsp_driver_info->rw_alloc);
			}
		}

		if (LINUX_SUCCESS == open_status) {
			if (filp->f_mode & FMODE_WRITE) {
				open_status = LINUX_SUCCESS;

		/*------------------------------------------------------------
		 Initialize minor
		------------------------------------------------------------*/
				filp->private_data = (void *)minor_number;

			/* If module param not overridden, set default here */
				if (0 == audioFramesPerKSec)
					audioFramesPerKSec =
					    AUDIO_FRAMES_PER_KSEC;

		/* Only reset to defaults if not configured flag is set!!!
		 (configuration protected by rw_alloc under write lock) */
				if (dsp_driver_info->
				    audio_state & AUDIOSS_NOT_CONFIGURED) {
				dsp_driver_info->pAudioSS->samplesPerSec =
							AUDIO_SAMPLES_PER_SEC;
				dsp_driver_info->pAudioSS->bStereo =
							AUDIO_IS_STEREO;
				dsp_driver_info->pAudioSS->framesPerKSec =
							 audioFramesPerKSec;
				dsp_driver_info->pAudioSS->bNonBlock =
					    (filp->
					     f_flags & O_NONBLOCK) ? TRUE :
					    FALSE;
				printk(KERN_INFO "bNonBlock %d flags:%x\n",
					       dsp_driver_info->pAudioSS->
					       bNonBlock, filp->f_flags);
				} else {
				/* check if blocking option has changed */
					boolean bNonBlock =
					    (filp->
					     f_flags & O_NONBLOCK) ? TRUE :
					    FALSE;
					if (bNonBlock !=
					    dsp_driver_info->pAudioSS->
					    bNonBlock) {
						dsp_driver_info->pAudioSS->
						    bNonBlock = bNonBlock;
						dsp_driver_info->audio_state |=
						    AUDIOSS_NOT_CONFIGURED;
					}
				}

				/* ConfigAudioSS delayed until first write
				ConfigAudioSS( dsp_driver_info->pAudioSS ); */
				/* dsp_driver_info->audio_state &=
					~AUDIOSS_NOT_CONFIGURED; */

				dsp_driver_info->allocated = TRUE;

	/* If module was previously waiting for put opportunity, then safe
		 to module_put now (this instance takes over previous setup).
		  need_module_put protected by rw_alloc in write lock.*/
				if (dsp_driver_info->need_module_put) {
		/* no need to TeardownAudioSS; leave configured */
					module_put(THIS_MODULE);
					dsp_driver_info->need_module_put--;
				}
			} else {
				/* We do not support read( ) at this time */
				/* open_status = -EBUSY; */
				open_status = -ENODEV;
			}

			/* Done. release locks */
			mutex_unlock(&dsp_driver_open_mutex);
			up_write(&dsp_driver_info->rw_alloc);
		}		/* we have locks (open, rw_alloc) */

		/*--------------------------------------------------------------
		 If we had an error, then decrement the module count since open
		 didn't complete successfully (ie, we haven't really opened
		 the driver).
		--------------------------------------------------------------*/
		if (open_status < LINUX_SUCCESS)
			module_put(THIS_MODULE);

	} else {			/* try_module_get failed */
		open_status = -EFAULT;
	}

bail_open:
	return open_status;
}				/* dsp_driver_open() */

/*===========================================================================
FUNCTION dsp_driver_close

DESCRIPTION
   Closes the driver and cleans up associated data structures

DEPENDENCIES
   None

RETURN VALUE
    LINUX_SUCCESS           successful close
   -EINVAL      illegal parameter
   -ERESTARTSYS blocking wait interrupted, retry
   -ENODEV      device wasn't previously opened

SIDE EFFECTS
   None
===========================================================================*/
static int dsp_driver_close
		(struct inode *inode, struct file *filp) {
	int close_status = LINUX_SUCCESS;
	int remaining;

	/*------------------------------------------------------------------
	 Do these basic sanity checks to make sure we don't crash the driver
	 and are in the correct state to provide this service.
	------------------------------------------------------------------*/
	if (!dsp_driver_initialized) {
		close_status = -EINVAL;
		goto bail_close;
	}
	/* Make sure all ioctls and writes are complete (not pending)
	 and check open/allocated status */
	/* close_status =
	down_write_interruptible( &dsp_driver_info->rw_alloc ); */
	down_write(&dsp_driver_info->rw_alloc);
	if (LINUX_SUCCESS == close_status) {
		if (!dsp_driver_info->allocated) {
			close_status = -EINVAL;
			up_write(&dsp_driver_info->rw_alloc);
			goto bail_close;
		}

		dsp_driver_info->allocated = FALSE;

		(void)FlushAudioSS(dsp_driver_info->pAudioSS);
		TeardownAudioSS(dsp_driver_info->pAudioSS,
				FALSE);
		module_put(THIS_MODULE);
		if (0 < dsp_driver_info->need_module_put)
			module_put(THIS_MODULE);

		dsp_driver_info->audio_state = AUDIOSS_INIT;

		up_write(&dsp_driver_info->rw_alloc);
	}

bail_close:
	return (close_status);
}				/* dsp_driver_close() */

/*===========================================================================
FUNCTION dsp_driver_ioctl

DESCRIPTION
   Performs a control operation on the device

DEPENDENCIES
   The driver should have been previously initialized

RETURN VALUE
    LINUX_SUCCESS           successful close
   -ENIVAL      the user buffer is too small to receive message
   -ERESTARTSYS blocking operation interrupted, retry
   -EFAULT      an error condition was encountered that caused the
		operation to fail

SIDE EFFECTS
   None
===========================================================================*/
static int dsp_driver_ioctl
    (struct inode *inode,
     struct file *filp, unsigned int cmd, unsigned long arg) {
	int ioctl_status = LINUX_SUCCESS;
	int return_arg;

	/*------------------------------------------------------------------
	 Do these basic sanity checks to make sure we don't crash the driver
	 and are in the correct state to provide this service.
	-------------------------------------------------------------------*/
	if (!dsp_driver_initialized) {
		ioctl_status = -EINVAL;
		goto bail_ioctl;
	}
	/* State of allocated must be held during ioctl call.  rw_alloc
	 locks it, but in a read only state so that multiple ioctl's
	 or write's don't have to wait on each other. */
	/*ioctl_status =
	 down_read_interruptible( &dsp_driver_info->rw_alloc ); */
	down_read(&dsp_driver_info->rw_alloc);
	if (LINUX_SUCCESS == ioctl_status) {
		if (!dsp_driver_info->allocated) {
			ioctl_status = -EINVAL;
			up_read(&dsp_driver_info->rw_alloc);
			goto bail_ioctl;
		}
	}
	/*----------------------------------------------------------------
	 Only one thread should be able to use each minor number at a time.
	-----------------------------------------------------------------*/
	ioctl_status = mutex_lock_interruptible(&dsp_driver_info->usage_mutex);
	if (ioctl_status == LINUX_SUCCESS) {
		switch (cmd) {
		case SNDCTL_DSP_SETFRAGMENT:
			{
				int user_arg;
				if (copy_from_user
				    (&user_arg, (int *)arg, sizeof(user_arg)))
					ioctl_status = -EFAULT;

	    /*---------------------------------------------------------------
		Set fragment is a suggestion, we are not required to use it, or
		return the actual value. The actual value should be returned by
		SNDCTL_DSP_GETBLKSIZE
		----------------------------------------------------*/
				break;
			}

		case SNDCTL_DSP_GETBLKSIZE:
			{
	    /*----------------------------------------------------------------
		Return our block size
		--------------------------------------------------------------*/
				if (arg) {
					return_arg =
					    GetAudioSSFrameSize
					    (dsp_driver_info->pAudioSS);

					if (copy_to_user
					    ((int *)arg, &return_arg,
					     sizeof(return_arg)))
						ioctl_status = -EFAULT;
				} else {
					ioctl_status = -EINVAL;
				}

				break;
			}

		case SNDCTL_DSP_SPEED:
			{
	    /*---------------------------------------------------------------
		Set the sample rate as close to the given rate as possible, and
		return the actual rate
	---------------------------------------------------------------*/
				if (arg) {
					int user_arg;
					if (copy_from_user
					    (&user_arg, (int *)arg,
					     sizeof(user_arg))) {
						ioctl_status = -EFAULT;
					} else {
						if (dsp_driver_info->
						    audio_state &
						    AUDIOSS_CHANGES_ALLOWED) {
							AudioSSSamplingRateType
							    sr =
						    (AudioSSSamplingRateType)
							    0;
							while (sr + 1 < SR_COUNT
							       && user_arg <
							       (AudioSSConfig.
								config[sr].
								rate +
								AudioSSConfig.
								config[sr +
								       1].
								rate) / 2)
								sr++;

							return_arg =
							    AudioSSConfig.
							    config[sr].rate;

							if (dsp_driver_info->
							    pAudioSS->
							    samplesPerSec !=
							    return_arg) {
							dsp_driver_info->
							    pAudioSS->
							    samplesPerSec
								    =
								    return_arg;
							dsp_driver_info->
								    audio_state
								    |=
							 AUDIOSS_NOT_CONFIGURED;
							}
						} else {
							return_arg =
							    dsp_driver_info->
							    pAudioSS->
							    samplesPerSec;
						}

						if (copy_to_user
						    ((int *)arg, &return_arg,
						     sizeof(return_arg)))
							ioctl_status = -EFAULT;

					}
				} else {
					ioctl_status = -EINVAL;
				}

				break;
			}

		case SNDCTL_DSP_STEREO:
			{
	    /*-----------------------------------------------------------------
		Attempt to set the audio to mono or stereo.  This IOCTL is
		obsolete, but some old code may want it. Treat it the same as
		DSP_CHANNELS.
		-------------------------------------------------------------*/
				if (arg) {
					int user_arg;
					if (copy_from_user
					    (&user_arg, (int *)arg,
					     sizeof(user_arg))) {
						ioctl_status = -EFAULT;
					} else {
						if (dsp_driver_info->
						    audio_state &
						    AUDIOSS_CHANGES_ALLOWED) {
							return_arg = user_arg;
						if (dsp_driver_info->
							    pAudioSS->bStereo !=
							    return_arg) {
							dsp_driver_info->
								    pAudioSS->
								    bStereo =
								    return_arg;
							dsp_driver_info->
								    audio_state
								    |=
						    AUDIOSS_NOT_CONFIGURED;
							}
						} else {
							return_arg =
							    dsp_driver_info->
							    pAudioSS->bStereo;
						}

						if (copy_to_user
						    ((int *)arg, &return_arg,
						     sizeof(return_arg)))
							ioctl_status = -EFAULT;

					}
				} else {
					ioctl_status = -EINVAL;
				}

				break;
		}

		case SNDCTL_DSP_CHANNELS:
			{
	    /*--------------------------------------------------------------
		Attempt to set the number of channels in the sound driver,
		return the actual setting.
		------------------------------------------------------------*/
				if (arg) {
					int user_arg;
					if (copy_from_user
					    (&user_arg, (int *)arg,
					     sizeof(user_arg))) {
						ioctl_status = -EFAULT;
					} else {
						if ((0 >= user_arg)
						    || (2 < user_arg)) {
							ioctl_status = -EFAULT;
							break;
						}

						if (dsp_driver_info->
						    audio_state &
						    AUDIOSS_CHANGES_ALLOWED) {
							return_arg = user_arg;
							user_arg--;
							if (dsp_driver_info->
							    pAudioSS->bStereo !=
							    user_arg) {
							dsp_driver_info->
								    pAudioSS->
								    bStereo =
								    user_arg;
							dsp_driver_info->
								    audio_state
								    |=
							 AUDIOSS_NOT_CONFIGURED;
							}
						} else {
							return_arg =
							    dsp_driver_info->
							    pAudioSS->bStereo;
						}

						if (copy_to_user
						    ((int *)arg, &return_arg,
						     sizeof(return_arg)))
							ioctl_status = -EFAULT;

					}
				} else {
					ioctl_status = -EINVAL;
				}

				break;
			}

		case SNDCTL_DSP_SETFMT:
			{
	    /*------------------------------------------------------------------
		Attempt to set the sound format, return the actual setting
		-------------------------------------------------------------*/
				if (arg) {
					int user_arg;
					if (copy_from_user
					    (&user_arg, (int *)arg,
					     sizeof(user_arg))) {
						ioctl_status = -EFAULT;
						break;
					}

	       /*---------------------------------------------------------------
		We only support 16-bit little endian, always return that,
		since we can't change it
		------------------------------------------------------------*/
					return_arg = AFMT_S16_LE;
					if (copy_to_user
					    ((int *)arg, &return_arg,
					     sizeof(return_arg)))
						ioctl_status = -EFAULT;

				} else {
					ioctl_status = -EINVAL;
				}

				break;
			}

		case SNDCTL_DSP_GETFMTS:
			{
	    /*------------------------------------------------------------------
		Return all of the sound formats we support
		-----------------------------------------------------------*/
				if (arg) {
	       /*---------------------------------------------------------------
		We OR together the bit masks for all of the formats we support
		-------------------------------------------------------------*/
					return_arg = AFMT_S16_LE;
					if (copy_to_user
					    ((int *)arg, &return_arg,
					     sizeof(return_arg)))
						ioctl_status = -EFAULT;

				} else {
					ioctl_status = -EINVAL;
				}

				break;
			}

		case SNDCTL_DSP_GETOSPACE:
			{
	    /*------------------------------------------------------------------
		Return the amount of data that can be written right now
		--------------------------------------------------------------*/
				if (arg) {
					audio_buf_info info;

					GetAudioSSBufInfo(dsp_driver_info->
							  pAudioSS, &info);

					if (copy_to_user
					    ((audio_buf_info *) arg, &info,
					     sizeof(info)))
						ioctl_status = -EFAULT;

				} else {
					ioctl_status = -EFAULT;
				}

				break;
			}
		default:
			{
				ioctl_status = -EINVAL;
				break;
			}
		}
		mutex_unlock(&dsp_driver_info->usage_mutex);
	}

	up_read(&dsp_driver_info->rw_alloc);

bail_ioctl:
	return ioctl_status;
}				/* dsp_driver_ioctl() */

/*============================================================================
FUNCTION dsp_driver_write

DESCRIPTION
   Writes data to the /dev/dsp driver for output (WriteAudioSSMono/Stereo).

DEPENDENCIES

RETURN VALUE
   if >=LINUX_SUCCESS      the number of bytes written
   EAGAIN                  Non-blocking when no space to write
   ENIVAL                  the user buffer is too large to send
   ERESTARTSYS             blocking operation interrupted, retry
   EFAULT                  an error condition was encountered that caused the
				operation to fail
   EIO                     Device has gone bad.

SIDE EFFECTS
   None
============================================================================*/
static ssize_t dsp_driver_write
    (struct file *filp,
     const char *user_space_buffer, size_t size, loff_t *f_pos) {
	int write_status = LINUX_SUCCESS;
	int len;
	static int cnt;

	if ((cnt++ & 0x3FF) == 0) {
		printk(KERN_INFO "bNonBlock %d flags:%x O_NONBLOCK:%x\n",
		       dsp_driver_info->pAudioSS->bNonBlock, filp->f_flags,
		       O_NONBLOCK);
	}
/* dsp_driver_info->pAudioSS->bNonBlock = FALSE; */

	/*---------------------------------------------------------------------
	 Do these basic sanity checks to make sure we don't crash the driver
	 and are in the correct state to provide this service.
	----------------------------------------------------------------------*/

	if (!dsp_driver_initialized) {
		write_status = -EINVAL;
		goto bail_write;
	}

	if (user_space_buffer == NULL) {
		write_status = -EINVAL;
		goto bail_write;
	}
	/* State of allocated must be held during write call.  rw_alloc
	 locks it, but in a read only state so that multiple ioctl's
	 or write's don't have to wait on each other. */
	/*write_status =
		down_read_interruptible( &dsp_driver_info->rw_alloc ); */
	down_read(&dsp_driver_info->rw_alloc);
	if (LINUX_SUCCESS == write_status) {
		if (!dsp_driver_info->allocated) {
			write_status = -EINVAL;
			goto bail_write_up;
		}
	} else {
		write_status = -EFAULT;
		goto bail_write;
	}

	/*--------------------------------------------------------------------
	 ConfigAudioSS delayed until first write.  Likewise, no more
	 configuration changes after first write.
	----------------------------------------------------------------------*/
	if (dsp_driver_info->
	    audio_state & (AUDIOSS_NOT_CONFIGURED | AUDIOSS_CHANGES_ALLOWED)) {
		write_status =
		    mutex_lock_interruptible(&dsp_driver_info->usage_mutex);
		if (write_status == LINUX_SUCCESS) {
		/* If audio has not been re-configured yet, try to read the
		 wave file header. If available configure audio from the
			header. */
			if (dsp_driver_info->
			    audio_state & AUDIOSS_CHANGES_ALLOWED) {
				WaveHeaderType header;
				memcpy(&header, user_space_buffer,
				       sizeof(WaveHeaderType));
				if (header.ChunkID == RIFF_CHUNK_ID
				    && header.Format == WAVE_FORMAT) {
					if (dsp_driver_info->pAudioSS->
					    samplesPerSec != header.SampleRate
					    || dsp_driver_info->pAudioSS->
					    bStereo != (header.NumChannels ==
							2)) {
			/* New sample rate or #channels, re-configure. */
						dsp_driver_info->pAudioSS->
						    samplesPerSec =
						    header.SampleRate;
						dsp_driver_info->pAudioSS->
						    bStereo =
						    (header.NumChannels == 2);
						dsp_driver_info->audio_state |=
						    AUDIOSS_NOT_CONFIGURED;
					}
					/* Skip the header */
					user_space_buffer +=
					    sizeof(WaveHeaderType);
					size -= sizeof(WaveHeaderType);
				}
			}
		/* once configured, AudioSS does not get reconfigured
						 (until next open)*/
			if (dsp_driver_info->
			    audio_state & AUDIOSS_NOT_CONFIGURED) {
				ConfigAudioSS(dsp_driver_info->pAudioSS);
				dsp_driver_info->audio_state &=
				    ~AUDIOSS_NOT_CONFIGURED;
			}
		/* once changes are disallowed, they are not
					allowed again (until next open)*/
			dsp_driver_info->audio_state &=
			    ~AUDIOSS_CHANGES_ALLOWED;

			mutex_unlock(&dsp_driver_info->usage_mutex);
		} else {
			write_status = -EFAULT;
			goto bail_write_up;
		}
	}

   /*--------------------------------------------------------------------------
      Sanity checks are done.  See if we can write.
   --------------------------------------------------------------------------*/
	write_status = mutex_lock_interruptible(&dsp_driver_info->write_mutex);
	if (write_status == LINUX_SUCCESS) {
		dsp_driver_info->pAudioSS->bNonBlock =
			 (filp->f_flags & O_NONBLOCK) == O_NONBLOCK;
		if (dsp_driver_info->pAudioSS->bStereo) {
			len = WriteAudioSSStereo(dsp_driver_info->pAudioSS,
				 (int *)user_space_buffer, size / (2 * 2));
			len *= 2 * 2;	/* sample size(2) * 2 channels */
		} else {
			len = WriteAudioSSMono(dsp_driver_info->pAudioSS,
				(short *)user_space_buffer, size / 2);
			len *= 2;	/* sample size(2) */
		}

		if (size == len) {
			write_status = len;
		} else if (0 == len) {
		/*if ( dsp_driver_info->pAudioSS->bNonBlock )
					write_status = -EAGAIN; */
			if (filp->f_flags & O_NONBLOCK)
				write_status = -EAGAIN;
			else
				write_status = -EIO;
					/* BAD ERROR on WriteAudioSS___ */
		} else if (0 > len) {
			write_status = -EIO;
			/* BAD ERROR on WriteAudioSS___ */
		} else {
			write_status = len;
		}

		mutex_unlock(&dsp_driver_info->write_mutex);
	} else {
		write_status = -EFAULT;
	}

bail_write_up:
	up_read(&dsp_driver_info->rw_alloc);

bail_write:
	return write_status;
}				/* dsp_driver_write() */

/*===========================================================================
FUNCTION dsp_driver_init

DESCRIPTION
   Initializes the driver module and its data structures.

DEPENDENCIES

RETURN VALUE
   LINUX_SUCCESS if initialization passed
   <LINUX_SUCCESS for an error (error given by return code)

SIDE EFFECTS
===========================================================================*/
static int dsp_driver_init(void)
{
	int init_status = LINUX_SUCCESS;
	int ret;
	uint32 sizeof_dsp_driver_info_mem;
	dev_t dev;

	/*--------------------------------------------------------------------
	 Don't reinitialize the driver if its already initialized.
	---------------------------------------------------------------------*/
	if (dsp_driver_initialized)
		return LINUX_SUCCESS;

	/*---------------------------------------------------------------------
	Initialize open control mutex and lock it (prevents open's until ready)
	--------------------------------------------------------------------*/
	mutex_init(&dsp_driver_open_mutex);
	mutex_lock(&dsp_driver_open_mutex);

	/*----------------------------------------------------------------------
	Allocate and initialize memory for dsp_driver_info (one for each minor)
	--------------------------------------------------------------------*/
	sizeof_dsp_driver_info_mem = sizeof(dsp_driver_info_type);
	dsp_driver_info = kmalloc(sizeof_dsp_driver_info_mem, GFP_KERNEL);
	if (dsp_driver_info == NULL)
		return -EINVAL;


	dsp_driver_info->allocated = FALSE;
	dsp_driver_info->audio_state = AUDIOSS_INIT;
	dsp_driver_info->need_module_put = 0;
	/* Initialize the semaphores so that we can grab them once
						before blocking */
	init_rwsem(&dsp_driver_info->rw_alloc);
	mutex_init(&dsp_driver_info->usage_mutex);
	mutex_init(&dsp_driver_info->write_mutex);

	/*---------------------------------------------------------------------
	 Initialize the Audio SubSystem (for each minor)
	------------------------------------------------------------------*/
	if (0 == audioFIFOBufferLen)
		audioFIFOBufferLen = AUDIO_BUFFER_LEN;
	dsp_driver_info->pAudioSS = InitAudioSS(audioFIFOBufferLen);

   /*-------------------------------------------------------------------------
     Register the driver with the kernel using the 2.6 registration
     mechanism.
   -------------------------------------------------------------------------*/
	dev = MKDEV(DSP_MAJOR, DSP_MINOR);
	ret = register_chrdev_region(dev, 1, DSP_DRIVER_DEVICE_NAME);
	if (ret < 0)
		return -EINVAL;


	cdev_init(&dsp_cdev, &dsp_fops);
	ret = cdev_add(&dsp_cdev, dev, 1);
	if (ret < 0)
		return -EINVAL;

   /*-------------------------------------------------------------------------
     All is well...
   -------------------------------------------------------------------------*/

	/*----------------------------------------------------------------
	 Create the procfs entry to allow us to emit debug info.  For more
	 info on how this works, look at Linux/Documentation/DocBook/
	 procfs_example.c under the LINUX source code tree.
	----------------------------------------------------------------*/
	dsp_driver_procfs_qcom_dir = proc_mkdir("qcom", NULL);
	if (dsp_driver_procfs_qcom_dir != NULL) {
		if (create_proc_read_entry("dsp", 0, dsp_driver_procfs_qcom_dir,
					   dsp_driver_read_proc_dsp,
					   NULL) == NULL) {
			printk(KERN_DEBUG
	       "dsp_driver_init error, create_proc_read_entry failed\n");
		}
	} else {
		printk(KERN_DEBUG "dsp_driver_init error, proc_mkdir error\n");
	}

	/*---------------------------------------------------------------------
	Set this to TRUE so that other functions will know that the driver
	 is able to provide service.
	----------------------------------------------------------------------*/
	dsp_driver_initialized = TRUE;

	mutex_unlock(&dsp_driver_open_mutex);	/* enable opens */

	return init_status;
}				/* dsp_driver_init() */

/*============================================================================
FUNCTION dsp_driver_cleanup

DESCRIPTION
   Terminates the driver module by deallocating resources and disconnecting
   services.

DEPENDENCIES
   None

RETURN VALUE
   None

SIDE EFFECTS
   None
============================================================================*/
static void dsp_driver_cleanup(void)
{
	dev_t dev;

	if (dsp_driver_initialized) {
		dsp_driver_initialized = FALSE;

		/*--------------------------------------------------------------
		 Remove the proc fs entries that were added before. This is a
		 void function, so there is no need for error checking.
		--------------------------------------------------------------*/
		remove_proc_entry("dsp", dsp_driver_procfs_qcom_dir);
		remove_proc_entry("qcom", NULL);

		/*-------------------------------------------------------------
		We must unregister all previously registered major device nos.
		as part of our cleanup. This will allow other drivers to reuse
		this major number.
		------------------------------------------------------------*/
		dev = MKDEV(DSP_MAJOR, DSP_MINOR);
		unregister_chrdev_region(dev, 1);
		cdev_del(&dsp_cdev);

		/* release AudioSS resources */
		CleanupAudioSS();

		/* account for delayed module_put's */
		while (0 < dsp_driver_info->need_module_put) {
			module_put(THIS_MODULE);
			dsp_driver_info->need_module_put--;
		}

		kfree(dsp_driver_info);
		dsp_driver_info = NULL;
	}
}				/* dsp_driver_cleanup() */

/*---------------------------------------------------------------------------
  The file_operations structure is registered with Linux when the driver is
  loaded, and points to the functions we export to handle the various
  system calls. It also indicates which system calls are not supported.
---------------------------------------------------------------------------*/
static struct file_operations dsp_fops = {
	.owner = DSP_OWNER,
	.llseek = DSP_NOTIMPLEMENTED,
	.read = DSP_NOTIMPLEMENTED,
	.aio_read = DSP_NOTIMPLEMENTED,
	.write = dsp_driver_write,
	.aio_write = DSP_NOTIMPLEMENTED,
	.readdir = DSP_NOTIMPLEMENTED,
	.poll = DSP_NOTIMPLEMENTED,
	.ioctl = dsp_driver_ioctl,
	.mmap = DSP_NOTIMPLEMENTED,
	.open = dsp_driver_open,
	.flush = DSP_NOTIMPLEMENTED,
	.release = dsp_driver_close,
	.fsync = DSP_NOTIMPLEMENTED,
	.aio_fsync = DSP_NOTIMPLEMENTED,
	.fasync = DSP_NOTIMPLEMENTED,
	.lock = DSP_NOTIMPLEMENTED,
	 /*
	   .readv              = DSP_NOTIMPLEMENTED,
	   .writev             = DSP_NOTIMPLEMENTED,
	   .sendfile           = DSP_NOTIMPLEMENTED,
	 */
	.sendpage = DSP_NOTIMPLEMENTED,
	.get_unmapped_area = DSP_NOTIMPLEMENTED
};

/*-------------------------------------------------------------------------
 Declare to LINUX the names of the init and cleanup module functions.
 This must be done after these functions declarations, hence these
 macros are used at the end of this file. If the driver is statically
 linked into the kernel, then they will be automatically called
 during LINUX's booting process(by  virtue of the __init, added
 by the module macros).
--------------------------------------------------------------------------*/
module_init(dsp_driver_init);
module_exit(dsp_driver_cleanup);

