/*
 * ether_cdc_ecm.c -- Ethernet Function driver, with CDC
 *
 * This file has been derived from gadget/ether.c
 *
 * Copyright (C) 2003-2005 David Brownell
 * Copyright (C) 2003-2004 Robert Schwebel, Benedikt Spranger
 *
 * Copyright (c) 2008 QUALCOMM USA, INC.
 *
 * All source code in this file is licensed under the following license except
 * where indicated.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 */

/* #define VERBOSE_DEBUG */

#include <linux/kernel.h>
#include <linux/utsname.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>

#include <linux/usb/ch9.h>
#include <linux/usb/cdc.h>

#include "usb_function.h"

#ifdef FEATURE_E5832
#include <linux/netlink.h>
#endif

/*-------------------------------------------------------------------------*/

/*
 * Ethernet function driver -- with CDC options
 * Builds on hardware support for a full duplex link.
 *
 * CDC Ethernet is the standard USB solution for sending Ethernet frames
 * using USB.  Real hardware tends to use the same framing protocol but look
 * different for control features.  This driver strongly prefers to use
 * this USB-IF standard as its open-systems interoperability solution;
 * most host side USB stacks (except from Microsoft) support it.
 */

#define DRIVER_DESC		"Ethernet Function CDC ECM"
#define DRIVER_VERSION		"1.0"

static const char shortname [] = "ether";
static const char driver_desc [] = DRIVER_DESC;

static unsigned int string_data;
static unsigned int string_control;
static unsigned int string_ethaddr;
#define RX_EXTRA	20		/* guard against rx overflows */

#ifdef CONFIG_USB_FUNCTION_ETH_QMI
#include "qmux.h"
#endif

/* outgoing packet filters. */
#define	DEFAULT_FILTER	(USB_CDC_PACKET_TYPE_BROADCAST \
			| USB_CDC_PACKET_TYPE_ALL_MULTICAST \
			| USB_CDC_PACKET_TYPE_PROMISCUOUS \
			| USB_CDC_PACKET_TYPE_DIRECTED)

/*-------------------------------------------------------------------------*/

struct eth_dev {
	spinlock_t		lock;
	struct usb_request	*req;		/* for control responses */
	struct usb_request	*stat_req;	/* for cdc status */

	unsigned		configured:1;
	struct usb_endpoint	*in_ep, *out_ep, *status_ep;
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	const struct usb_endpoint_descriptor
				*in, *out, *status;
#endif
#endif

	spinlock_t		req_lock;
	struct list_head	tx_reqs, rx_reqs;

	struct net_device	*net;
	struct net_device_stats	stats;
	atomic_t		tx_qlen;

	struct work_struct	work;
	unsigned		zlp:1;
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	unsigned		cdc:1;
	unsigned		qmi:1;
#endif
#endif
	unsigned		suspended:1;
	u16			cdc_filter;
	unsigned long		todo;
#define	WORK_RX_MEMORY		0
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	int			qmi_config;
#endif
#endif
	u8			host_mac [ETH_ALEN];

	int alt_set;
};

static struct usb_function usb_func_ether;

/* Ethernet function descriptors */
#define USB_DT_IAD_SIZE		8
struct usb_interface_assoc_descriptor	eth_IAD = {
	.bLength           = USB_DT_IAD_SIZE,
	.bDescriptorType   = USB_DT_INTERFACE_ASSOCIATION,
	.bInterfaceCount   = 2,
	.bFunctionClass    = USB_CLASS_COMM,
	.bFunctionSubClass = USB_CDC_SUBCLASS_ETHERNET,
	.bFunctionProtocol = USB_CDC_PROTO_NONE,
	.iFunction         = 0,
};

struct usb_interface_descriptor		eth_control_intf = {
	.bLength =  USB_DT_INTERFACE_SIZE,
	.bDescriptorType =	USB_DT_INTERFACE,
	.bNumEndpoints =	1,
	.bInterfaceClass =	USB_CLASS_COMM,
	.bInterfaceSubClass =	USB_CDC_SUBCLASS_ETHERNET,
	.bInterfaceProtocol =	USB_CDC_ACM_PROTO_VENDOR,//USB_CDC_PROTO_NONE,
};

struct usb_cdc_header_desc		eth_header_desc = {
	.bLength =		sizeof(struct usb_cdc_header_desc),
	.bDescriptorType =	USB_DT_CS_INTERFACE,
	.bDescriptorSubType =	USB_CDC_HEADER_TYPE,
	.bcdCDC =		__constant_cpu_to_le16(0x0110),
};

struct usb_cdc_union_desc		eth_union_desc = {
	.bLength =		sizeof(struct usb_cdc_union_desc),
	.bDescriptorType =	USB_DT_CS_INTERFACE,
	.bDescriptorSubType =	USB_CDC_UNION_TYPE,
};

struct usb_cdc_ether_desc 		eth_ether_desc = {
	.bLength =		sizeof(struct usb_cdc_ether_desc),
	.bDescriptorType =	USB_DT_CS_INTERFACE,
	.bDescriptorSubType =	USB_CDC_ETHERNET_TYPE,
	/* this descriptor actually adds value, surprise! */
	.bmEthernetStatistics =	__constant_cpu_to_le32(0), /* no statistics */
	.wMaxSegmentSize =	__constant_cpu_to_le16(ETH_FRAME_LEN),
	.wNumberMCFilters =	__constant_cpu_to_le16(0),
	.bNumberPowerFilters =	0,
};

struct usb_endpoint_descriptor 		eth_control_intf_hs_int_in_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_IN,
	.bmAttributes =         USB_ENDPOINT_XFER_INT,
	.bInterval =           4,
	.wMaxPacketSize =       64,
};

struct usb_endpoint_descriptor 		eth_control_intf_fs_int_in_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_IN,
	.bmAttributes =         USB_ENDPOINT_XFER_INT,
	.bInterval =           4,
	.wMaxPacketSize =       64,
};

struct usb_interface_descriptor 	eth_data_alt_zero_intf = {
	.bLength =  USB_DT_INTERFACE_SIZE,
	.bDescriptorType =      USB_DT_INTERFACE,
	.bAlternateSetting =    0,
	.bNumEndpoints =        0,
	.bInterfaceClass =      USB_CLASS_CDC_DATA,
	.bInterfaceSubClass =   0,
	.bInterfaceProtocol =   0,
};

struct usb_interface_descriptor 	eth_data_alt_one_intf = {
	.bLength =              USB_DT_INTERFACE_SIZE,
	.bDescriptorType =      USB_DT_INTERFACE,
	.bAlternateSetting =    0,/*1,*/
	.bNumEndpoints =        2,
	.bInterfaceClass =      USB_CLASS_CDC_DATA ,
	.bInterfaceSubClass =   0,
	.bInterfaceProtocol =   USB_CDC_PROTO_NONE,
};

struct usb_endpoint_descriptor 		eth_data_intf_hs_bulk_out_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_OUT,
	.bmAttributes =         USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =       __constant_cpu_to_le16(512),
};

struct usb_endpoint_descriptor 		eth_data_intf_fs_bulk_out_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_OUT,
	.bmAttributes =         USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =       __constant_cpu_to_le16(64),
};

struct usb_endpoint_descriptor 		eth_data_intf_hs_bulk_in_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_IN,
	.bmAttributes =         USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =       __constant_cpu_to_le16(512),
};

struct usb_endpoint_descriptor 		eth_data_intf_fs_bulk_in_ep_desc = {
	.bDescriptorType =      USB_DT_ENDPOINT,
	.bLength =              USB_DT_ENDPOINT_SIZE,
	.bEndpointAddress =     USB_DIR_IN,
	.bmAttributes =         USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =       __constant_cpu_to_le16(64),
};

#ifdef FEATURE_E5832
#ifdef FEATURE_E5832//CONFIG_USB_FUNCTION_ETH_QMI
static struct usb_interface_descriptor qmi_intf = {
    .bLength =		USB_DT_INTERFACE_SIZE,
    .bDescriptorType =	USB_DT_INTERFACE,
    .bInterfaceNumber =	0,
    .bAlternateSetting = 0,
    .bNumEndpoints =	3,
    .bInterfaceClass =	USB_CLASS_VENDOR_SPEC,
    .bInterfaceSubClass =	0xff,
    .bInterfaceProtocol =	USB_CDC_ACM_PROTO_VENDOR,
};
struct usb_endpoint_descriptor 	qmi_control_intf_hs_int_in_ep_desc = {
    .bDescriptorType =      USB_DT_ENDPOINT,
    .bLength =              USB_DT_ENDPOINT_SIZE,
    .bEndpointAddress =     USB_DIR_IN,
    .bmAttributes =         USB_ENDPOINT_XFER_INT,
    .bInterval =               0x05,
    .wMaxPacketSize =   __constant_cpu_to_le16(64),
};
struct usb_endpoint_descriptor  qmi_data_intf_hs_bulk_in_ep_desc=
{
    .bLength =USB_DT_ENDPOINT_SIZE,
    .bDescriptorType =USB_DT_ENDPOINT,
    .bEndpointAddress =	USB_DIR_IN,
    .bmAttributes =USB_ENDPOINT_XFER_BULK,
    .wMaxPacketSize = __constant_cpu_to_le16(512),
    .bInterval =               0x20,
};
struct usb_endpoint_descriptor  qmi_data_intf_hs_bulk_out_ep_desc=
{
    .bLength =USB_DT_ENDPOINT_SIZE,
    .bDescriptorType =USB_DT_ENDPOINT,
    .bEndpointAddress =	USB_DIR_OUT,
    .bmAttributes =USB_ENDPOINT_XFER_BULK,
    .wMaxPacketSize = __constant_cpu_to_le16(512),
    .bInterval =               0x20,
};
struct usb_endpoint_descriptor 	qmi_control_intf_fs_int_in_ep_desc = {
    .bDescriptorType =      USB_DT_ENDPOINT,
    .bLength =              USB_DT_ENDPOINT_SIZE,
    .bEndpointAddress =     USB_DIR_IN,
    .bmAttributes =         USB_ENDPOINT_XFER_INT,
    .bInterval =               0x05,
    .wMaxPacketSize =   __constant_cpu_to_le16(64),
};
struct usb_endpoint_descriptor  qmi_data_intf_fs_bulk_in_ep_desc=
{
    .bLength =USB_DT_ENDPOINT_SIZE,
    .bDescriptorType =USB_DT_ENDPOINT,
    .bEndpointAddress =	USB_DIR_IN,
    .bmAttributes =USB_ENDPOINT_XFER_BULK,
    .wMaxPacketSize = __constant_cpu_to_le16(64),
    .bInterval =               0x20,
};
struct usb_endpoint_descriptor  qmi_data_intf_fs_bulk_out_ep_desc=
{
    .bLength =USB_DT_ENDPOINT_SIZE,
    .bDescriptorType =USB_DT_ENDPOINT,
    .bEndpointAddress =	USB_DIR_OUT,
    .bmAttributes =USB_ENDPOINT_XFER_BULK,
    .wMaxPacketSize = __constant_cpu_to_le16(64),
    .bInterval =               0x20,
};
#endif
#endif
struct eth_dev *eth_device;

#ifdef FEATURE_E5832
static void qmi_command_complete (struct usb_endpoint *, struct usb_request *);
#endif
/* Some systems will want different product identifers published in the
 * device descriptor, either numbers or strings or both.  These string
 * parameters are in UTF-8 (superset of ASCII's 7 bit characters).
 */

static int os_type;
static char *dev_addr;
/* this address is invisible to ifconfig */
static char *host_addr;

static char ethaddr [2 * ETH_ALEN + 1];

#ifdef FEATURE_E5832
/*-------------------------------------------------------------------------*/
/* maxpacket and other transfer characteristics vary by speed. */
static inline struct usb_endpoint_descriptor *
ept_desc(struct usb_endpoint_descriptor *hs,
		struct usb_endpoint_descriptor *fs)
{
	if (usb_msm_get_speed() == USB_SPEED_HIGH)
		return hs;
	return fs;
}

#ifdef CONFIG_USB_FUNCTION_ETH_QMI
/* "main" config is either CDC, or its simple subset */
static inline int is_cdc(struct eth_dev *dev)
{
	return 1;		/* only cdc possible */
}

/* "secondary" RNDIS config may sometimes be activated */
static inline int qmi_active(struct eth_dev *dev)
{
	return dev->qmi;
}
#define	subset_active(dev)	(!is_cdc(dev) && !qmi_active(dev))
#define	cdc_active(dev)		( is_cdc(dev) && !qmi_active(dev))
#endif
#endif
#define DEFAULT_QLEN	2	/* double buffering by default */

/* peak bulk transfer bits-per-second */
#ifdef FEATURE_E5832
#define	HS_BPS		100000000//(13 * 512 * 8 * 1000 * 8)
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
#define	FS_BPS		10000000//(19 *  64 * 1 * 1000 * 8)
#define	DEVSPEED	USB_SPEED_HIGH

static unsigned qmult = 5;

#endif
/* for dual-speed hardware, use deeper queues at highspeed */
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
#define qlen(speed) \
	(DEFAULT_QLEN*((speed == USB_SPEED_HIGH) ? qmult : 1))

static inline int BITRATE(unsigned speed)
{
	return (speed == USB_SPEED_HIGH) ? HS_BPS : FS_BPS;
}
#else
#define qlen (DEFAULT_QLEN * 5) /* High Speed */
#endif
#endif

/*-------------------------------------------------------------------------*/

#define xprintk(d, level, fmt, args...) \
	printk(level "%s: " fmt, (d)->net->name, ## args)

#ifdef DEBUG
#undef DEBUG
#define DEBUG(dev, fmt,args...) \
	xprintk(dev, KERN_DEBUG, fmt, ## args)
#else
#define DEBUG(dev, fmt, args...) \
	do { } while (0)
#endif /* DEBUG */

#ifdef VERBOSE_DEBUG
#define VDEBUG	DEBUG
#else
#define VDEBUG(dev, fmt, args...) \
	do { } while (0)
#endif /* DEBUG */

#define ERROR(dev, fmt, args...) \
	xprintk(dev, KERN_ERR, fmt, ## args)
#define WARN(dev, fmt, args...) \
	xprintk(dev, KERN_WARNING, fmt, ## args)
#define INFO(dev, fmt, args...) \
	xprintk(dev, KERN_INFO, fmt, ## args)

/*-------------------------------------------------------------------------*/

#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
/* USB DRIVER HOOKUP (to the hardware driver, below us), mostly
 * ep0 implementation:  descriptors, config management, setup().
 * also optional class-specific notification interrupt transfer.
 */

/*
 * DESCRIPTORS ... most are static, but strings and (full) configuration
 * descriptors are built on demand.  For now we do either full CDC, or
 * our simple subset, with RNDIS as an optional second configuration.
 *
 * RNDIS includes some CDC ACM descriptors ... like CDC Ethernet.  But
 * the class descriptors match a modem (they're ignored; it's really just
 * Ethernet functionality), they don't need the NOP altsetting, and the
 * status transfer endpoint isn't optional.
 */

#define STRING_MANUFACTURER		1
#define STRING_PRODUCT			2
#define STRING_ETHADDR			3
#define STRING_DATA			4
#define STRING_CONTROL			5
#define STRING_RNDIS_CONTROL		6
#define STRING_CDC			7
#define STRING_SUBSET			8
#define STRING_RNDIS			9
#define STRING_SERIALNUMBER		10

/* holds our biggest descriptor (or RNDIS response) */
#define USB_BUFSIZ	256

/*
 * This device advertises one configuration, eth_config, unless RNDIS
 * is enabled (rndis_config) on hardware supporting at least two configs.
 *
 * NOTE:  Controllers like superh_udc should probably be able to use
 * an RNDIS-only configuration.
 *
 * FIXME define some higher-powered configurations to make it easier
 * to recharge batteries ...
 */

static char				manufacturer [50];
#endif
#endif
/* include the status endpoint if we can, even where it's optional.
 * use wMaxPacketSize big enough to fit CDC_NOTIFY_SPEED_CHANGE in one
 * packet, to simplify cancellation; and a big transfer interval, to
 * waste less bandwidth.
 *
 * some drivers (like Linux 2.4 cdc-ether!) "need" it to exist even
 * if they ignore the connect/disconnect notifications that real ether
 * can provide.  more advanced cdc configurations might want to support
 * encapsulated commands (vendor-specific, using control-OUT).
 */
#define STATUS_BYTECOUNT		16	/* 8 byte header + data */


static void eth_start(struct eth_dev *dev, gfp_t gfp_flags);
static int alloc_requests(struct eth_dev *dev, unsigned n, gfp_t gfp_flags);

static int set_ether_config(struct eth_dev *dev, gfp_t gfp_flags)
{
	int result = 0;

#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
    unsigned int speed = 0;
	/* status endpoint used for RNDIS and (optionally) CDC */
       if(os_type == WINDOWS_TYPE)
       {
		if (!subset_active(dev) && dev->status_ep) {
			dev->status = ept_desc(&qmi_control_intf_hs_int_in_ep_desc, &qmi_control_intf_fs_int_in_ep_desc);
			usb_ept_enable (dev->status_ep, 1);
		}

		dev->in = ept_desc(&qmi_data_intf_hs_bulk_in_ep_desc, &qmi_data_intf_fs_bulk_in_ep_desc);

		dev->out = ept_desc(&qmi_data_intf_hs_bulk_out_ep_desc, &qmi_data_intf_fs_bulk_out_ep_desc);
       }
	else
	{
	      if (!subset_active(dev) && dev->status_ep) {
			dev->status = ept_desc(&eth_control_intf_hs_int_in_ep_desc, &eth_control_intf_fs_int_in_ep_desc);
			usb_ept_enable (dev->status_ep, 1);
		}

		dev->in = ept_desc(&eth_data_intf_hs_bulk_in_ep_desc, &eth_data_intf_fs_bulk_in_ep_desc);

		dev->out = ept_desc(&eth_data_intf_hs_bulk_out_ep_desc, &eth_data_intf_fs_bulk_out_ep_desc);
	}
	/* With CDC,  the host isn't allowed to use these two data
	 * endpoints in the default altsetting for the interface.
	 * so we don't activate them yet.  Reset from SET_INTERFACE.
	 *
	 * Strictly speaking RNDIS should work the same: activation is
	 * a side effect of setting a packet filter.  Deactivation is
	 * from REMOTE_NDIS_HALT_MSG, reset from REMOTE_NDIS_RESET_MSG.
	 */
	if (!cdc_active(dev)) {
		usb_ept_enable (dev->in_ep, 1);
        usb_ept_enable (dev->out_ep, 1);
		
	}
    dev->cdc_filter = DEFAULT_FILTER;
	speed = usb_msm_get_speed();
	result = alloc_requests(dev, qlen(speed) , gfp_flags);
   	
	/* activate non-CDC configs right away
	 * this isn't strictly according to the RNDIS spec
	 */
	if (!cdc_active (dev)) {
		netif_carrier_on (dev->net);
		
#ifdef FEATURE_E5832       
        if (!syswatch_nl_inited)
        {
            syswatch_usb_state |= SYSWATCH_NDIS_UP;

            syswatch_usb_state &= ~SYSWATCH_NDIS_DOWN;
        } 
        else
        {
            syswatch_nl_send(NULL, NDIS_UP);
        }
#endif
		if (netif_running (dev->net)) {
			spin_lock (&dev->lock);
			eth_start (dev, GFP_ATOMIC);
			spin_unlock (&dev->lock);
		}
	}
	
	if (result == 0)
		DEBUG(dev, "qlen %d\n", qlen(speed));
#else
	if (dev->status_ep)
		usb_ept_enable(dev->status_ep, 1);

	result = alloc_requests(dev, qlen , gfp_flags);
	if (result == 0)
		DEBUG(dev, "qlen %d\n", qlen);
#endif
#endif
	/* caller is responsible for cleanup on error */
	return result;
}

static void eth_reset_config(struct eth_dev *dev)
{
	struct usb_request	*req;
	unsigned long  flags;

	DEBUG(dev, "%s\n", __func__);

	if (!dev)
		return;
	if (!dev->net)
		return;

	if (dev->configured == 0)
		return;
	netif_stop_queue(dev->net);
	netif_carrier_off(dev->net);
    printk(KERN_EMERG "qmi_usb_uninit\n");
    qmi_usb_uninit(dev->qmi_config);
	/* disable endpoints, forcing (synchronous) completion of
	 * pending i/o.  then free the requests.
	 */
	if (dev->in_ep) {
		usb_ept_enable(dev->in_ep, 0);
		spin_lock_irqsave(&dev->req_lock, flags);
		while (likely(!list_empty(&dev->tx_reqs))) {
			req = container_of(dev->tx_reqs.next,
						struct usb_request, list);
			list_del(&req->list);
			spin_unlock_irqrestore(&dev->req_lock, flags);
			usb_ept_free_req(dev->in_ep, req);
			spin_lock_irqsave(&dev->req_lock, flags);
		}
		spin_unlock_irqrestore(&dev->req_lock, flags);
	}
	if (dev->out_ep) {
		usb_ept_enable(dev->out_ep, 0);
		spin_lock_irqsave(&dev->req_lock, flags);
		while (likely(!list_empty(&dev->rx_reqs))) {
			req = container_of(dev->rx_reqs.next,
						struct usb_request, list);
			list_del(&req->list);
			spin_unlock_irqrestore(&dev->req_lock, flags);
			usb_ept_free_req(dev->out_ep, req);
			spin_lock_irqsave(&dev->req_lock, flags);
		}
		spin_unlock_irqrestore(&dev->req_lock, flags);
	}

	if (dev->status_ep)
		usb_ept_free_req(dev->status_ep, 0);
	dev->qmi = 0;
	dev->cdc_filter = 0;
	dev->configured = 0;
}

/* change our operational config.  must agree with the code
 * that returns config descriptors, and altsetting code.
 */
static int eth_set_config(struct eth_dev *dev,  gfp_t gfp_flags)
{
	int result = 0;

	eth_reset_config(dev);
#ifdef FEATURE_E5832
    dev->qmi = 1;
    printk(KERN_EMERG "eth_set_config\n");
#endif
	result = set_ether_config(dev, gfp_flags);
	if (result)
		eth_reset_config(dev);
	else
		dev->configured = 1;
	return result;
}

#ifdef FEATURE_E5832
static void usb_configure_single_ep(struct eth_dev *dev)
{
    if (usb_msm_get_speed() == USB_SPEED_HIGH) 
    {
        usb_configure_endpoint(dev->status_ep,
        			&qmi_control_intf_hs_int_in_ep_desc);
        usb_configure_endpoint(dev->in_ep,
        			&qmi_data_intf_hs_bulk_in_ep_desc);
        usb_configure_endpoint(dev->out_ep,
        			&qmi_data_intf_hs_bulk_out_ep_desc);
    } 
    else 
    {
        usb_configure_endpoint(dev->status_ep,
        			&qmi_control_intf_fs_int_in_ep_desc);
        usb_configure_endpoint(dev->in_ep,
        			&qmi_data_intf_fs_bulk_in_ep_desc);
        usb_configure_endpoint(dev->out_ep,
        			&qmi_data_intf_fs_bulk_out_ep_desc);
    }
}
static void usb_configure_cdc_ep(struct eth_dev *dev)
{
    if (usb_msm_get_speed() == USB_SPEED_HIGH) 
    {
        usb_configure_endpoint(dev->status_ep,
        			&eth_control_intf_hs_int_in_ep_desc);
        usb_configure_endpoint(dev->in_ep,
        			&eth_data_intf_hs_bulk_in_ep_desc);
        usb_configure_endpoint(dev->out_ep,
        			&eth_data_intf_hs_bulk_out_ep_desc);
    } 
    else 
    {
        usb_configure_endpoint(dev->status_ep,
        			&eth_control_intf_fs_int_in_ep_desc);
        usb_configure_endpoint(dev->in_ep,
        			&eth_data_intf_fs_bulk_in_ep_desc);
        usb_configure_endpoint(dev->out_ep,
        			&eth_data_intf_fs_bulk_out_ep_desc);
    }
}
#endif
static void eth_configure(int configured, void *_ctxt)
{
    int                     result = 0;
    struct eth_dev *dev = (struct eth_dev *) _ctxt;

    printk(KERN_INFO "%s(%d)\n", __func__, configured);

    if (!configured) 
    {
        eth_reset_config(dev);
        return ;
    }
    if (dev->configured == 1)
    {
        return ;
    }
#ifdef FEATURE_E5832
    if(os_type == WINDOWS_TYPE)
    {
        usb_configure_single_ep(dev);
    }
    else
    {
        usb_configure_cdc_ep(dev);
    }

#endif
	result = eth_set_config(dev, GFP_ATOMIC);
}
/* The interrupt endpoint is used in CDC networking models (Ethernet, ATM)
 * only to notify the host about link status changes (which we support)
 * Since we want this CDC Ethernet code to be vendor-neutral, only one
 * status request is ever queued.
 */

static void
eth_status_complete(struct usb_endpoint *ep, struct usb_request *req)
{
	struct usb_cdc_notification	*event = req->buf;
	int				value = req->status;

	/* issue the second notification if host reads the first */
	if (event->bNotificationType == USB_CDC_NOTIFY_NETWORK_CONNECTION
			&& value == 0) {
		__le32	*data = req->buf + sizeof *event;

		event->bmRequestType = 0xA1;
		event->bNotificationType = USB_CDC_NOTIFY_SPEED_CHANGE;
		event->wValue = __constant_cpu_to_le16(0);
#ifdef FEATURE_E5832
if(os_type == WINDOWS_TYPE)
{
		event->wIndex =	__constant_cpu_to_le16(
				qmi_intf.bInterfaceNumber);
}
else
{
		event->wIndex =	__constant_cpu_to_le16(
				eth_data_alt_one_intf.bInterfaceNumber);
}
#endif
		event->wLength = __constant_cpu_to_le16(8);

		/* SPEED_CHANGE data is up/down speeds in bits/sec */
		data [0] = data [1] = cpu_to_le32(HS_BPS);

		req->length = STATUS_BYTECOUNT;
		value = usb_ept_queue_xfer(ep, req);
		DEBUG(dev, "send SPEED_CHANGE --> %d\n", value);
		if (value == 0)
			return;
	} else if (value != -ECONNRESET)
		DEBUG(dev, "event %02x --> %d\n",
			event->bNotificationType, value);
	req->context = NULL;
}

static void issue_start_status(struct eth_dev *dev)
{
	struct usb_request		*req = dev->stat_req;
	struct usb_cdc_notification	*event;
	int				value;

	DEBUG(dev, "%s, flush old status first\n", __func__);

	/* flush old status
	 *
	 * FIXME ugly idiom, maybe we'd be better with just
	 * a "cancel the whole queue" primitive since any
	 * unlink-one primitive has way too many error modes.
	 * here, we "know" toggle is already clear...
	 *
	 * FIXME iff req->context != null just dequeue it
	 */
	usb_ept_enable(dev->status_ep,  0);
	usb_ept_enable(dev->status_ep, 1);

	/* 3.8.1 says to issue first NETWORK_CONNECTION, then
	 * a SPEED_CHANGE.  could be useful in some configs.
	 */
	event = req->buf;
	event->bmRequestType = 0xA1;
	event->bNotificationType = USB_CDC_NOTIFY_NETWORK_CONNECTION;
	event->wValue = __constant_cpu_to_le16(1);	/* connected */
#ifdef FEATURE_E5832
      if(os_type == WINDOWS_TYPE)
      {
		event->wIndex = __constant_cpu_to_le16(
					qmi_intf.bInterfaceNumber);
	}
	else
	{
		event->wIndex = __constant_cpu_to_le16(
					eth_data_alt_one_intf.bInterfaceNumber);
	}
#endif
	event->wLength = 0;

	req->length = sizeof *event;
	req->complete = eth_status_complete;
	req->context = dev;

	value = usb_ept_queue_xfer(dev->status_ep, req);
	if (value < 0)
		DEBUG(dev, "status buf queue --> %d\n", value);
}

static int  eth_set_interface(int  wIndex, int wValue, void *_ctxt)
{
#ifdef FEATURE_E5832
	struct usb_request *req;
	int			value = -EOPNOTSUPP;
	struct eth_dev *dev = eth_device;
	if (dev == NULL)
		return 1;
	req = dev->req;
    if(os_type == WINDOWS_TYPE)
    {
	    if (wIndex == qmi_intf.bInterfaceNumber)
	    {
        if (wValue > 1)
        {
            return 0;
        }
        usb_ept_enable(dev->status_ep, 0);
        usb_ept_enable(dev->in_ep, 0);
        usb_ept_enable(dev->out_ep, 0);
        if(wValue == 1)
        {
            if (!cdc_active (dev))
            {
                return 0;
            }
            dev->alt_set = 0;
            usb_ept_enable(dev->status_ep, 1);	
            usb_ept_enable(dev->in_ep, 1);
            usb_ept_enable(dev->out_ep, 1);
 			printk(KERN_EMERG "eth_set_interface: cdc_filter will be DEFAULT_FILTER\n");	
            dev->cdc_filter = DEFAULT_FILTER;
            netif_carrier_on(dev->net);
            
            if (dev->status)
            {
                issue_start_status(dev);
            }
            
            if (netif_running(dev->net)) 
            {
                spin_lock (&dev->lock);
                eth_start(dev, GFP_ATOMIC);
                spin_unlock(&dev->lock);
            }
            else
            {
                dev->alt_set = 0;
                netif_stop_queue(dev->net);
                netif_carrier_off(dev->net);
            }
        }
        value = 0;	       
        } 
    } 
    else
    {
        if (wIndex == eth_data_alt_one_intf.bInterfaceNumber)
        {
            if (wValue > 1)
            return 0;
            usb_ept_enable(dev->in_ep, 0);
            usb_ept_enable(dev->out_ep, 0);
            if(wValue == 1)
            {
                if (!cdc_active (dev))
                return 0;
                
                dev->alt_set = 1;
                usb_ept_enable(dev->in_ep, 1);
                usb_ept_enable(dev->out_ep, 1);
                dev->cdc_filter = DEFAULT_FILTER;
                netif_carrier_on(dev->net);
                
                if (dev->status)
                {
                    issue_start_status(dev);
                }
                
                if (netif_running(dev->net)) 
                {
                    spin_lock(&dev->lock);
                    eth_start(dev, GFP_ATOMIC);
                    spin_unlock(&dev->lock);
                }
                else
                {
                    dev->alt_set = 0;
                    netif_stop_queue(dev->net);
                    netif_carrier_off(dev->net);
                }
            }
            value = 0;
        }
        else if (wIndex == eth_control_intf.bInterfaceNumber)
        {
            dev->alt_set = 0;
            if (wValue != 0)
                return 0;
            
            if (dev->status) 
            {
                usb_ept_enable(dev->status_ep, 0);
                usb_ept_enable(dev->status_ep, 1);
            }
            value = 0;
        }
    }
    if (value >= 0) 
    {
        req->length = value;		
		usb_ept0out_queue(req);
    }
#endif
	return 0;
}

static int eth_get_interface(int wIndex, void *_ctxt)
{
	struct eth_dev *dev = eth_device;
#ifdef FEATURE_E5832
	if (!(cdc_active(dev) || qmi_active(dev)) && wIndex != 0)
		return 0;

	/* for CDC, iff carrier is on, data interface is active. */
	if (qmi_active(dev) || wIndex != 1)
		return 0;
	else
		return netif_carrier_ok (dev->net) ? 1 : 0;
#endif
}

/*
 * The setup() callback implements all the ep0 functionality that's not
 * handled lower down.  CDC has a number of less-common features:
 *
 *  - class-specific descriptors for the control interface
 *  - class-specific control requests
 */
static int
eth_setup(struct usb_ctrlrequest *ctrl, void *buf, int len, void *_ctxt)
{
	struct eth_dev	*dev = (struct eth_dev *) _ctxt;
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	struct usb_request *req;
#endif
#endif
	int		value = -EOPNOTSUPP;
	u16		wIndex = le16_to_cpu(ctrl->wIndex);
	u16		wValue = le16_to_cpu(ctrl->wValue);
	u16		wLength = le16_to_cpu(ctrl->wLength);
	u16         data_int = 0;
	u16		ctrl_int = 0;
#ifdef FEATURE_E5832
       if(os_type == WINDOWS_TYPE)
       {
             data_int = qmi_intf.bInterfaceNumber;
	         ctrl_int = 	 qmi_intf.bInterfaceNumber;
       }
        else
       {
		data_int = eth_data_alt_one_intf.bInterfaceNumber;
		ctrl_int = eth_control_intf.bInterfaceNumber;
	}
#endif
	switch (ctrl->bRequest) {
	case USB_CDC_SET_ETHERNET_PACKET_FILTER:
		/* see 6.2.30: no data, wIndex = interface,
		 * wValue = packet filter bitmap
		 */
		if (ctrl->bRequestType != (USB_TYPE_CLASS|USB_RECIP_INTERFACE)
			|| wLength != 0
			|| ((wIndex != data_int) && (wIndex != ctrl_int)))
			break;
		DEBUG(dev, "packet filter %02x\n", wValue);
		dev->cdc_filter = wValue;
		value = 0;
		break;

	/* and potentially:
	 * case USB_CDC_SET_ETHERNET_MULTICAST_FILTERS:
	 * case USB_CDC_SET_ETHERNET_PM_PATTERN_FILTER:
	 * case USB_CDC_GET_ETHERNET_PM_PATTERN_FILTER:
	 * case USB_CDC_GET_ETHERNET_STATISTIC:
	 */

#ifdef FEATURE_E5832
	/* RNDIS uses the CDC command encapsulation mechanism to implement
	 * an RPC scheme, with much getting/setting of attributes by OID.
	 */
	case USB_CDC_SEND_ENCAPSULATED_COMMAND:
		if (ctrl->bRequestType != (USB_TYPE_CLASS|USB_RECIP_INTERFACE)
				|| !qmi_active(dev)
				|| wLength > USB_BUFSIZ
				|| wValue
				/*|| rndis_control_intf.bInterfaceNumber
					!= wIndex*/)
		    {
			break;
		    }
		/* read the request, then process it */
		req = dev->req;
		value = wLength;
    	req->complete = qmi_command_complete;
		/* later, rndis_control_ack () sends a notification */
        if (value >= 0) 
       	{
    		req->length = value;		
		    usb_ept0out_queue(req);
        }
		break;

	case USB_CDC_GET_ENCAPSULATED_RESPONSE:
		if ((USB_DIR_IN|USB_TYPE_CLASS|USB_RECIP_INTERFACE)
					== ctrl->bRequestType
				&& qmi_active(dev)
				// && wLength >= 0x0400
				&& !wValue
				/*&& rndis_control_intf.bInterfaceNumber
					== wIndex*/) 
		{
			u8 *buffer;
			u32 n;
                     req = usb_get_ep0in_req();
			/* return the result */
			buffer = qmi_get_next_response(dev->qmi_config, &n);
            
			if (buffer) {
			    if(req->buf)
				    memcpy(req->buf, buffer, n);
				qmi_free_response(dev->qmi_config, buffer);
				value = n;
				if (value >= 0) 
	       	        {
					    req->length = value;
					    //usb_ept0out_queue(req);
				}
			}
			/* if with no data ready, just send empty ack*/	
			else{
				value = 0;
			}
			/* else stalls ... spec says to avoid that */
            break;
		}
		else if ((USB_DIR_IN|USB_TYPE_CLASS|USB_RECIP_INTERFACE)
					== ctrl->bRequestType
				&& !qmi_active(dev)					
				&& !wValue)
		{
		     printk("wenp: ENCAPSULATED_RESPONSE qmi not active");
             req = usb_get_ep0in_req();		
		     value = 0;
		     req->length = 0;			 
		}
		break;
#endif
	default:
		VDEBUG(dev,
			"unknown control req%02x.%02x v%04x i%04x l%d\n",
			ctrl->bRequestType, ctrl->bRequest,
			wValue, wIndex, wLength);
	}
	return value;
}


static void eth_disconnect(void *_ctxt)
{
	struct eth_dev		*dev = (struct eth_dev *) _ctxt;
	unsigned long		flags;

	printk(KERN_INFO "eth_disconnect()\n");
	spin_lock_irqsave(&dev->lock, flags);
	netif_stop_queue(dev->net);
	netif_carrier_off(dev->net);
#ifdef FEATURE_E5832
    if (!syswatch_nl_inited )
    {
        syswatch_usb_state |= SYSWATCH_NDIS_DOWN;
        syswatch_usb_state &= ~SYSWATCH_NDIS_UP;
    }
    else
    {
        syswatch_nl_send(NULL, NDIS_DOWN);
    }
#endif
	eth_reset_config(dev);
	spin_unlock_irqrestore(&dev->lock, flags);
}

/*-------------------------------------------------------------------------*/

/* NETWORK DRIVER HOOKUP (to the layer above this driver) */

static int eth_change_mtu(struct net_device *net, int new_mtu)
{
	struct eth_dev	*dev = netdev_priv(net);
	if (dev->qmi)
		return -EBUSY;

	if (new_mtu <= ETH_HLEN || new_mtu > ETH_FRAME_LEN)
		return -ERANGE;
	/* no zero-length packet read wanted after mtu-sized packets */
	if (((new_mtu + sizeof(struct ethhdr)) %
			(usb_ept_get_max_packet(dev->in_ep))) == 0)
		return -EDOM;
	net->mtu = new_mtu;
	return 0;
}

static struct net_device_stats *eth_get_stats(struct net_device *net)
{
	return &((struct eth_dev *)netdev_priv(net))->stats;
}

static void eth_get_drvinfo(struct net_device *net, struct ethtool_drvinfo *p)
{
	strlcpy(p->driver, shortname, sizeof p->driver);
	strlcpy(p->version, DRIVER_VERSION, sizeof p->version);
	strlcpy(p->fw_version, "ethernet", sizeof p->fw_version);
}

static u32 eth_get_link(struct net_device *net)
{
	return 1;
}

static struct ethtool_ops ops = {
	.get_drvinfo = eth_get_drvinfo,
	.get_link = eth_get_link
};

static void defer_kevent(struct eth_dev *dev, int flag)
{
	if (test_and_set_bit(flag, &dev->todo))
		return;
	if (!schedule_work(&dev->work))
		ERROR(dev, "kevent %d may have been dropped\n", flag);
	else
		DEBUG(dev, "kevent %d scheduled\n", flag);
}

static void rx_complete(struct usb_endpoint *ep, struct usb_request *req);

static int
rx_submit(struct eth_dev *dev, struct usb_request *req, gfp_t gfp_flags)
{
	struct sk_buff		*skb;
	int			retval = -ENOMEM;
	size_t			size;
	unsigned long		flags;
	/* Padding up to RX_EXTRA handles minor disagreements with host.
	 * Normally we use the USB "terminate on short read" convention;
	 * so allow up to (N*max_pkt), since that memory is normally
	 * already allocated.  Some hardware doesn't deal well with short
	 * reads (e.g. DMA must be N*max_pkt), so for now don't trim a
	 * byte off the end (to force hardware errors on overflow).
	 */
	size = (sizeof(struct ethhdr) + dev->net->mtu + RX_EXTRA);
	size += usb_ept_get_max_packet(dev->out_ep) - 1;
	size -= size % usb_ept_get_max_packet(dev->out_ep);
	skb = alloc_skb(size + NET_IP_ALIGN, gfp_flags);
	if (skb  == NULL) {
		DEBUG(dev, "no rx skb\n");
		goto enomem;
	}

	/* Some platforms perform better when IP packets are aligned,
	 * but on at least one, checksumming fails otherwise.
	 */
	skb_reserve(skb, NET_IP_ALIGN);

	req->buf = skb->data;
	req->length = size;
	req->complete = rx_complete;
	req->context = skb;

	retval = usb_ept_queue_xfer(dev->out_ep, req);
	if (retval == -ENOMEM)
enomem:
		defer_kevent(dev, WORK_RX_MEMORY);
	if (retval) {
		DEBUG(dev, "rx submit --> %d\n", retval);
		if (skb)
			dev_kfree_skb_any(skb);
		spin_lock_irqsave(&dev->req_lock, flags);
		list_add(&req->list, &dev->rx_reqs);
		spin_unlock_irqrestore(&dev->req_lock, flags);
	}
	return retval;
}

static void rx_complete(struct usb_endpoint *ep, struct usb_request *req)
{
	struct sk_buff	*skb = req->context;
	struct eth_dev	*dev = eth_device;
	int		status = req->status;
	switch (status) {

	/* normal completion */
	case 0:
		skb_put(skb, req->actual);
		/* we know MaxPacketsPerTransfer == 1 here */
		if (status < 0
				|| ETH_HLEN > skb->len
				|| skb->len > ETH_FRAME_LEN) {
			dev->stats.rx_errors++;
			dev->stats.rx_length_errors++;
			DEBUG(dev, "rx length %d\n", skb->len);
			break;
		}

		skb->protocol = eth_type_trans(skb, dev->net);
		dev->stats.rx_packets++;
		dev->stats.rx_bytes += skb->len;

		/* no buffer copies needed, unless hardware can't
		 * use skb buffers.
		 */
		status = netif_rx(skb);
		skb = NULL;
		break;

	/* software-driven interface shutdown */
	case -ECONNRESET:		/* unlink */
	case -ESHUTDOWN:		/* disconnect etc */
		VDEBUG(dev, "rx shutdown, code %d\n", status);
		goto quiesce;

	/* for hardware automagic (such as pxa) */
	case -ECONNABORTED:		/* endpoint reset */
		DEBUG(dev, "rx %s reset\n", ep->name);
		defer_kevent(dev, WORK_RX_MEMORY);
quiesce:
		dev_kfree_skb_any(skb);
		goto clean;

	/* data overrun */
	case -EOVERFLOW:
		dev->stats.rx_over_errors++;
		/* FALLTHROUGH */

	default:
		dev->stats.rx_errors++;
		DEBUG(dev, "rx status %d\n", status);
		break;
	}

	if (skb)
		dev_kfree_skb_any(skb);
	if (!netif_running(dev->net)) {
clean:
		spin_lock(&dev->req_lock);
		list_add(&req->list, &dev->rx_reqs);
		spin_unlock(&dev->req_lock);
		req = NULL;
	}
	if (req)
		rx_submit(dev, req, GFP_ATOMIC);
}

static int prealloc(struct list_head *list, struct usb_endpoint *ep,
			unsigned n, gfp_t gfp_flags)
{
	unsigned		i;
	struct usb_request	*req;

	if (!n)
		return -ENOMEM;

	/* queue/recycle up to N requests */
	i = n;
	list_for_each_entry(req, list, list) {
		if (i-- == 0)
			goto extra;
	}
	while (i--) {
		/* CDC ECM uses skb buffer pointer for requests */
		req = usb_ept_alloc_req(ep, 0);
		if (!req)
			return list_empty(list) ? -ENOMEM : 0;
		list_add(&req->list, list);
	}
	return 0;

extra:
	/* free extras */
	for (;;) {
		struct list_head	*next;

		next = req->list.next;
		list_del(&req->list);
		usb_ept_free_req(ep, req);

		if (next == list)
			break;

		req = container_of(next, struct usb_request, list);
	}
	return 0;
}

static int alloc_requests(struct eth_dev *dev, unsigned n, gfp_t gfp_flags)
{
	int status;
	unsigned long flags;

	spin_lock_irqsave(&dev->req_lock, flags);
	status = prealloc(&dev->tx_reqs, dev->in_ep, n, gfp_flags);
	if (status < 0)
		goto fail;
	status = prealloc(&dev->rx_reqs, dev->out_ep, n, gfp_flags);
	if (status < 0)
		goto fail;
	goto done;
fail:
	DEBUG(dev, "can't alloc requests\n");
done:
	spin_unlock_irqrestore(&dev->req_lock, flags);
	return status;
}

static void rx_fill(struct eth_dev *dev, gfp_t gfp_flags)
{
	struct usb_request	*req;
	unsigned long		flags;
	/* fill unused rxq slots with some skb */
	spin_lock_irqsave(&dev->req_lock, flags);
	while (!list_empty(&dev->rx_reqs)) {
		req = container_of(dev->rx_reqs.next,
				struct usb_request, list);
		list_del_init(&req->list);
		spin_unlock_irqrestore(&dev->req_lock, flags);

		if (rx_submit(dev, req, gfp_flags) < 0) {
			defer_kevent(dev, WORK_RX_MEMORY);
			return;
		}

		spin_lock_irqsave(&dev->req_lock, flags);
	}
	spin_unlock_irqrestore(&dev->req_lock, flags);
}

static void eth_work(struct work_struct *work)
{
	struct eth_dev	*dev = container_of(work, struct eth_dev, work);

	if (test_and_clear_bit(WORK_RX_MEMORY, &dev->todo)) {
		if (netif_running(dev->net))
			rx_fill(dev, GFP_KERNEL);
	}

	if (dev->todo)
		DEBUG(dev, "work done, flags = 0x%lx\n", dev->todo);
}

static void tx_complete(struct usb_endpoint *ep, struct usb_request *req)
{
	struct sk_buff	*skb = req->context;
	struct eth_dev	*dev = eth_device;

	switch (req->status) {
	default:
		dev->stats.tx_errors++;
		VDEBUG(dev, "tx err %d\n", req->status);
		/* FALLTHROUGH */
	case -ECONNRESET:		/* unlink */
	case -ESHUTDOWN:		/* disconnect etc */
		break;
	case 0:
		dev->stats.tx_bytes += skb->len;
	}
	dev->stats.tx_packets++;

	spin_lock(&dev->req_lock);
	list_add(&req->list, &dev->tx_reqs);
	spin_unlock(&dev->req_lock);
	dev_kfree_skb_any(skb);

	atomic_dec(&dev->tx_qlen);
	if (netif_carrier_ok(dev->net))
		netif_wake_queue(dev->net);
}

static inline int eth_is_promisc(struct eth_dev *dev)
{
	return dev->cdc_filter & USB_CDC_PACKET_TYPE_PROMISCUOUS;
}

static int eth_start_xmit(struct sk_buff *skb, struct net_device *net)
{
	struct eth_dev		*dev = netdev_priv(net);
	int			length = skb->len;
	int			retval;
	struct usb_request	*req = NULL;
	unsigned long		flags;

	/* apply outgoing CDC filters */
	if (!eth_is_promisc(dev)) {
		u8		*dest = skb->data;

		if (is_multicast_ether_addr(dest)) {
			u16	type;

			/* ignores USB_CDC_PACKET_TYPE_MULTICAST and host
			 * SET_ETHERNET_MULTICAST_FILTERS requests
			 */
			if (is_broadcast_ether_addr(dest))
				type = USB_CDC_PACKET_TYPE_BROADCAST;
			else
				type = USB_CDC_PACKET_TYPE_ALL_MULTICAST;
			if (!(dev->cdc_filter & type)) {
				dev_kfree_skb_any(skb);
				return 0;
			}
		}
		/* ignores USB_CDC_PACKET_TYPE_DIRECTED */
	}

	spin_lock_irqsave(&dev->req_lock, flags);
	/*
	 * this freelist can be empty if an interrupt triggered disconnect()
	 * and reconfigured the function (shutting down this queue) after the
	 * network stack decided to xmit but before we got the spinlock.
	 */
	if (list_empty(&dev->tx_reqs)) {
		spin_unlock_irqrestore(&dev->req_lock, flags);
		return 1;
	}

	req = container_of(dev->tx_reqs.next, struct usb_request, list);
	list_del(&req->list);

	/* temporarily stop TX queue when the freelist empties */
	if (list_empty(&dev->tx_reqs))
		netif_stop_queue(net);
	spin_unlock_irqrestore(&dev->req_lock, flags);

	/* no buffer copies needed, unless the network stack did it
	 * or the hardware can't use skb buffers.
	 */
	req->buf = skb->data;
	req->context = skb;
	req->complete = tx_complete;

	/* use zlp framing on tx for strict CDC-Ether conformance,
	 * though any robust network rx path ignores extra padding.
	 * and some hardware doesn't like to write zlps.
	 */
	if (!dev->zlp && (length % usb_ept_get_max_packet(dev->in_ep)) == 0)
		length++;

	req->length = length;

	retval = usb_ept_queue_xfer(dev->in_ep, req);
	switch (retval) {
	default:
		DEBUG(dev, "tx queue err %d\n", retval);
		break;
	case 0:
		net->trans_start = jiffies;
		atomic_inc(&dev->tx_qlen);
	}
	if (retval) {
		dev->stats.tx_dropped++;
		dev_kfree_skb_any(skb);
		spin_lock_irqsave(&dev->req_lock, flags);
		if (list_empty(&dev->tx_reqs))
			netif_start_queue(net);
		list_add(&req->list, &dev->tx_reqs);
		spin_unlock_irqrestore(&dev->req_lock, flags);
	}
	return 0;
}


#ifdef FEATURE_E5832
/*wenping 2009-06-26 add begin*/
 /* The RNDIS control queue is processed by GET_ENCAPSULATED_RESPONSE, and
 * normally just one notification will be queued.
 */

static void
qmi_control_ack_complete (struct usb_endpoint *ep, struct usb_request *req)
{
	struct eth_dev          *dev = eth_device;

	if (req->status || req->actual != req->length)
		DEBUG (dev,
			"rndis control ack complete --> %d, %d/%d\n",
			req->status, req->actual, req->length);
	req->context = NULL;

	if (req != dev->stat_req)
		usb_ept_free_req(ep, req);
}

static int qmi_control_ack (struct net_device *net)
{
	struct eth_dev          *dev = eth_device;//netdev_priv(net);
	int                     length;
	struct usb_request      *resp = dev->stat_req;
	/* in case qmi calls this after disconnect */
	if (!dev->status) {
		DEBUG (dev, "status ENODEV\n");
		return -ENODEV;
	}

	/* in case queue length > 1 */
	if (resp->context) {
		resp = usb_ept_alloc_req(dev->status_ep, 8);
		if (!resp)
		{
			return -ENOMEM;
		}
	}

	/* Send QMI RESPONSE_AVAILABLE notification;
	 * USB_CDC_NOTIFY_RESPONSE_AVAILABLE should work too
	 */
	resp->length = 8;
	resp->complete = qmi_control_ack_complete;
	resp->context = dev;

        /*ack buffer "a1 01 00 00 01 00 00 00"*/
	*((__le32 *) resp->buf) = __constant_cpu_to_le32 (QMI_ACK_DWORD1);
	*((__le32 *) resp->buf + 1) = __constant_cpu_to_le32 (QMI_ACK_DWORD2);

	length = usb_ept_queue_xfer (dev->status_ep, resp);
	if (length < 0) {
		resp->status = 0;
		qmi_control_ack_complete (dev->status_ep, resp);
	}

	return 0;
}
static void qmi_command_complete (struct usb_endpoint *ep, struct usb_request *req)
{
	struct eth_dev          *dev = eth_device;
	int			status;
	/* received qmi command from USB_CDC_SEND_ENCAPSULATED_COMMAND */
	spin_lock(&dev->lock);
	status = qmi_msg_parser_p(dev->qmi_config, (u8 *) req->buf);
	if (status < 0)
		ERROR(dev, "%s: rndis parse error %d\n", __FUNCTION__, status);
	spin_unlock(&dev->lock);
}       
/*wenping 2009-06-26 add end*/


#endif

static void eth_start(struct eth_dev *dev, gfp_t gfp_flags)
{
	DEBUG(dev, "%s\n", __func__);

	/* fill the rx queue */
	rx_fill(dev, gfp_flags);

	/* and open the tx floodgates */
	atomic_set(&dev->tx_qlen, 0);
	netif_wake_queue(dev->net);
#ifdef FEATURE_E5832

#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	if (qmi_active(dev)) {
	    unsigned speed = usb_msm_get_speed();
		qmi_set_param_medium (dev->qmi_config,
					NDIS_MEDIUM_802_3,
					BITRATE(speed)/100);
		(void) qmi_signal_connect (dev->qmi_config);
	}
#endif	/* RNDIS */
#endif
}

static int eth_open(struct net_device *net)
{
	struct eth_dev		*dev = netdev_priv(net);

	DEBUG(dev, "%s\n", __func__);
	if (netif_carrier_ok(dev->net))
		eth_start(dev, GFP_KERNEL);
	return 0;
}

static int eth_stop(struct net_device *net)
{
	struct eth_dev		*dev = netdev_priv(net);

	VDEBUG(dev, "%s\n", __func__);
	netif_stop_queue(net);

	DEBUG(dev, "stop stats: rx/tx %ld/%ld, errs %ld/%ld\n",
		dev->stats.rx_packets, dev->stats.tx_packets,
		dev->stats.rx_errors, dev->stats.tx_errors
		);

	/* ensure there are no more active requests */
	if (dev->configured) {
		usb_ept_enable(dev->in_ep, 0);
		usb_ept_enable(dev->out_ep, 0);
		if (netif_carrier_ok(dev->net)) {
			DEBUG(dev, "host still using in/out endpoints\n");
			/* FIXME idiom may leave toggle wrong here */
			usb_ept_enable(dev->in_ep, 1);
			usb_ept_enable(dev->out_ep, 1);
		}
		if (dev->status_ep) {
			usb_ept_enable(dev->status_ep, 0);
			usb_ept_enable(dev->status_ep,  1);
		}
	}
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	if (qmi_active(dev)) {
		qmi_set_param_medium(dev->qmi_config, NDIS_MEDIUM_802_3, 0);
		(void) qmi_signal_disconnect (dev->qmi_config);
	}
#endif	/* RNDIS */
#endif

	return 0;
}


static u8 __devinit nibble(unsigned char c)
{
	if (likely(isdigit(c)))
		return c - '0';
	c = toupper(c);
	if (likely(isxdigit(c)))
		return 10 + c - 'A';
	return 0;
}

static int __devinit get_ether_addr(const char *str, u8 *dev_addr)
{
	if (str) {
		unsigned	i;

		for (i = 0; i < 6; i++) {
			unsigned char num;

			if ((*str == '.') || (*str == ':'))
				str++;
			num = nibble(*str++) << 4;
			num |= (nibble(*str++));
			dev_addr [i] = num;
		}
	    dev_addr[0] = dev_addr[0] | 0xFE;
		if (is_valid_ether_addr(dev_addr))
			return 0;
	}
	random_ether_addr(dev_addr);
	dev_addr[0] = dev_addr[0] | 0xFE;
	return 1;
}

static void  eth_unbind(void *_ctxt)
{
	struct eth_dev   *dev = (struct eth_dev *)_ctxt ;

	if (!dev)
		return ;
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
    qmi_deregister (dev->qmi_config);
    qmi_exit();
#endif	/* RNDIS */
#endif
	if (dev->in_ep) {
		usb_ept_fifo_flush(dev->in_ep);
		usb_ept_enable(dev->in_ep, 0);
		usb_free_endpoint(dev->in_ep);
	}
	if (dev->out_ep) {
		usb_ept_fifo_flush(dev->out_ep);
		usb_ept_enable(dev->out_ep, 0);
		usb_free_endpoint(dev->out_ep);
	}
	if (dev->status_ep) {
		usb_ept_fifo_flush(dev->status_ep);
		usb_ept_enable(dev->status_ep, 0);
		usb_free_endpoint(dev->status_ep);
	}
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
/* we've already been disconnected ... no i/o is active */
	if (dev->req) {
		usb_free_ep0out_req(dev->req);
		dev->req = NULL;
	}
	if (dev->stat_req) {
		usb_ept_free_req (dev->status_ep, dev->stat_req);
		dev->stat_req = NULL;
	}
#endif	/* RNDIS */
#endif


	if (dev->net) {
		unregister_netdev(dev->net);
		free_netdev(dev->net);
	}
	return ;
}
static int eth_resume(void *_ctxt)
{
    struct eth_dev *dev = eth_device;

    if(NULL == dev)
	{
	    return 0;
	}

    printk(KERN_ERR"eth_resume\n");

    /*contrl the qmi function*/
    if(qmi_active(dev))
	{
	    printk(KERN_INFO "qmi is active and resume qmi\n");
	    qmi_resume(dev->qmi_config);
	}
	return 0;

}
static void  eth_bind(void *_ctxt)
{
	struct eth_dev		*dev;
	struct net_device	*net;
	u8			zlp = 0;
#ifdef FEATURE_E5832
    u8          qmi = 1;
#endif
	struct usb_endpoint     *in_ep, *out_ep, *status_ep = NULL;
	int			status = -ENOMEM;
	int			ret;
	struct device		*get_dev;

	get_dev = usb_get_device();
#ifdef FEATURE_E5832
    /* Enable IAD */
    if(os_type != WINDOWS_TYPE)
    {
        usb_msm_enable_iad();
    }
    /* Configuring STATUS endpoint */
    status_ep = usb_alloc_endpoint(USB_DIR_IN);
    status_ep->max_pkt = 64;

    /* Configuring OUT endpoint */
    out_ep = usb_alloc_endpoint(USB_DIR_OUT);
    out_ep->max_pkt = 512;

    /*Configuring IN Endpoint*/
    in_ep = usb_alloc_endpoint(USB_DIR_IN);
    in_ep->max_pkt = 512;

    ret = usb_msm_get_next_ifc_number(&usb_func_ether);
	snprintf (manufacturer, sizeof manufacturer, "%s %s/%s",
		init_utsname()->sysname, init_utsname()->release,
		"Huawei Incorporated");
    string_data = usb_msm_get_next_strdesc_id("Ethernet Data");
    
    qmi = 1;
    if (string_data != 0) 
    {
    	string_control = usb_msm_get_next_strdesc_id
    			 ("RNDIS Communications Control");
    	if (string_control != 0) {
    		string_ethaddr = usb_msm_get_next_strdesc_id(ethaddr);
    		if (string_ethaddr != 0) {
    			eth_ether_desc.iMACAddress = string_ethaddr;
    		}
    	}
    }
    if(os_type == WINDOWS_TYPE)
    {
        qmi_intf.bInterfaceNumber = ret;
        qmi_intf.iInterface = string_data;
        qmi_control_intf_hs_int_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | status_ep->num;
      
	     qmi_control_intf_fs_int_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | status_ep->num;

        qmi_data_intf_hs_bulk_out_ep_desc.bEndpointAddress =
        					USB_DIR_OUT | out_ep->num;

	    qmi_data_intf_fs_bulk_out_ep_desc.bEndpointAddress =
        					USB_DIR_OUT | out_ep->num;  

	    qmi_data_intf_hs_bulk_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | in_ep->num;
       qmi_data_intf_fs_bulk_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | in_ep->num;
    }
    else
    {
        eth_control_intf.bInterfaceNumber = ret;
        eth_control_intf.iInterface = string_control;
        eth_IAD.bFirstInterface = ret;
        eth_union_desc.bMasterInterface0 = ret;  
        ret = usb_msm_get_next_ifc_number(&usb_func_ether);

        eth_data_alt_zero_intf.bInterfaceNumber = ret;
        eth_data_alt_zero_intf.iInterface = 0;
        eth_data_alt_one_intf.bInterfaceNumber = ret;
        eth_data_alt_one_intf.iInterface = string_data;
        eth_union_desc.bSlaveInterface0 = ret;
        
        eth_control_intf_hs_int_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | status_ep->num;
        eth_control_intf_hs_int_in_ep_desc.wMaxPacketSize =
        					status_ep->max_pkt;
        eth_control_intf_fs_int_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | status_ep->num;
        eth_control_intf_hs_int_in_ep_desc.bInterval = 4;    
        
        eth_data_intf_hs_bulk_out_ep_desc.bEndpointAddress =
        					USB_DIR_OUT | out_ep->num;
        eth_data_intf_hs_bulk_out_ep_desc.wMaxPacketSize = out_ep->max_pkt;
        eth_data_intf_fs_bulk_out_ep_desc.bEndpointAddress =
        					USB_DIR_OUT | out_ep->num;   
        eth_data_intf_hs_bulk_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | in_ep->num;
        eth_data_intf_hs_bulk_in_ep_desc.wMaxPacketSize = in_ep->max_pkt;
        eth_data_intf_fs_bulk_in_ep_desc.bEndpointAddress =
        					USB_DIR_IN | in_ep->num;
 }
    
#endif

	net = alloc_etherdev(sizeof *dev);
	if (!net) {
		printk(KERN_DEBUG "eth_bind: alloc_etherdev failed \n");
		return ;
	}
	dev = netdev_priv(net);
	spin_lock_init(&dev->lock);
	spin_lock_init(&dev->req_lock);
	INIT_WORK(&dev->work, eth_work);
	INIT_LIST_HEAD(&dev->tx_reqs);
	INIT_LIST_HEAD(&dev->rx_reqs);

	/* network device setup */
	dev->net = net;
	strcpy(net->name, "usb%d");
	dev->zlp = zlp;
	dev->in_ep = in_ep;
	dev->out_ep = out_ep;
	dev->status_ep = status_ep;

	eth_device = dev;
	usb_func_ether.context = eth_device;

	/* Module params for these addresses should come from ID proms.
	 * The host side address is used with CDC and RNDIS, and commonly
	 * ends up in a persistent config database.  It's not clear if
	 * host side code for the SAFE thing cares -- its original BLAN
	 * thing didn't, Sharp never assigned those addresses on Zaurii.
	 */
	if (get_ether_addr(dev_addr, net->dev_addr))
		dev_warn(get_dev,
			"using random %s ethernet address\n", "self");
	if (get_ether_addr(host_addr, dev->host_mac))
		dev_warn(get_dev,
			"using random %s ethernet address\n", "host");
	snprintf(ethaddr, sizeof ethaddr, "%02X%02X%02X%02X%02X%02X",
		dev->host_mac [0], dev->host_mac [1],
		dev->host_mac [2], dev->host_mac [3],
		dev->host_mac [4], dev->host_mac [5]);
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
    if (qmi) 
    {
        status = qmi_usb_init();
        if (status < 0) 
        {
        	dev_err (get_dev, "can't init RNDIS, %d\n", status);
        	goto fail;
        }
    }
#endif	/* RNDIS */
#endif

	net->change_mtu = eth_change_mtu;
	net->get_stats = eth_get_stats;
	net->hard_start_xmit = eth_start_xmit;
	net->open = eth_open;
	net->stop = eth_stop;
	/* watchdog_timeo, tx_timeout ...
	 * set_multicast_list */
	SET_ETHTOOL_OPS(net, &ops);
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
	dev->req = usb_get_ep0out_req();
	if (!dev->req)
		goto fail;
#endif	/* RNDIS */
#endif
	/* ... and maybe likewise for status transfer */
	if (dev->status_ep) {
	    /* preallocate control message data and buffer */
		dev->stat_req = usb_ept_alloc_req(dev->status_ep,
					STATUS_BYTECOUNT);
		if (!dev->stat_req) {
#ifdef FEATURE_E5832
#ifdef CONFIG_USB_FUNCTION_ETH_QMI
            usb_free_ep0out_req(dev->req);
#endif	/* RNDIS */
#endif
			goto fail;
		}
		dev->stat_req->context = NULL;
	}
	/* finish hookup to lower layer ... */
	/* two kinds of host-initiated state changes:
	 *  - iff DATA transfer is active, carrier is "on"
	 *  - tx queueing enabled if open *and* carrier is "on"
	 */
	netif_stop_queue(dev->net);
	netif_carrier_off(dev->net);

	SET_NETDEV_DEV(dev->net, get_dev);
	status = register_netdev(dev->net);
	if (status < 0)
	{
		goto fail1;
	}

	INFO(dev, "%s, version: " DRIVER_VERSION "\n", driver_desc);
	INFO(dev, "MAC %02x:%02x:%02x:%02x:%02x:%02x\n",
		net->dev_addr [0], net->dev_addr [1],
		net->dev_addr [2], net->dev_addr [3],
		net->dev_addr [4], net->dev_addr [5]);

	INFO(dev, "HOST MAC %02x:%02x:%02x:%02x:%02x:%02x\n",
		dev->host_mac [0], dev->host_mac [1],
		dev->host_mac [2], dev->host_mac [3],
		dev->host_mac [4], dev->host_mac [5]);

#ifdef FEATURE_E5832
	if (qmi) {
		u32	vendorID = 0x12d1;

		/* FIXME RNDIS vendor id == "vendor NIC code" == ? */
		dev->qmi_config = qmi_register (qmi_control_ack);
		if (dev->qmi_config < 0) {
            printk(KERN_INFO"eth_bind: dev->qmi_config < 0\n");
fail0:
			unregister_netdev (dev->net);
			status = -ENODEV;
			goto fail;
		}

		/* these set up a lot of the OIDs that RNDIS needs */
		qmi_set_host_mac (dev->qmi_config, dev->host_mac);
		if (qmi_set_param_dev (dev->qmi_config, dev->net,
					 &dev->stats, &dev->cdc_filter))
			goto fail0;
		if (qmi_set_param_vendor(dev->qmi_config, vendorID,
					manufacturer))
			goto fail0;
		if (qmi_set_param_medium(dev->qmi_config,
					NDIS_MEDIUM_802_3, 0))
			goto fail0;
		INFO (dev, "RNDIS ready\n");
	}
    return ;

    fail1:
        dev_dbg(get_dev, "register_netdev failed, %d\n", status);
    fail:
        printk(KERN_INFO"eth_bind: returning from eth_bind\n");
        eth_unbind (eth_device);
#endif
	return ;
}


static struct usb_function usb_func_ether = {
	.name		= "ethernet",
	.bind		= eth_bind,
	.unbind		= eth_unbind,
	.configure	= eth_configure,
	.disconnect	= eth_disconnect,
	.setup		= eth_setup,
	.set_interface	= eth_set_interface,
	.get_interface	= eth_get_interface,
	.resume         = eth_resume, 
};

#define TOTAL_ETH_DESCRIPTORS 11
struct usb_descriptor_header *eth_hs_descriptors[TOTAL_ETH_DESCRIPTORS];
struct usb_descriptor_header *eth_fs_descriptors[TOTAL_ETH_DESCRIPTORS];

#ifdef FEATURE_E5832
int init_usb_cdc_desc(void)
{
	eth_hs_descriptors[0] = (struct usb_descriptor_header *)
				&eth_IAD;
	eth_hs_descriptors[1] = (struct usb_descriptor_header *)
				&eth_control_intf;
	eth_hs_descriptors[2] = (struct usb_descriptor_header *)
				&eth_header_desc;
	eth_hs_descriptors[3] = (struct usb_descriptor_header *)
				&eth_ether_desc;
	eth_hs_descriptors[4] = (struct usb_descriptor_header *)
				&eth_union_desc;
	eth_hs_descriptors[5] = (struct usb_descriptor_header *)
				&eth_control_intf_hs_int_in_ep_desc;
	eth_hs_descriptors[6] = (struct usb_descriptor_header *)
				&eth_data_alt_one_intf;
	eth_hs_descriptors[7] = (struct usb_descriptor_header *)
				&eth_data_intf_hs_bulk_in_ep_desc;
	eth_hs_descriptors[8] = (struct usb_descriptor_header *)
				&eth_data_intf_hs_bulk_out_ep_desc;
	eth_hs_descriptors[9] = (struct usb_descriptor_header *)
				NULL;

	eth_fs_descriptors[0] = (struct usb_descriptor_header *)
	            &eth_IAD;
	eth_fs_descriptors[1] = (struct usb_descriptor_header *)
				&eth_control_intf;
	eth_fs_descriptors[2] = (struct usb_descriptor_header *)
				&eth_header_desc;
	eth_fs_descriptors[3] = (struct usb_descriptor_header *)
	            &eth_ether_desc;
	eth_fs_descriptors[4] = (struct usb_descriptor_header *)
	            &eth_union_desc;
	eth_fs_descriptors[5] = (struct usb_descriptor_header *)
				&eth_control_intf_fs_int_in_ep_desc;
	eth_fs_descriptors[6] = (struct usb_descriptor_header *)
				&eth_data_alt_one_intf;
	eth_fs_descriptors[7] = (struct usb_descriptor_header *)
				&eth_data_intf_fs_bulk_in_ep_desc;
	eth_fs_descriptors[8] = (struct usb_descriptor_header *)
				&eth_data_intf_fs_bulk_out_ep_desc;
	eth_fs_descriptors[9] = (struct usb_descriptor_header *)
				NULL;

	usb_func_ether.hs_descriptors = eth_hs_descriptors;
	usb_func_ether.fs_descriptors = eth_fs_descriptors;
    return 0;
}

int init_usb_single_interface_desc(void)
{
	eth_hs_descriptors[0] = (struct usb_descriptor_header *)
				&qmi_intf;
	eth_hs_descriptors[1] = (struct usb_descriptor_header *)
				&qmi_control_intf_hs_int_in_ep_desc;
	eth_hs_descriptors[2] = (struct usb_descriptor_header *)
				&qmi_data_intf_hs_bulk_in_ep_desc;
	eth_hs_descriptors[3] = (struct usb_descriptor_header *)
				&qmi_data_intf_hs_bulk_out_ep_desc;
	eth_hs_descriptors[4] = (struct usb_descriptor_header *)
				NULL;

	eth_fs_descriptors[0] = (struct usb_descriptor_header *)
	            &qmi_intf;
	eth_fs_descriptors[1] = (struct usb_descriptor_header *)
				&qmi_control_intf_fs_int_in_ep_desc;
	eth_fs_descriptors[2] = (struct usb_descriptor_header *)
				&qmi_data_intf_fs_bulk_in_ep_desc;
	eth_fs_descriptors[3] = (struct usb_descriptor_header *)
	            &qmi_data_intf_fs_bulk_out_ep_desc;
	eth_fs_descriptors[4] = (struct usb_descriptor_header *)
	            NULL;

	usb_func_ether.hs_descriptors = eth_hs_descriptors;
	usb_func_ether.fs_descriptors = eth_fs_descriptors;
    return 0;
}
#endif

void usb_ether_descriptor_init(int type){

    os_type = type;

    if(type == WINDOWS_TYPE)
    {
    	init_usb_single_interface_desc();
    }
    else
    {
        init_usb_cdc_desc();
    }

}
EXPORT_SYMBOL(usb_ether_descriptor_init);
	
 int  usb_eth_init(void)
{
	int rc;

	usb_ether_descriptor_init(WINDOWS_TYPE);
		
	rc = usb_function_register(&usb_func_ether);

	if (rc < 0)
		printk(KERN_INFO "cdcecm init:usb function register failed \n");

    qmux_init();
	return rc;
}
EXPORT_SYMBOL(usb_eth_init);

 void usb_eth_cleanup(void)
{
	struct eth_dev          *dev = eth_device;

	usb_function_unregister(&usb_func_ether);
	if (dev) {
		dev->net = NULL;
		dev = NULL;
	}
}
EXPORT_SYMBOL(usb_eth_cleanup);

