/* ehci-msm.c - HSUSB Host Controller Driver Implementation
 *
 * Copyright (c) 2008 QUALCOMM USA, INC.
 *
 * Partly derived from ehci-fsl.c and ehci-hcd.c
 * Copyright (c) 2000-2004 by David Brownell
 * Copyright (c) 2005 MontaVista Software
 *
 * All source code in this file is licensed under the following license except
 * where indicated.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it at http://www.fsf.org
 */

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/clk.h>
#include <linux/spinlock.h>

#include <asm/arch/board.h>
#include <asm/arch/rpc_hsusb.h>

#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

#include "ehci-msm.h"

#define MSM_USB_BASE (hcd->regs)

#ifdef CONFIG_ANDROID_POWER
struct usb_wakelock {
	spinlock_t lock;
	android_suspend_lock_t wlock;
	unsigned acquired;
};
#endif

struct msm_hc_device {
	struct clk *clk;
	struct clk *pclk;
	struct usb_hcd *msm_hcd_data;
	unsigned phy_info;
	unsigned in_lpm;
	struct work_struct lpm_exit_work;
	spinlock_t lock;
#ifdef CONFIG_ANDROID_POWER
	struct usb_wakelock wake_lock;
#endif
};

#ifdef CONFIG_ANDROID_POWER
static void usb_release_wake_lock(void);
static void usb_acquire_wake_lock(void);
#else
static void usb_release_wake_lock(void) { }
static void usb_acquire_wake_lock(void) { }
#endif

static struct msm_hc_device msm_hc_dev;

#ifdef CONFIG_ANDROID_POWER
static void usb_release_wake_lock(void)
{
	unsigned flags;

	spin_lock_irqsave(&msm_hc_dev.wake_lock.lock, flags);
	if (msm_hc_dev.wake_lock.acquired) {
		android_unlock_suspend(&msm_hc_dev.wake_lock.wlock);
		msm_hc_dev.wake_lock.acquired = 0;
	}
	spin_unlock_irqrestore(&msm_hc_dev.wake_lock.lock, flags);
}

static void usb_acquire_wake_lock(void)
{
	unsigned flags;

	spin_lock_irqsave(&msm_hc_dev.wake_lock.lock, flags);
	if (!msm_hc_dev.wake_lock.acquired) {
		android_lock_suspend(&msm_hc_dev.wake_lock.wlock);
		msm_hc_dev.wake_lock.acquired = 1;
	}
	spin_unlock_irqrestore(&msm_hc_dev.wake_lock.lock, flags);
}
#endif

static void msm_vbus_online(void)
{
	pr_info("msm_hsusb: vbus online, clock enable\n");
	clk_enable(msm_hc_dev.clk);
	clk_enable(msm_hc_dev.pclk);
}

static void msm_vbus_offline(void)
{
	pr_info("msm_hsusb: vbus offline, clock disable\n");
	clk_disable(msm_hc_dev.clk);
	clk_disable(msm_hc_dev.pclk);
}

static unsigned ulpi_read(struct usb_hcd *hcd, unsigned reg)
{
	unsigned timeout = 100000;

	/* initiate read operation */
	writel(ULPI_RUN | ULPI_READ | ULPI_ADDR(reg),
	       USB_ULPI_VIEWPORT);

	/* wait for completion */
	while ((readl(USB_ULPI_VIEWPORT) & ULPI_RUN) && (--timeout))
		cpu_relax();

	if (timeout == 0) {
		printk(KERN_ERR "ulpi_read: timeout %08x\n",
			readl(USB_ULPI_VIEWPORT));
		return -ETIMEDOUT;
	}
	return ULPI_DATA_READ(readl(USB_ULPI_VIEWPORT));
}

static int ulpi_write(struct usb_hcd *hcd, unsigned val, unsigned reg)
{
	unsigned timeout = 10000;

	/* initiate write operation */
	writel(ULPI_RUN | ULPI_WRITE |
	       ULPI_ADDR(reg) | ULPI_DATA(val),
	       USB_ULPI_VIEWPORT);

	/* wait for completion */
	while ((readl(USB_ULPI_VIEWPORT) & ULPI_RUN) && (--timeout))
		cpu_relax();

	if (timeout == 0) {
		printk(KERN_ERR "ulpi_write: timeout\n");
		return -ETIMEDOUT;
	}

	return 0;
}

static int usb_wakeup_phy(struct usb_hcd *hcd)
{
	int i;

	pr_info("%s: start\n", __func__);

	writel(readl(USB_USBCMD) & ~ULPI_STP_CTRL, USB_USBCMD);

	/* some circuits automatically clear PHCD bit */
	for (i = 0; i < 5 && (readl(USB_PORTSC) & PORTSC_PHCD); i++) {
		writel(readl(USB_PORTSC) & ~PORTSC_PHCD, USB_PORTSC);
		msleep(1);
	}

	if ((readl(USB_PORTSC) & PORTSC_PHCD)) {
		pr_err("%s: cannot clear phcd bit\n", __func__);
		return -EAGAIN;
	}

	switch (PHY_TYPE(msm_hc_dev.phy_info)) {
	case USB_PHY_INTEGRATED:
		break;
	case USB_PHY_EXTERNAL:
		ulpi_write(hcd, 0x0A, 0x0E);
		ulpi_write(hcd, 0x0A, 0x11);
		break;
	default:
		pr_err("%s: undefined phy type\n", __func__);
		return -1;
	}

	pr_info("%s: end\n", __func__);

	return 0;
}
static int usb_suspend_phy(struct usb_hcd *hcd)
{
	int i;

	switch (PHY_TYPE(msm_hc_dev.phy_info)) {
	case USB_PHY_EXTERNAL:
		pr_info("%s: external phy\n", __func__);
		/* clear VBusValid and SessionEnd rising interrupts */
		ulpi_write(hcd, ULPI_C_VBUS_VALID_RISE |
				ULPI_C_SESS_END_RISE, 0x0f);
		/* clear VBusValid and SessionEnd falling interrupts */
		ulpi_write(hcd, ULPI_C_VBUS_VALID_FALL |
				ULPI_C_SESS_END_FALL, 0x12);

		/* spec talks about following bits in LPM but they are ignored
		 * because
		 * 1. disabling interface protection circuit: by disabling
		 * interface protection curcuit we cannot come out
		 * of lpm as asyn interrupts would be disabled
		 * 2. setting the suspendM bit: this bit would be set by usb
		 * controller once we set phcd bit.
		 */

		break;

	case USB_PHY_INTEGRATED:
		pr_info("%s: integrated phy\n", __func__);
		/* clearing latch register, keeping phy comparators ON and
		   turning off PLL are done because of h/w bugs */
		ulpi_read(hcd, 0x14);/* clear PHY interrupt latch register */
		ulpi_write(hcd, 0x01, 0x30);/* PHY comparators on in LPM */
		ulpi_write(hcd, 0x08, 0x09);/* turn off PLL on integrated phy */
		break;

	case USB_PHY_UNDEFINED:
		pr_err("%s: undefined phy type\n", __func__);
		return -1;
	}

	/* loop for 500 times with a delay of 1ms, input by AMSS USB team */
	for (i = 0; i < 500; i++) {
		/* set phy to be in lpm */
		writel(readl(USB_PORTSC) | PORTSC_PHCD, USB_PORTSC);

		msleep(1);
		if (readl(USB_PORTSC) & PORTSC_PHCD)
			goto blk_stp_sig;
	}

	if (!(readl(USB_PORTSC) & PORTSC_PHCD)) {
		pr_err("unable to set phcd of portsc reg\n");
		return -1;
	}

	/* we have to set this bit again to work-around h/w bug */
	writel(readl(USB_PORTSC) | PORTSC_PHCD, USB_PORTSC);

blk_stp_sig:
	/* block the stop signal */
	writel(readl(USB_USBCMD) | ULPI_STP_CTRL, USB_USBCMD);

	return 0;
}

static int usb_lpm_enter(struct usb_hcd *hcd)
{
	unsigned flags;
	spin_lock_irqsave(&msm_hc_dev.lock, flags);
	if (msm_hc_dev.in_lpm == 1) {
		pr_info("%s: already in lpm. nothing to do\n", __func__);
		spin_unlock_irqrestore(&msm_hc_dev.lock, flags);
		return 0;
	}

	if (HC_IS_RUNNING(hcd->state)) {
		pr_info("%s: can't enter into lpm. controller is runnning\n",
			__func__);
		spin_unlock_irqrestore(&msm_hc_dev.lock, flags);
		return -1;
	}

	pr_info("%s: lpm enter procedure started\n", __func__);

	msm_hc_dev.in_lpm = 1;
	disable_irq(hcd->irq);
	spin_unlock_irqrestore(&msm_hc_dev.lock, flags);

	if (usb_suspend_phy(hcd)) {
		msm_hc_dev.in_lpm = 0;
		enable_irq(hcd->irq);
		pr_info("phy suspend failed\n");
		pr_info("%s: lpm enter procedure end\n", __func__);
		return -1;
	}

	/* enable async interrupt */
	writel(readl(USB_USBCMD) | ASYNC_INTR_CTRL, USB_USBCMD);

	msm_vbus_offline();

	usb_release_wake_lock();
	enable_irq_wake(hcd->irq);
	enable_irq(hcd->irq);
	pr_info("%s: lpm enter procedure end\n", __func__);
	return 0;
}

void usb_lpm_exit_w(struct work_struct *work)
{
	struct usb_hcd *hcd = msm_hc_dev.msm_hcd_data;

	msm_vbus_online();
	writel(readl(USB_USBCMD) & ~ASYNC_INTR_CTRL, USB_USBCMD);

	if (usb_wakeup_phy(hcd)) {
		pr_err("fatal error: cannot bring phy out of lpm\n");
		return;
	}

	disable_irq_wake(hcd->irq);
	enable_irq(hcd->irq);
}


static void usb_lpm_exit(struct usb_hcd *hcd)
{
	unsigned flags;
	spin_lock_irqsave(&msm_hc_dev.lock, flags);
	if (!msm_hc_dev.in_lpm) {
		spin_unlock_irqrestore(&msm_hc_dev.lock, flags);
		return;
	}
	usb_acquire_wake_lock();
	msm_hc_dev.in_lpm = 0;
	disable_irq(hcd->irq);
	spin_unlock_irqrestore(&msm_hc_dev.lock, flags);
	schedule_work(&(msm_hc_dev.lpm_exit_work));
}

static irqreturn_t ehci_msm_irq(struct usb_hcd *hcd)
{
	if (msm_hc_dev.in_lpm == 1) {
		usb_lpm_exit(hcd);
		return IRQ_HANDLED;
	}
	return ehci_irq(hcd);
}

static int ehci_msm_bus_suspend(struct usb_hcd *hcd)
{
	int rc;
	rc = ehci_bus_suspend(hcd);
	if (rc)
		return rc;
	rc = usb_lpm_enter(hcd);
	if (rc)
		return ehci_bus_resume(hcd);
	return rc;
}

static int ehci_msm_bus_resume(struct usb_hcd *hcd)
{
	usb_lpm_exit(hcd);
	if (cancel_work_sync(&(msm_hc_dev.lpm_exit_work)))
		usb_lpm_exit_w(NULL);
	return ehci_bus_resume(hcd);

}

static int ehci_msm_init(struct ehci_hcd *ehci)
{
	int	port	= HCS_N_PORTS(ehci->hcs_params);
	u32	__iomem	*reg_ptr;
	u32	command	= ehci_readl(ehci, &ehci->regs->command);

	/* port configuration - phy, port speed, port power, port enable */
	while (port--) {
		ehci_writel(ehci, (PORT_ULPI | PORT_POWER | PORT_PE),
				&ehci->regs->port_status[port]);
		msleep(20);
	}

	reg_ptr = (u32 __iomem *)(((u8 __iomem *)ehci->regs) + USBMODE);
	/* set hostmode */
	ehci_writel(ehci, (USBMODE_VBUS | USBMODE_CM_HC |
				USBMODE_SDIS), reg_ptr);

	command |= CMD_RUN;
	ehci_writel(ehci, command, &ehci->regs->command);

	ehci_to_hcd(ehci)->state = HC_STATE_RUNNING;

	return 0;
}

static int ehci_msm_reset(struct usb_hcd *hcd)
{
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	int retval;

	ehci->caps = USB_CAPLENGTH;
	ehci->regs = USB_CAPLENGTH +
		HC_LENGTH(ehci_readl(ehci, &ehci->caps->hc_capbase));

	/* cache the data to minimize the chip reads*/
	ehci->hcs_params = ehci_readl(ehci, &ehci->caps->hcs_params);

	retval = ehci_halt(ehci);
	if (retval)
		return retval;

	retval = ehci_init(hcd);
	if (retval)
		return retval;

	ehci->is_tdi_rh_tt = 1;

	ehci->sbrn = HCD_USB2;

	ehci_reset(ehci);

	retval = ehci_msm_init(ehci);

	return retval;
}

/* SW workarounds
 * Issue#2		- Integrated PHY Calibration
 * Symptom		- Electrical compliance failure in eye-diagram tests
 * SW workaround        - Try to raise amplitude to 400mV

 * Issue#3		- AHB Posted Writes
 * Symptom		- USB stability
 * SW workaround	- This programs xtor ON, BURST disabled and
 *			unspecified length of INCR burst enabled
 */
static int ehci_msm_run(struct usb_hcd *hcd)
{
	struct ehci_hcd *ehci  = hcd_to_ehci(hcd);
	int             retval = 0;
	int     	port   = HCS_N_PORTS(ehci->hcs_params);
	u32 __iomem     *reg_ptr;
	u32             hcc_params;

	hcd->uses_new_polling = 1;
	hcd->poll_rh = 0;

	retval = ehci_halt(ehci);
	if (retval) {
		ehci_mem_cleanup(ehci);
		return retval;
	}

	/* EHCI spec section 4.1 */
	retval = ehci_reset(ehci);
	if (retval) {
		ehci_mem_cleanup(ehci);
		return retval;
	}

	/* set hostmode */
	reg_ptr = (u32 __iomem *)(((u8 __iomem *)ehci->regs) + USBMODE);
	ehci_writel(ehci, (USBMODE_VBUS | USBMODE_CM_HC |
				USBMODE_SDIS), reg_ptr);

	/* port configuration - phy, port speed, port power, port enable */
	while (port--) {
		ehci_writel(ehci, (PORT_ULPI | PORT_POWER | PORT_PE),
				&ehci->regs->port_status[port]);
		msleep(20);
	}

	if (PHY_TYPE(msm_hc_dev.phy_info) == USB_PHY_INTEGRATED) {
		/* SW workaround, Issue#2 */
		ulpi_write(hcd, 0x31, 0x0C);
		/* SW workaround, Issue#3 */
		writel(0x0, USB_AHB_MODE);
		writel(0x0, USB_SBUSCFG);
	} else {
		/* INCR4 BURST mode */
		writel(SBUSCFG_AHBBRST_INCR4, USB_SBUSCFG);
	}

	ehci_writel(ehci, ehci->periodic_dma, &ehci->regs->frame_list);
	ehci_writel(ehci, (u32)ehci->async->qh_dma, &ehci->regs->async_next);

	hcc_params = ehci_readl(ehci, &ehci->caps->hcc_params);
	if (HCC_64BIT_ADDR(hcc_params))
		ehci_writel(ehci, 0, &ehci->regs->segment);

	ehci->command &= ~(CMD_LRESET|CMD_IAAD|CMD_PSE|CMD_ASE|CMD_RESET);
	ehci->command |= CMD_RUN;
	ehci_writel(ehci, ehci->command, &ehci->regs->command);

	/*
	 * Start, enabling full USB 2.0 functionality ... usb 1.1 devices
	 * are explicitly handed to companion controller(s), so no TT is
	 * involved with the root hub.  (Except where one is integrated,
	 * and there's no companion controller unless maybe for USB OTG.)
	 *
	 * Turning on the CF flag will transfer ownership of all ports
	 * from the companions to the EHCI controller.  If any of the
	 * companions are in the middle of a port reset at the time, it
	 * could cause trouble.  Write-locking ehci_cf_port_reset_rwsem
	 * guarantees that no resets are in progress.  After we set CF,
	 * a short delay lets the hardware catch up; new resets shouldn't
	 * be started before the port switching actions could complete.
	 */

	down_write(&ehci_cf_port_reset_rwsem);
	hcd->state = HC_STATE_RUNNING;
	ehci_writel(ehci, FLAG_CF, &ehci->regs->configured_flag);
	ehci_readl(ehci, &ehci->regs->command); /* unblock posted writes */
	msleep(5);
	up_write(&ehci_cf_port_reset_rwsem);

	/*Enable appropriate Interrupts*/
	ehci_writel(ehci, INTR_MASK,
			&ehci->regs->intr_enable);

	/* power supply on vbus for peripherals */
	retval = msm_hsusb_vbus_powerup();

	return retval;
}

static struct hc_driver msm_hc_driver = {
	.description		= hcd_name,
	.product_desc 		= "Qualcomm On-Chip EHCI Host Controller",
	.hcd_priv_size 		= sizeof(struct ehci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq 			= ehci_msm_irq,
	.flags 			= HCD_USB2,

	.reset 			= ehci_msm_reset,
	.start 			= ehci_msm_run,
	.stop			= ehci_stop,
	.shutdown		= ehci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue		= ehci_urb_enqueue,
	.urb_dequeue		= ehci_urb_dequeue,
	.endpoint_disable	= ehci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number	= ehci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data	= ehci_hub_status_data,
	.hub_control		= ehci_hub_control,
	.bus_suspend		= ehci_msm_bus_suspend,
	.bus_resume		= ehci_msm_bus_resume,
	.relinquish_port	= ehci_relinquish_port,
};

static int ehci_msm_probe(struct platform_device *pdev)
{
	struct usb_hcd *hcd;
	struct resource *res;
	struct hc_driver *driver = &msm_hc_driver;
	int irq;
	int retval;
	int msec;
	u32 result;

	msm_hc_dev.msm_hcd_data = NULL;
       /* initializing phy type */
	if (pdev->dev.platform_data) {
		struct msm_hsusb_platform_data *pdata = pdev->dev.platform_data;
		msm_hc_dev.phy_info = pdata->phy_info;
		if (PHY_TYPE(msm_hc_dev.phy_info) == USB_PHY_UNDEFINED) {
			printk(KERN_ERR "undefined phytype:(%d)\n",
			      msm_hc_dev.phy_info);
			return -ENODEV;
		}
	} else
		return -ENODEV;

	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (!res) {
		printk(KERN_ALERT "Found HC with no IRQ, \
					check %s set up!\n", pdev->dev.bus_id);
		retval = -ENODEV;
		goto err_hcd_create;
	}

	irq = res->start;

	hcd = usb_create_hcd(driver, &pdev->dev, pdev->dev.bus_id);
	if (!hcd) {
		printk(KERN_ALERT "failed @ usb_create_hcd\n");
		retval = -ENOMEM;
		goto err_hcd_create;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		printk(KERN_ALERT "Found HC with no register address, \
					check %s setup!\n", pdev->dev.bus_id);
		retval = -ENODEV;
		goto err_no_regs;
	}

	hcd->rsrc_start = res->start;
	hcd->rsrc_len = res->end - res->start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len,
						driver->description)) {
		printk(KERN_ALERT "Controller already in use!!!\n");
		retval = -EBUSY;
		goto err_res_busy;
	}

	hcd->regs = ioremap(hcd->rsrc_start, hcd->rsrc_len);

	if (hcd->regs == NULL) {
		printk(KERN_ALERT "Error mapping the memory\n");
		retval = -EFAULT;
		goto err_map;
	}

	/* get usb clocks */
	msm_hc_dev.clk = clk_get(NULL, "usb_hs_clk");
	if (IS_ERR(msm_hc_dev.clk)) {
		printk(KERN_ALERT "usb_hs_clk error\n");
		retval = -EFAULT;
		goto err_map;
	}

	msm_hc_dev.pclk = clk_get(NULL, "usb_hs_pclk");
	if (IS_ERR(msm_hc_dev.pclk)) {
		printk(KERN_ALERT "usb_hs_pclk error\n");
		retval = -EFAULT;
		goto err_map;
	}

#ifdef CONFIG_ANDROID_POWER
	spin_lock_init(&msm_hc_dev.wake_lock.lock);
	msm_hc_dev.wake_lock.wlock.name = "ehci_msm_lock";
	msm_hc_dev.wake_lock.acquired = 0;
	android_init_suspend_lock(&msm_hc_dev.wake_lock.wlock);
#endif
	usb_acquire_wake_lock();
	/* enable usb clocks */
	msm_vbus_online();

	msm_hsusb_rpc_connect();

	retval = msm_hsusb_phy_reset();

	if (retval < 0)
		goto err_hcd_add;

	/* reset hc core */
	writel((readl(USB_USBCMD) | CMD_RESET), USB_USBCMD);

	/* waiting for hc core reset upto 250ms */
	for (msec = 250; msec > 0; msec--) {
		result = readl(USB_USBCMD);
		if (result == ~(u32)0) {
			retval = -ENODEV;
			goto err_hcd_add;
		}

		if ((result & CMD_RESET) == 0)
			break;

		msleep(1);
	}
	if (msec <= 0) {
		retval = -ETIMEDOUT;
		goto err_hcd_add;
	}

	/* set hostmode */
	writel((USBMODE_VBUS | USBMODE_CM_HC), USB_USBMODE);
	msm_hc_dev.in_lpm = 0;
	spin_lock_init(&msm_hc_dev.lock);
	INIT_WORK(&(msm_hc_dev.lpm_exit_work), usb_lpm_exit_w);

	retval = usb_add_hcd(hcd, irq, 0);
	if (retval != 0)
		goto err_hcd_add;

	msm_hc_dev.msm_hcd_data = hcd;

	return retval;

err_hcd_add:
	iounmap(hcd->regs);
err_map:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err_res_busy:
err_no_regs:
	usb_put_hcd(hcd);
err_hcd_create:
	msm_hsusb_rpc_close();
	msm_vbus_offline();
	printk(KERN_ALERT "Init %s fail, %d\n", pdev->dev.bus_id, retval);
	return retval;
}

static int ehci_msm_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = msm_hc_dev.msm_hcd_data;
	int retval = 0;

	cancel_work_sync(&msm_hc_dev.lpm_exit_work);
#ifdef CONFIG_ANDROID_POWER
	android_uninit_suspend_lock(&msm_hc_dev.wake_lock.wlock);
#endif
	if (msm_hc_dev.msm_hcd_data) {
		usb_remove_hcd(hcd);
		iounmap(hcd->regs);
		release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
		usb_put_hcd(hcd);
	}

	retval = msm_hsusb_vbus_shutdown();
	if (retval < 0)
		return retval;

	retval = msm_hsusb_rpc_close();
	if (retval < 0)
		return retval;

	msm_vbus_offline();
	clk_put(msm_hc_dev.clk);
	clk_put(msm_hc_dev.pclk);

	return retval;
}

static struct platform_driver ehci_msm_driver = {
	.probe	= ehci_msm_probe,
	.remove	= ehci_msm_remove,
	.driver	= {.name = "msm_hsusb"},
};
