/* drivers/video/msm/src/drv/mddi/mddi.c
 *
 * Copyright (C) 2008 QUALCOMM USA, INC.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/system.h>
#include <asm/mach-types.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/clk.h>
#include <linux/platform_device.h>

#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

#include "msm_fb.h"
#include "mddihosti.h"
#include "mddihost.h"
#include <asm/arch-msm/gpio.h>

static int mddi_probe(struct platform_device *pdev);
static int mddi_remove(struct platform_device *pdev);

static int mddi_off(struct platform_device *pdev);
static int mddi_on(struct platform_device *pdev);

static struct platform_device *pdev_list[MSM_FB_MAX_DEV_LIST];
static int pdev_list_cnt;
static struct clk *mddi_clk;

#ifdef CONFIG_ANDROID_POWER
static android_early_suspend_t early_suspend;
#endif

static struct platform_driver mddi_driver = {
	.probe = mddi_probe,
	.remove = mddi_remove,
	.suspend = NULL,
	.suspend_late = NULL,
	.resume_early = NULL,
	.resume = NULL,
	.shutdown = NULL,
	.driver = {
		   .name = "mddi",
		   },
};

extern int int_mddi_pri_flag;

static int mddi_off(struct platform_device *pdev)
{
	int ret = 0;

	ret = panel_next_off(pdev);

	if (machine_is_msm7201a_ffa())
		gpio_direction_output(88, 0);

	return ret;
}

static int mddi_on(struct platform_device *pdev)
{
	int ret = 0;

	if (machine_is_msm7201a_ffa())
		gpio_direction_output(88, 1);

	ret = panel_next_on(pdev);

	return ret;
}

static int mddi_resource_initialized;
static struct mddi_platform_data *mddi_pdata;

static int mddi_probe(struct platform_device *pdev)
{
	struct msm_fb_data_type *mfd;
	struct platform_device *mdp_dev = NULL;
	struct msm_fb_panel_data *pdata = NULL;
	int rc;

	if ((pdev->id == 0) && (pdev->num_resources >= 0)) {
		mddi_pdata = pdev->dev.platform_data;
		msm_pmdh_base = (u8 *)pdev->resource[0].start;
		MSM_FB_INFO("primary mddi base address = 0x%x\n",
				(int)msm_pmdh_base);
		mddi_resource_initialized = 1;
		return 0;
	}

	if (!mddi_resource_initialized)
		return -EPERM;

	mfd = platform_get_drvdata(pdev);

	if (!mfd)
		return -ENODEV;

	if (mfd->key != MFD_KEY)
		return -EINVAL;

	if (pdev_list_cnt >= MSM_FB_MAX_DEV_LIST)
		return -ENOMEM;

	mdp_dev = platform_device_alloc("mdp", pdev->id);
	if (!mdp_dev)
		return -ENOMEM;

	/////////////////////////////////////////
	// link to the latest pdev
	/////////////////////////////////////////
	mfd->pdev = mdp_dev;
	mfd->dest = DISPLAY_LCD;

	/////////////////////////////////////////
	// alloc panel device data
	/////////////////////////////////////////
	if (platform_device_add_data
	    (mdp_dev, pdev->dev.platform_data,
	     sizeof(struct msm_fb_panel_data))) {
		printk(KERN_ERR "mddi_probe: platform_device_add_data failed!\n");
		platform_device_put(mdp_dev);
		return -ENOMEM;
	}
	/////////////////////////////////////////
	// data chain
	/////////////////////////////////////////
	pdata = mdp_dev->dev.platform_data;
	pdata->on = mddi_on;
	pdata->off = mddi_off;
	pdata->next = pdev;

	/////////////////////////////////////////
	// get/set panel specific fb info
	/////////////////////////////////////////
	mfd->panel_info = pdata->panel_info;
	mfd->fb_imgType = MDP_RGB_565;

	if (mfd->panel_info.pdest == DISPLAY_1)
		mfd->mddi_vdopkt = MDDI_DEFAULT_PRIM_PIX_ATTR;
	else
		mfd->mddi_vdopkt = 0x400;

	mddi_host_clock_config(mfd, mddi_clk);

	/////////////////////////////////////////
	// set driver data
	/////////////////////////////////////////
	platform_set_drvdata(mdp_dev, mfd);

	/////////////////////////////////////////
	// register in mdp driver
	/////////////////////////////////////////
	rc = platform_device_add(mdp_dev);
	if (rc) {
		goto mddi_probe_err;
	}

	pdev_list[pdev_list_cnt++] = pdev;
	return 0;

      mddi_probe_err:
	platform_device_put(mdp_dev);
	return rc;
}

#ifdef CONFIG_ANDROID_POWER
static int mddi_pad_ctrl;

static void mddi_early_suspend(android_early_suspend_t *h)
{
	mddi_host_type host_idx = MDDI_HOST_PRIM;

	mddi_pad_ctrl = mddi_host_reg_in(PAD_CTL);
	mddi_host_reg_out(PAD_CTL, 0x0);
	clk_disable(mddi_clk);
	disable_irq(INT_MDDI_PRI);

	if (mddi_pdata && mddi_pdata->mddi_power_save)
		mddi_pdata->mddi_power_save(0);
}

static void mddi_early_resume(android_early_suspend_t * h)
{
	mddi_host_type host_idx = MDDI_HOST_PRIM;

	if (mddi_pdata && mddi_pdata->mddi_power_save)
		mddi_pdata->mddi_power_save(1);

	mddi_host_reg_out(PAD_CTL, mddi_pad_ctrl);
	enable_irq(INT_MDDI_PRI);
	clk_enable(mddi_clk);
}
#endif

extern struct timer_list mddi_host_timer;

static int mddi_remove(struct platform_device *pdev)
{
	if (mddi_host_timer.function)
		del_timer_sync(&mddi_host_timer);

	return 0;
}

static int mddi_register_driver(void)
{
#ifdef CONFIG_ANDROID_POWER
	early_suspend.level = ANDROID_EARLY_SUSPEND_LEVEL_DISABLE_FB;
	early_suspend.suspend = mddi_early_suspend;
	early_suspend.resume = mddi_early_resume;
	android_register_early_suspend(&early_suspend);
#endif

	return platform_driver_register(&mddi_driver);
}

static int __init mddi_driver_init(void)
{
	int ret;

	mddi_clk = clk_get(NULL, "pmdh_clk");
	if (IS_ERR(mddi_clk)) {
		printk(KERN_ERR "can't find pmdh_clk\n");
		return PTR_ERR(mddi_clk);
	}
	clk_enable(mddi_clk);

	ret = mddi_register_driver();
	if (ret) {
		clk_disable(mddi_clk);
		clk_put(mddi_clk);
		printk(KERN_ERR "mddi_register_driver() failed!\n");
		return ret;
	}
	mddi_init();

	return ret;
}

module_init(mddi_driver_init);
