# Script to test the clock driver.

verbose=0
total=0
passed=0
failed=0
skipped=0
rate_checked=0

high_rate=0
low_rate=0
skip_reason="Test error"

# If a clock is not listed here and hits the default rate of 0, it indicates
# that the rate for the clock is read-only.
get_high_rate ()
{
  case "$1" in
    mdp_lcdc_pclk_clk)
      high_rate=64000000;;
    icodec_*)
      high_rate=12288000;;
    sdc*_clk)
      high_rate=49152000;;
    vfe_clk)
      high_rate=64000000;;
    uart*dm_clk)
      high_rate=64000000;;
    *)
      high_rate=0;;
  esac
}

get_low_rate ()
{
  case "$1" in
    mdp_lcdc_pclk_clk)
      low_rate=30720000;;
    icodec_*)
      low_rate=3072000;;
    sdc*_clk)
      low_rate=20000000;;
    vfe_clk)
      low_rate=36000000;;
    uart*dm_clk)
      low_rate=48000000;;
    *)
      low_rate=0;;
  esac
}

is_testable ()
{
  # Skip clocks that shouldn't be messed with.
  # vfe_mdc_clk is skipped for now because it's writable only for 8k. Once
  # the target info is available in sysfs, one could fix the script to black
  # list the clocks based on the target.
  for skip_clk in acpu_clk pmdh_clk uart2_clk uart2dm_clk vdc_clk vfe_mdc_clk;
  do
    if [ $1 = $skip_clk ]; then
      skip_reason="blacklisted"
      return 0
    fi
  done

  # Skip testing a clock if it's already enabled. Might be used by something
  # else.
  status=`cat $1`
  if [ $status -eq 1 ]; then
    skip_reason="already enabled"
    return 0
  fi

  return 1
}

print ()
{
  if [ $verbose -gt $1 ]; then
    echo $2
  fi
}

# Parse options
while [ $# -ge 1 ]
do
  case $1 in
    -v|--verbose)
      verbose=$(( $verbose + 1 ))
  esac
  shift
done

cd /debug/clk_rate/ 2>/dev/null
if [ $? -ne 0 ]; then
  echo "Unable to find /debug/clk_rate/"
  echo "Debug fs support required to proceed with tests."
  exit 1;
fi

cd /debug/clk_enable/ 2>/dev/null
if [ $? -ne 0 ]; then
  echo "Unable to find /debug/clk_enable/"
  echo "Debug fs support required to proceed with tests."
  exit 1;
fi

echo "Testing clocks:"
echo "==============="

for clk_name in *
do
  total=$(( $total + 1 ))

  # Check if it's okay to test the clock
  is_testable $clk_name
  if [ $? -ne 1 ]; then
    print 1 "$clk_name... Skipped ($skip_reason)"
    skipped=$(( $skipped + 1 ))
    continue;
  fi

  # Some clocks need the rate to be set before they can be enabled. So we do
  # the rate test before doing the enable/disable test. We use two test rates
  # to prevent false positives when the old rate is same as the test rate.
  get_high_rate $clk_name
  get_low_rate $clk_name

  # If the clock rate is read only, then skip the rate tests.
  if [ $high_rate -ne 0 ]; then

    rate_checked=$(( $rate_checked+ 1 ))
    old_rate=`cat ../clk_rate/$clk_name`

    echo $low_rate > ../clk_rate/$clk_name
    rate=`cat ../clk_rate/$clk_name`

    if [ $rate -ne $low_rate ]; then
      print 0 "$clk_name... Failed setting low rate ($low_rate Hz)."
      failed=$(( $failed + 1 ))
      # Disable the failed clock to avoid skipping it in future runs of this
      # test suite.
      echo "0" > $clk_name
      continue
    fi

    echo $high_rate > ../clk_rate/$clk_name
    rate=`cat ../clk_rate/$clk_name`

    if [ $rate -ne $high_rate ]; then
      print 0 "$clk_name... Failed setting high rate ($high_rate Hz)."
      failed=$(( $failed + 1 ))
      # Disable the failed clock to avoid skipping it in future runs of this
      # test suite.
      echo "0" > $clk_name
      continue
    fi

    # If old rate is 0 then don't set it because clocks on some targets
    # can't be enabled if the rate is 0.
    if [ $old_rate -ne 0 ]; then
      echo $old_rate > ../clk_rate/$clk_name
    fi

    # Some targets enable the clock when a set rate is performed. Disable
    # those clocks so that an explicit enable can be tested.
    echo "0" > $clk_name
  fi

  # Enable clock and verify
  echo "1" > $clk_name
  status=`cat $clk_name`

  if [ $status -eq 0 ]; then
    print 0 "$clk_name... Failed enabling."
    failed=$(( $failed + 1 ))
    continue
  fi

  # Disable clock and verify
  echo "0" > $clk_name
  status=`cat $clk_name`

  if [ $status -eq 1 ]; then
    print 0 "$clk_name... Failed disabling."
    failed=$(( $failed + 1 ))
    continue
  fi

  # Done testing this clock
  if [ $high_rate -ne 0 ]; then
    print 2 "$clk_name... Passed both."
  else
    print 2 "$clk_name... Passed."
  fi
  passed=$(( $passed + 1 ))

done

echo
echo "Passed:         $passed"
echo "Failed:         $failed"
echo "Skipped:        $skipped"
echo "Rate checked:   $rate_checked"
echo "Total:          $total"
echo

# If any test failed or if no clocks were checked for set rate then indicate
# an error
if [ $failed -ne 0 -o $rate_checked -lt 1 ]; then
  exit 1;
else
  exit 0;
fi
